;; Constructive skolemization
;; Matt Kaufmann, CLI

(in-package "USER")

(defun defn-sk-constructive-fn (name args body witness-alist hints defn-sk-hints)
  (let (skolemized-body new-functions-and-formals)
       (match! (chk-acceptable-defn-sk name args body defn-sk-hints)
               (list name args body defn-sk-hints
                     skolemized-body new-functions-and-formals))
       (nconc (list 'constrain (pack (list name '-intro)) '(rewrite) skolemized-body
                    witness-alist)
              (if hints (list hints) nil))))

(defmacro defn-sk-constructive
  (name args body witness-alist &optional hints defn-sk-hints)
  (defn-sk-constructive-fn name args body witness-alist hints defn-sk-hints))

(defmacro defn-sk-constructive-help (name args body)
  (let (skolemized-body new-functions-and-formals)
    (match! (chk-acceptable-defn-sk name args body hints)
            (list name args body hints
                  skolemized-body new-functions-and-formals))
    (PRINEVAL (PQUOTE (PROGN CR CR
                             |The| |Skolemized| |formula| |will| |be| |:|))
              nil
              0 PROVE-FILE)
    (PRINEVAL (PQUOTE (PROGN CR (!term skolemized-body (quote |.|)) CR
                             |The| |witnessing| |function| |alist|
                             |should| |have| |roughly| |the| |form| |:|))
              `((skolemized-body . ,skolemized-body))
              0 PROVE-FILE)
    (PRINEVAL (PQUOTE (PROGN CR (!ppr x (quote |.|)) CR CR))
              `((X . ,(iterate for fn-formals in new-functions-and-formals
                               collect (list (car fn-formals)
                                             '???))))
              0 PROVE-FILE)
    t))

(defun defn-sk-nonconstructive-fn (name args body defn-sk-hints)
  (let (skolemized-body new-functions-and-formals)
    (match! (chk-acceptable-defn-sk name args body defn-sk-hints)
            (list name args body defn-sk-hints
                  skolemized-body new-functions-and-formals))
    `(let (undone-events)
       (do-events 
        ',(iterate for fn-args in new-functions-and-formals
                   with temp
                   collect
                   `(dcl ,(car fn-args) ,(cdr fn-args))
                   into temp
                   finally
                   (return
                    (append temp
                            `((add-axiom ,(pack (list name '-intro))
                                         (rewrite)
                                         ,(untranslate skolemized-body))
                              (add-axiom ,(pack (list name '-boolean))
                                         (rewrite)
                                         (or (truep (,name ,@args))
                                             (falsep (,name ,@args))))))))))))

(defmacro defn-sk-nonconstructive (name args body &optional defn-sk-hints)
  (defn-sk-nonconstructive-fn name args body defn-sk-hints))

(defun functionally-one-way-unify-rec (pat term fns acc &aux pair)
  ;; Returns a functional substitution fs with domain contained in
  ;; fns, no member of which is a constructor, such that
  ;; pat \ (fs U acc) = term; or else returns 'fail.  Here we're doing very
  ;; simple one-way unification -- no fancy higher-order stuff,
  ;; and every function symbol that's bound is bound to a symbol.
  ;; Let's say that the "free" function symbols are those in fns
  ;; which aren't bound in acc.
  ;; We assume that no function symbol of fns occurs in term.
  (cond
   ((variablep pat)
    (if (eq pat term)
        acc
      'fail))
   ((fquotep pat)
    (if (equal pat term)
        acc
      'fail))
   ((variablep term)
    'fail)
   ((eq (ffn-symb pat) (fn-symb term))
    ;; then by assumption, (ffn-symb pat) is not in fns, and hence
    ;; we'll never replace it
    (functionally-one-way-unify-lst (fargs pat) (sargs term) fns acc))
   ((setq pair (assoc-eq (ffn-symb pat) acc))
    (if (eq (cdr pair) (fn-symb term))
        (functionally-one-way-unify-lst (fargs pat) (sargs term) fns acc)
      'fail))
   ((not (member-eq (ffn-symb pat) fns))
    'fail)
   ((= (arity (ffn-symb pat)) (arity (fn-symb term)))
    ;; otherwise we are allowed to bind the function symbol of pat
    (functionally-one-way-unify-lst
     (fargs pat)
     (sargs term)
     fns
     (cons (cons (ffn-symb pat) (fn-symb term)) acc)))
   (t
    'fail)))

(defun functionally-one-way-unify-lst (patlist termlist fns acc)
  (if patlist
      (let ((new-acc (functionally-one-way-unify-rec
                      (car patlist) (car termlist) fns acc)))
        (if (eq new-acc 'fail)
            'fail
          (functionally-one-way-unify-lst
           (cdr patlist) (cdr termlist) fns new-acc)))
    acc))

(defun functionally-one-way-unify (patlist termlist fns)
  (functionally-one-way-unify-rec patlist termlist fns nil))

(defun fnnamep-formula (fn x &aux (args (formula-args x)) (op (formula-op x)))
  (cond
   ((eq op 'term)
    (fnnamep fn args))
   ((member-eq op '(forall exists))
    (fnnamep-formula fn (cadr args)))
   (t (iterate for arg in args
               thereis (fnnamep-formula fn arg)))))

(defun is-skolemization (term formula &optional translate-flg)
  ;; adapted from CHK-ACCEPTABLE-DEFN-SK
  (let ((translated-formula
         (if translate-flg
             (translate-to-formula formula)
           formula))
        (untranslated-formula
         (if translate-flg
             formula
           (untranslate-formula formula))))
    (when translate-flg (setq term (translate term)))
    (let ((x (skolemize translated-formula t))
          (arity-alist arity-alist))
      (iterate for pair in *new-functions-and-formals*
               do
               (setq arity-alist
                     (cons (cons (car pair) (length (cdr pair)))
                           arity-alist)))
      (setq unify-subst (functionally-one-way-unify
                         x
                         term
                         (mapcar 'car *new-functions-and-formals*)))
      (if (eq unify-subst 'fail)
          (progn (PRINEVAL (PQUOTE (PROGN CR |The| |Skolemization| |of| CR
                                          (!ppr untranslated-formula nil)
                                          |is| CR (!term x nil) |,| |but| |the| |term|
                                          (!term term nil) |is| |not| |a|
                                          |functional| |instance|
                                          |of| |that| |term| |.| CR CR))
                           `((untranslated-formula . ,untranslated-formula)
                             (x . ,x)
                             (term . ,term))
                           0 PROVE-FILE)
                 nil)
        t))))

(defun check-quant-with-fs (name witness-alist)
  ;; This is adapted from Nqthm code for CHK-ACCEPTABLE-CONSTRAIN.
  ;; Remember, the basic idea is that to show that a quantified
  ;; notion is correctly represented, we need to show that a
  ;; functional instantiation of a Skolemization is a theorem.
  ;; I'm probably could allow lambda expressions, but I won't
  ;; at this time.

;   It is important that the function symbols being witnessed, i.e.,
;   the new function symbols, not occur in the terms in the range of
;   the substitution.  This is insured by our not binding ARITY-ALIST
;   until after the TRANSLATE on the range terms.

  (MATCH! (CHK-ACCEPTABLE-FUNCTIONAL-SUBSTITUTION WITNESS-ALIST T)
          (LIST WITNESS-ALIST))
  (OR (NO-DUPLICATESP (CONS NAME (ITERATE FOR DOUBLET IN WITNESS-ALIST
                                          COLLECT (CAR DOUBLET))))
      (ER SOFT NIL
          |it| |is| |illegal| |to| |use| |the| |name| |of| |a|
          |Skolem| |function|
          |for| |the| |name| |of| |the| |function| |that| |is|
          |representing| |the| |quantified| |notion| |,|
          |or| |to| |duplicate| |a| |function| |in| |the| |domain| |of| |the|
          |witnessing| |alist| |of| |the| |constraint| |.|))
  (or (iterate for doublet in witness-alist always (symbolp (cadr doublet)))
      (er soft nil |We| |do| |not| |currently| |allow| |lambda| |expressions|
          |in| |the| |witnessing| |alist| |for| |CHECK-QUANT| |.|))
  WITNESS-ALIST)

(defun check-quant-fn (NAME ARGS BODY event-name-or-witness-alist hints &aux formula)
  ;; Checks that an existing notion NAME defines the notion in BODY
  ;; in the sense that (IFF (NAME . ARGS) BODY).
  (or (get name 'type-prescription-lst)
      (er soft (name) (!ppr name nil) |is| |not| |a| |function|
          |known| |in| |the| |current| |history| |.|))
  (chk-arglist name args)
  (setq formula (translate-to-formula body))
  (if (fnnamep-formula name formula)
      (progn (PRINEVAL (PQUOTE (PROGN CR |The| |function| |symbol| (!ppr name nil)
                                      |occurs| |in| |the| |formula| (!ppr body nil) |,|
                                      |and| |hence| (!ppr x nil) |is| |not| |what|
                                      |we| |call| |a|
                                      |representation| |of| |the| |indicated| 
                                      |quantified| |notion| |.| CR CR))
                       `((name . ,name)
                         (body . ,body)
                         (x . (,name ,@args)))
                       0 PROVE-FILE)
             (return-from check-quant-fn nil)))
  ;; could probably omit the following
  (free-var-chk-formula name args formula)
  (cond
   ((symbolp event-name-or-witness-alist)
    (if (formula-of event-name-or-witness-alist)
        (is-skolemization (formula-of event-name-or-witness-alist)
                          (make-defn-sk-formula name args formula))
      (progn (PRINEVAL (PQUOTE (PROGN CR |The| |symbol|
                                      (!ppr event-name-or-witness-alist nil)
                                      |is| |not| |the| |name| |of| |an| |event|
                                      |with| |a| |formula| |.| CR CR))
                       `((event-name-or-witness-alist . ,event-name-or-witness-alist))
                       0 PROVE-FILE)
             nil)))
   (t
    ;; Otherwise, we check that the result of functionally substituting
    ;; the witness-alist into the Skolemization is really a theorem, even
    ;; though it would in fact be legal to further substitute and get a
    ;; theorem then instead.
    (let ((term (skolemize (make-defn-sk-formula name args formula) t))
          (arity-alist arity-alist))
      (iterate for pair in *new-functions-and-formals*
               do
               (setq arity-alist
                     (cons (cons (car pair) (length (cdr pair)))
                           arity-alist)))
      (iterate for pair in event-name-or-witness-alist
               do (or (assoc-eq (car pair) *new-functions-and-formals*)
                      (er soft ((fn (car pair))) |The| |function| |symbol|
                          (!ppr fn nil) |appears| |in| |the| |domain| |of|
                          |the| |witnessing| |alist| |but| |is| |not| |a|
                          |Skolem| |function| |generated| |by| |the|
                          |Skolemizer| |in| |the| |current| |context| |.|)))
      (iterate for pair in *new-functions-and-formals*
               do (or (assoc-eq (car pair) event-name-or-witness-alist)
                      (er soft ((fn (car pair)) term) |The| |function| |symbol|
                          (!ppr fn nil) |was| |generated| |as| |a|
                          |Skolem| |function| |but| |does| |not| |appear|
                          |in| |the| |witnessing| |alist| |that|
                          |was| |supplied| |.| |The| |Skolemization|
                          |is| |:| CR (!term term '|.|))))
      (let ((term-to-prove
             (sublis-fn (iterate for doublet
                                 in (check-quant-with-fs
                                     name event-name-or-witness-alist)
                                 collect
                                 (cons (car doublet)
                                       (cadr doublet)))
                        term)))
        (PRINEVAL (PQUOTE (PROGN CR |Our| |proof| |obligation| |is| |:|
                                 (!ppr term-to-prove '|.|) CR CR))
                  `((term-to-prove . ,term-to-prove))
                  0 PROVE-FILE) 
        ;; The following is adapted from the code for PROVE-LEMMA
        (LET ((IN-PROVE-LEMMA-FLG T) PROVE-ANS)
             (MATCH! (CHK-ACCEPTABLE-HINTS HINTS)
                     (LIST HINTS))
             (UNWIND-PROTECT
              (progn
                ;; Call translate first so that if arities is wrong, we
                ;; find out now.  Kind of a shame, since PROVE calls translate
                ;; too, but that's OK.
                (or (error1-set (setq term-to-prove
                                      (translate (APPLY-HINTS HINTS TERM-to-prove))))
                    (er soft nil |Aborting| |,| |since| |our| |proof| |obligation|
                        |is| |not| |well-formed| |.|))
                (and (PROVE term-to-prove)
                     t))
              (PROGN (ITERATE FOR X IN HINT-VARIABLE-ALIST
                              DO (SET (CADR X) (CADDDR X)))
                     (SETQ LOCAL-DISABLEDP-HASH-FLAG NIL)))))))))

(defmacro check-quant (&optional NAME ARGS BODY event-name hints)
  (if (or (null name)
          (null body)
          (null event-name))
      (er soft nil CHECK-QUANT |has| |four| |required| |arguments| |.|)
    (if (and event-name (symbolp event-name) hints)
        (er soft (event-name hints) |Since| |a| |previous| |event|
            |name| |was| |specified| |,| |namely| (!ppr event-name nil)
            |,| |it| |is| |not| |sensible| |or| |legal| |to| |supply|
            |hints| |here| |.| |The| |final| |argument| (!ppr hints nil)
            |to| CHECK-QUANT |is| |thus| |illegal| |.|)
      `(check-quant-fn ',name ',args ',body ',event-name ',hints))))

(defun check-quant-help-fn (NAME ARGS BODY &aux formula)
  ;; Checks that an existing notion NAME defines the notion in BODY
  ;; in the sense that (IFF (NAME . ARGS) BODY).
  (or (arity name)
      (er soft (name) (!ppr name nil) |is| |not| |a| |function|
          |known| |in| |the| |current| |history| |.|))
  (chk-arglist name args)
  (setq formula (translate-to-formula body))
  ;; could probably omit the following
  (free-var-chk-formula name args formula)
  (let ((x (skolemize (make-defn-sk-formula name args formula) t)))
    (PRINEVAL (PQUOTE (PROGN CR |The| |new| |function|
                             (plural?
                              y
                              (progn |symbols| |with| |their| |argument| |lists| |are|)
                              (progn |symbol| |with| |its| |argument| |list| |is|))
                             (!ppr-list y (quote |.|)) CR CR))
              `((y . ,*new-functions-and-formals*))
              0 PROVE-FILE)
    (untranslate x)))

(defmacro check-quant-help (name args body)
  `(check-quant-help-fn ',name ',args ',body))

#|
(bdd-forall-sk-model-aux qvar upper args body)
(defn forall-p4-sk-model-aux (x i j y)
  (if (lessp x j)
      (if (not (p4 (fix x) i j y))
          (fix x)
        (forall-p4-sk-model-aux (add1 x) i j y))
    j)
  ((lessp (difference (add1 j) x))))

|#
(defun bdd-forall-sk-model-aux (bdd-forall-sk-model-aux-name qvar upper args body)
  `(defn ,bdd-forall-sk-model-aux-name ,(cons qvar args)
     (if (lessp ,qvar ,upper)
         (if (not ,body)
             ,qvar
           (,bdd-forall-sk-model-aux-name (add1 ,qvar) ,@args))
       ,upper)
     ((lessp (difference (add1 ,upper) ,qvar)))))

#|
(bdd-forall-sk-model lower args)
(defn forall-p4-sk-model (i j y)
  (forall-p4-sk-model-aux (fix i) i j y))

|#
(defun bdd-forall-sk-model
  (bdd-forall-sk-model-aux-name bdd-forall-sk-model-name lower args)
  `(defn ,bdd-forall-sk-model-name ,args
     (,bdd-forall-sk-model-aux-name (fix ,lower) ,@args)))

#|
(bdd-forall-model-aux qvar upper args body)
(defn forall-p4-model-aux (x i j y)
  (if (lessp x j)
      (if (p4 x i j y)
          (forall-p4-model-aux (add1 x) i j y)
        f)
    t)
  ((lessp (difference (add1 j) x))))

|#
(defun bdd-forall-model-aux (bdd-forall-model-aux-name qvar upper args body)
  `(defn ,bdd-forall-model-aux-name ,(cons qvar args)
     (if (lessp ,qvar ,upper)
         (if ,body
             (,bdd-forall-model-aux-name (add1 ,qvar) ,@args)
           f)
       t)
     ((lessp (difference (add1 ,upper) ,qvar)))))

#|
(bdd-forall-model lower args)
(defn forall-p4-model (i j y)
  (forall-p4-model-aux (fix i) i j y))

|#
(defun bdd-forall-model
  (bdd-forall-model-aux-name bdd-forall-model-name lower args)
  `(defn ,bdd-forall-model-name ,args
     (,bdd-forall-model-aux-name (fix ,lower) ,@args)))

#|
(bdd-forall-model-aux-necc qvar upper args body)
(prove-lemma forall-p4-model-aux-necc ()
  (implies (not (implies (and (numberp x)
                              (numberp x**)
                              (not (lessp x x**))
                              (lessp x j))
                         (p4 x i j y)))
           (not (forall-p4-model-aux x** i j y))))

|#
(defun bdd-forall-model-aux-necc (bdd-forall-model-aux-necc-name
                                  bdd-forall-model-aux-name
                                  qvar qvar** upper args body)
  `(prove-lemma
    ,bdd-forall-model-aux-necc-name
    ()
    (implies (not (implies (and (numberp ,qvar)
                                (numberp ,qvar**)
                                (not (lessp ,qvar ,qvar**))
                                (lessp ,qvar ,upper))
                           ,body))
             (not (,bdd-forall-model-aux-name ,qvar** ,@args)))))

#|
(bdd-forall-model-necc qvar lower upper args body)
(prove-lemma forall-p4-model-necc ()
  (implies (not (implies (and (numberp x)
                              (not (lessp x i))
                              (lessp x j))
                         (p4 x i j y)))
           (not (forall-p4-model i j y)))
  ((use (forall-p4-model-aux-necc (x** (fix i))))))

|#
(defun bdd-forall-model-necc (bdd-forall-model-necc-name
                              bdd-forall-model-name
                              bdd-forall-model-aux-necc-name
                              qvar qvar** lower upper args body)
  `(prove-lemma
    ,bdd-forall-model-necc-name
    ()
    (implies (not (implies (and (numberp ,qvar)
                                (not (lessp ,qvar ,lower))
                                (lessp ,qvar ,upper))
                           ,body))
             (not (,bdd-forall-model-name ,@args)))
    ((use (,bdd-forall-model-aux-necc-name (,qvar** (fix ,lower)))))))

#|
(bdd-forall-model-aux-suff qvar upper args body)
(prove-lemma forall-p4-model-aux-suff ()
  (implies (and (numberp x)
                (not (lessp x i))
                (let ((x (forall-p4-sk-model-aux x i j y)))
                  (implies (and (numberp x)
                                (not (lessp x i))
                                (lessp x j))
                           (p4 x i j y))))
           (forall-p4-model-aux x i j y))
  ((induct (forall-p4-model-aux x i j y))))

|#
(defun bdd-forall-model-aux-suff (bdd-forall-model-aux-suff-name
                                  bdd-forall-sk-model-aux-name
                                  bdd-forall-model-aux-name
                                  qvar lower upper args body)
  `(prove-lemma
    ,bdd-forall-model-aux-suff-name
    ()
    (implies (and (numberp ,qvar)
                  (not (lessp ,qvar i))
                  (let ((,qvar (,bdd-forall-sk-model-aux-name ,qvar ,@args)))
                    (implies (and (numberp ,qvar)
                                  (not (lessp ,qvar ,lower))
                                  (lessp ,qvar ,upper))
                             ,body)))
             (,bdd-forall-model-aux-name ,qvar ,@args))
    ((induct (,bdd-forall-model-aux-name ,qvar ,@args)))))

#|
(bdd-forall-model-suff qvar upper args body)
(prove-lemma forall-p4-model-suff ()
  (let ((x (forall-p4-sk-model i j y)))
    (implies (implies (and (numberp x)
                           (not (lessp x i))
                           (lessp x j))
                      (p4 x i j y))
             (forall-p4-model i j y)))
  ((use (forall-p4-model-aux-suff (x (fix i))))))

|#
(defun bdd-forall-model-suff (bdd-forall-model-aux-suff-name
                              bdd-forall-model-suff-name
                              bdd-forall-sk-model-name
                              bdd-forall-model-name
                              qvar lower upper args body)
  `(prove-lemma
    ,bdd-forall-model-suff-name
    ()
    (let ((,qvar (,bdd-forall-sk-model-name ,@args)))
      (implies (implies (and (numberp ,qvar)
                             (not (lessp ,qvar ,lower))
                             (lessp ,qvar ,upper))
                        ,body)
               (,bdd-forall-model-name ,@args)))
    ((use (,bdd-forall-model-aux-suff-name (,qvar (fix ,lower)))))))

#|
(bdd-forall-defn-sk-constructive name qvar lower upper args body)
;; x-1 here is (make-new-skolem-fn-1 fn 0 (list (cons name args)))
(defn-sk-constructive forall-p4 (i j y)
  (forall x (implies (and (numberp x) (not (lessp x i)) (lessp x j))
                     (p4 x i j y)))
  ((x-1 forall-p4-sk-model)
   (forall-p4 forall-p4-model))
  ((use (forall-p4-model-necc)
        (forall-p4-model-suff))
   (disable-theory t)
   (enable-theory ground-zero)))

|#
(defun bdd-forall-defn-sk-constructive
  (name bdd-forall-sk-model-name bdd-forall-model-necc-name
        bdd-forall-model-suff-name bdd-forall-model-name
        qvar lower upper args body)
  `(defn-sk-constructive ,name ,args
     (forall ,qvar (implies (and (numberp ,qvar)
                                 (not (lessp ,qvar ,lower))
                                 (lessp ,qvar ,upper))
                            ,body))
     ((,(make-new-skolem-fn-1 qvar 0 (list (cons name args)))
       ,bdd-forall-sk-model-name)
      (,name ,bdd-forall-model-name))
     ((use (,bdd-forall-model-necc-name)
           (,bdd-forall-model-suff-name))
      (disable-theory t)
      (enable-theory ground-zero))))

(defun bdd-forall-events (name qvar lower upper args body)
  ;; defines (name .args) to be
  ;; (forall qvar (implies (and (numberp qvar)
  ;;                            (not (lessp qvar lower))
  ;;                            (lessp qvar upper))
  ;;                       body))
  (chk-new-name name t)
  (chk-arglist name args)
  (if (or (nvariablep qvar)
          (member-eq qvar args))
      (er soft (qvar args) |The| |quantified| |variable| (!ppr qvar nil)
          |is| |supposed| |to| |be| |a| |variable| |that| |is| |not| |a|
          |member| |of| |the| |argument| |list| (!ppr args (quote |.|))))
  ;; let's make sure these all translate before we even start
  (translate lower)
  (translate upper)
  (translate body)
  (let ((bdd-forall-sk-model-aux-name (pack (list name '-sk-model-aux)))
        (bdd-forall-sk-model-name (pack (list name '-sk-model)))
        (bdd-forall-model-aux-name (pack (list name '-model-aux)))
        (bdd-forall-model-name (pack (list name '-model)))
        (bdd-forall-model-aux-necc-name (pack (list name '-model-aux-necc)))
        (qvar** (pack (list qvar '**)))
        (bdd-forall-model-necc-name (pack (list name '-model-necc)))
        (bdd-forall-model-aux-suff-name (pack (list name '-model-aux-suff)))
        (bdd-forall-model-suff-name (pack (list name '-model-suff))))
    (list (bdd-forall-sk-model-aux
           bdd-forall-sk-model-aux-name qvar upper args body)
          (bdd-forall-sk-model
           bdd-forall-sk-model-aux-name bdd-forall-sk-model-name lower args)
          (bdd-forall-model-aux bdd-forall-model-aux-name qvar upper args body)
          (bdd-forall-model bdd-forall-model-aux-name bdd-forall-model-name lower args)
          (bdd-forall-model-aux-necc
           bdd-forall-model-aux-necc-name bdd-forall-model-aux-name
           qvar qvar** upper args body)
          (bdd-forall-model-necc
           bdd-forall-model-necc-name bdd-forall-model-name
           bdd-forall-model-aux-necc-name
           qvar qvar** lower upper args body)
          (bdd-forall-model-aux-suff
           bdd-forall-model-aux-suff-name
           bdd-forall-sk-model-aux-name
           bdd-forall-model-aux-name
           qvar lower upper args body)
          (bdd-forall-model-suff
           bdd-forall-model-aux-suff-name
           bdd-forall-model-suff-name
           bdd-forall-sk-model-name
           bdd-forall-model-name
           qvar lower upper args body)
          (bdd-forall-defn-sk-constructive
           name bdd-forall-sk-model-name bdd-forall-model-necc-name
           bdd-forall-model-suff-name bdd-forall-model-name
           qvar lower upper args body))))

(defmacro bdd-forall (name qvar lower upper args body)
  `(do-events (bdd-forall-events
               ',name ',qvar ',lower ',upper ',args
               ',body)))

