;; Here is the file to be loaded for running demos.  To run the demos:

;; 1. Optionally, edit the following form by replacing NIL with the
;;     current directory, such as

;; (defvar *demo-directory* "/usr/home/smith/nqthm/")

;;     Don't forget the trailing slash (in Unix)!

(defvar *demo-directory* NIL)

(or *demo-directory*
    (progn (format t "Please enter a directory name, including the separator (such as ~%~
                      `/' for Unix systems), without quotes, and then hit <RETURN> ~%~
                      (suggestion:  ./) ")
	   (setq *demo-directory*
		 (iterate with temp
			  while (equal (setq temp (read-line t)) "")
			  finally (return temp)))
	   (format t "~&Running from directory ~a" *demo-directory*)))

;; 2. (Optional) Replace T by NIL below if you are running kcl or akcl on
;;    top of Unix, and you prefer scrolling to clearing the screen.  (The
;;    current setting is probably optimal, though.  And, it is probably better
;;    not to run these demos under emacs, though they will work there.)

(defvar *clear-screen-for-execute-at-prompt*
  #+(and kcl unix) T
  #-(and kcl unix) NIL)

;; 3. Go to a shell in a window with at least 37 lines.
;; 4. In that shell, change to the directory of this file with the cd command.
;; 5. In that shell, start up the Boyer-Moore theorem prover
;; 6. Type (load "run.lisp") in that shell.
;; 7. Finally, hit the <RETURN> key in that shell and follow the directions.

(defun clear-screen ()
  #+(and kcl unix)
  (cond
   (*clear-screen-for-execute-at-prompt*
    (system "clear")
    t)
   (t nil))
  #-(and kcl unix)
  nil)

(defmacro no-clear (form)
  `(let ((*clear-screen-for-execute-at-prompt* nil))
     ,form))

(defun clear-screen-p ()
  #+(and kcl unix)
  *clear-screen-for-execute-at-prompt*
  #-(and kcl unix)
  nil)

(clear-screen)

(defvar *demo-disclaimer-printed* nil)

(defun print-demo-disclaimer ()
  (if *demo-disclaimer-printed*

      (format t "~&
;;;; Remember, our goal here is to illustrate a style of mathematical
;;;; modeling, but to give only a taste of how the Boyer-Moore theorem
;;;; prover is used.~&")

    (progn (format t "~&
;;;; WELCOME!  We hope that these demos of the Boyer-Moore theorem
;;;; prover will give you some idea of how it is used and what it can
;;;; do.

;;;; AUTHORSHIP/ACKNOWLEDGEMENT:  These demos have been put together
;;;; by Matt Kaufmann of Computational Logic, Inc., who thanks Ken
;;;; Albin, Rich Cohen, John Cowles, Art Flatau, Don Good, and Warren
;;;; Hunt for a number of useful suggestions.

;;;; DISCLAIMER:  We are attempting here to illustrate a methodology
;;;; for mathematical modeling that is amenable to the application of
;;;; automated reasoning.  We are also attempting to give some flavor
;;;; for how that automated reasoning takes place.  However, we do NOT
;;;; intend in this short demo to try to give more than a brief
;;;; introduction to the use of the Boyer-Moore theorem prover.
;;;; Therefore, we do not attempt to translate its output into
;;;; traditional prefix notation.  By the way, most users of this
;;;; system would probably agree with the following statement:  Even
;;;; though this notation may appear at first to be rather arcane, it
;;;; tends to appear quite natural after one has begun to use the
;;;; prover in earnest.

;;;; WARNING:  Therefore, do not attempt to read the formulas printed
;;;; out by the theorem prover.  Instead, read the comments that have
;;;; been inserted.

;;;; NOTE:  All text appears on lines beginning with four semicolons,
;;;; like this one, EXCEPT for text that is actual theorem prover
;;;; output.  Again, we suggest that you concentrate primarily on
;;;; those textual comments, and not too much on the prover's output.
")
	   (setq *demo-disclaimer-printed* t))))

(print-demo-disclaimer)

;; Recall that DEFVAR will not re-set the value of an already-set
;; variable.
(defvar *do-not-query-for-demo* nil)

(defmacro eval-if-query (form &optional otherwise)
  (if *do-not-query-for-demo*
      (or otherwise t)
    form))

(defmacro my-y-or-n-p (format-string &rest args)
  `(progn (format t "~&~%|||| ")
	  (format t ,format-string ,@args)
	  (format t "  (Y or N) ")
	  (eval-if-query
	   (loop
	    (let ((ans (string-upcase (string-trim " " (read-line *terminal-io*)))))
	      (cond
	       ((equalp ans ""))
	       ((equalp ans "Y")
		(return t))
	       ((equalp ans "N")
		(return nil))
	       (t (format t ";;;; You were supposed to answer Y or N, but
;;;; we'll take your answer of ~s as an N.~%"
			  ans)
		  (return nil)))))
	   (progn (format t "Y~%")
		  t))))

(defvar *interactive-demo-session* nil)

(defmacro my-y-or-n-p! (format-string &rest args)
  `(let ((*do-not-query-for-demo* *interactive-demo-session*))
    (my-y-or-n-p ,format-string ,@ args)))

(defparameter *run-all-demos*
  (my-y-or-n-p! "Do you want to run all of the demos?"))

(defmacro execute-at-prompt (format-string form &rest format-args)
  `(progn
     (format t ,format-string ,@format-args)
     (format t "~&
+++++++++++++ Hit <RETURN> to submit the following form +++++++++++++
")
     (ppr ',form nil)
     (eval-if-query
      (progn (read-line *terminal-io*)
	     (or (clear-screen)
		 (format t "~%")))
      (format t "~%"))
     (car (error1-set ,form))))

(defmacro format-when-ready (format-string &rest format-args)
  `(progn (format t "~&
++++++++++++++++ Hit <RETURN> when ready to continue ++++++++++++++++")
	  (eval-if-query
	   (progn (read-line *terminal-io*)
		  (or (clear-screen)
		      (format t "~%")))
	   (format t "~%~%"))
	  (format t ,format-string ,@format-args)
	  (values)))

(defparameter *just-did-one* nil)

(defun print-demo-separator ()
  (when *just-did-one*
	(format t "~&~%========================================~%~%")
	(format t "Are you ready for the next demo?.....")
	(format-when-ready ""))
  (setq *just-did-one* nil))

(defun demo-file (filename)
  (concatenate 'string *demo-directory* filename))

;; factorial-sum.ev:  simple introduction, with no lists and no lemma
;;                    discovery, but with recursion
;;;(print-demo-separator) -- no need for the first one!
(when (or *run-all-demos*
	  (my-y-or-n-p! "Do you want to begin a demo of the Boyer-Moore theorem prover,
        on ``A Simple Example with Factorials''?"))

      (load (demo-file "factorial-sum.ev")
	    :verbose nil)
      (setq *just-did-one* t))

;; subset-append-append.ev:  lemma discovery
(print-demo-separator)
(when (or *run-all-demos*
	  (my-y-or-n-p! "Do you want to begin a demo of the Boyer-Moore theorem prover,
        demonstrating lemma discovery?"))

      (load (demo-file "subset-append-append.ev")
	    :verbose nil)
      (setq *just-did-one* t))

;; bv-to-nat.ev:  more complicated hardware intro (naturals to bit-vectors
;;                and back), with bit-vectors as lists of 0s and 1s.  Lemma
;;                discovery is needed both for definition acceptance and for
;;                proof of theorem.
(print-demo-separator)
(when (or *run-all-demos*
	  (my-y-or-n-p! "Do you want to begin a demo of the Boyer-Moore theorem prover,
        about converting bit vectors to natural numbers?"))

      (load (demo-file "bv-to-nat.ev")
	    :verbose nil)
      (setq *just-did-one* t))

;; adder.ev:  simple hardware intro (full adder module correctness),
;;            with bit-vectors as lists of booleans.  No lemma discovery.
;;            It requires the LET patch.
(print-demo-separator)
(when (or *run-all-demos*
	  (my-y-or-n-p! "Do you want to begin a demo of the Boyer-Moore theorem prover,
        on ``A Very Simple Introduction to Hardware Proofs''?"))

(when (not (propertyless-symbolp 'let))

(clear-screen)

(format t "~%NOTE:  We are about to make minor modifications to the
theorem prover in order to support LET syntax.~%")

(DEFUN TRANSLATE (X)

;   TRANSLATE may be given any noncircular lisp object -- unlike almost all of
;   the other functions in the system, which expect terms or evgs.  TRANSLATE
;   either causes an error or returns a term.  TRANSLATE is intimately
;   connected with the functions TERMP and UNTRANSLATE.	 The former recognizes
;   the output of TRANSLATE; the latter inverts output to acceptable input.

  (COND
   ((ATOM X)
    (COND ((INTEGERP X) (LIST (QUOTE QUOTE) X))
	  ((SYMBOLP X)
	   (COND ((EQ X T) TRUE)
		 ((EQ X (QUOTE F)) FALSE)
		 ((EQ X NIL) (QUOTE (QUOTE NIL)))
		 ((ILLEGAL-NAME X)
		  (ER SOFT (X) (!PPR X NIL) |is| |an| |illegal| |variable|
		      |name| |.|))
		 (T X)))
	  (T (ER SOFT (X) |unrecognized| |syntax:| (!PPR X (QUOTE |.|))))))
   ((NOT (CONSP X))
    (ER SOFT (X) |No| |hunks| |please:| (!PPR X (QUOTE |.|))))
   ((CDR (OUR-LAST X))
    (ER SOFT (X) |contrary| |to| |the| |rules| |of| |well-formedness| |,| |the|
	|last| CDR |of| (!PPR X NIL) |is| |non-NIL| |.|))
   ((NOT (SYMBOLP (CAR X)))
    (ER SOFT (X) |a| |function| |symbol| |must| |be| |a| |Lisp| SYMBOLP |,|
	|but| (!PPR (CAR X) NIL) |is| |not| EXCL))
   ((NOT (OK-SYMBOLP (CAR X)))
    (ER SOFT (X) (!PPR (CAR X) NIL) |is| |not| |interned| |in| |the| |right|
	|places| EXCL))
   ((PROPERTYLESS-SYMBOLP (CAR X))
    (COND ((EQ (CAR X) (QUOTE QUOTE))
	   (COND ((NOT (= 1 (LENGTH (CDR X))))
		  (ER SOFT (X) QUOTE |must| |be| |given| |exactly| |one|
		      |argument| |.| |In| (!PPR X NIL) |it| |is| |given| |the|
		      |wrong| |number| |of| |arguments| |.|))
		 ((NOT (EVG (CADR X)))
		  (ER SOFT (X) |the| |object| |QUOTEd| |in| |the| |expression|
		      (!PPR X NIL) |does| |not| |represent| |an| |explicit|
		      |value| |term| |.|))
		 (T X)))
	  ((MEMBER-EQ (CAR X) (QUOTE (NIL T F)))
	   (ER SOFT (X) (!PPR (CAR X) NIL) |is| |an| |illegal| |function|
	       |symbol| |.|))
	  ((EQ (CAR X) (QUOTE LET))
	   (COND ((NOT (= 2 (LENGTH (CDR X))))
		  (ER SOFT (X) LET |must| |be| |given| |exactly| |two| |arguments| |.| |In|
		      (!PPR X NIL) |it| |is| |given| |the| |wrong| |number| |of|
		      |arguments| |.|))
		 ((OR (CDR (OUR-LAST (CADR X)))
		      (NOT (ITERATE FOR PAIR IN (CADR X)
				    ALWAYS
				    (AND (CONSP PAIR)
					 (NULL (CDR (OUR-LAST PAIR)))
					 (= 2 (LENGTH PAIR))))))
		  (ER SOFT (X) |The| |first| |argument| |of| |a| LET |must| |be|
		      |a| |proper| |list| |of| |doublets| |,| |i.e.,| |lists| |of|
		      |the| |form| |(var term),| |and| |in| (!PPR X NIL) |it| |is|
		      |not| |.|))
		 (T (SUBLIS-VAR (ITERATE FOR PAIR IN (CADR X)
					 COLLECT
					 (COND ((NOT (VARIABLEP (TRANSLATE (CAR PAIR))))
						(ER SOFT ((VAR (CAR PAIR))
							  X)
						    |In| |the| |first| |argument| |to|
						    |the| LET |statement| (!PPR X NIL)
						    |there| |is| |an| |attempt| |to|
						    |bind| |the| |non-variable|
						    (!PPR VAR '|.|)))
					       (T (CONS (CAR PAIR)
							(TRANSLATE (CADR PAIR))))))
				(TRANSLATE (CADDR X))))))
			   
	  ((EQ (CAR X) (QUOTE LIST))
	   (COND ((NULL (CDR X)) (TRANSLATE NIL))
		 (T (XXXJOIN (QUOTE CONS)
			     (NCONC1 (ITERATE FOR ARG IN (CDR X)
					      COLLECT (TRANSLATE ARG))
				     (TRANSLATE NIL))))))
	  ((CAR-CDRP (CAR X))
	   (COND ((= 1 (LENGTH (CDR X)))
		  (FIXCAR-CDR (LIST (CAR X) (TRANSLATE (CADR X)))))
		 (T (ER SOFT (X) (!PPR (CAR X) NIL) |is| |a| |reserved|
			|abbreviation| |for| |a| CAR-CDR |nest| |and| |must|
			|be| |given| |exactly| |one| |argument| |.|))))
	  (T (ER HARD (X) PROPERTYLESS-SYMBOLP |and| TRANSLATE |do| |not|
		 |agree| |on| (!PPR (CAR X) (QUOTE |.|))))))
   ((NULL (ARITY (CAR X)))
    (COND (IN-BOOT-STRAP-FLG
	   (ER WARNING (X) (!PPR (CAR X) NIL) |has| |been| |encountered| |as|
	       |an| |undefined| |function| |by| TRANSLATE |.| |You| |should|
	       |add| |it| |to| |the| |binding| |of| ARITY-ALIST |in| BOOT-STRAP
	       |if| |you| |wish| |to| |suppress| |this| |message| EXCL))
	  (T (ER SOFT (X) |the| |function| (!PPR (CAR X) NIL) |is| |unknown|
		 |.| |Please| |delete| |all| |references| |to| |it| |,|
		 |define| |it| |or| |declare| |it| |as| |an| |undefined|
		 |function| |.|))))
   ((AND (MEMBER-EQ (CAR X) (QUOTE (AND OR PLUS TIMES)))
	 (< 2 (LENGTH (CDR X))))
    (XXXJOIN (CAR X) (ITERATE FOR ARG IN (CDR X) COLLECT (TRANSLATE ARG))))
   ((AND (EQ *1*THM-MODE$ (QUOTE NQTHM))
	 (OR (MATCH X (FOR & (QUOTE IN) & (QUOTE WHEN) & & &))
	     (MATCH X (FOR & (QUOTE IN) & & &))))
    (LET (V C R O B)
      (COND ((NOT (MATCH X (FOR V (QUOTE IN) R (QUOTE WHEN) C O B)))
	     (SETQ C T) (MATCH X (FOR V (QUOTE IN) R O B))))
      (OR (AND (SYMBOLP V) (EQ V (TRANSLATE V)))
	  (ER SOFT (V) FOR |didn't| |like| (!PPR V NIL) |as| |its| |first|
	      |argument| |.|))
      (OR (MEMBER-EQ O FOR-OPS)
	  (ER SOFT (FOR-OPS O) |A| |quantifer| |operation| |must| |be| |one|
	      |of| (!PPR-LIST FOR-OPS (QUOTE  |.|))  (!PPR O NIL)
	      |is| |not| |.|))
      (FCONS-TERM (QUOTE FOR)
		  (LIST (KWOTE V)
			(TRANSLATE R)
			(KWOTE (SETQ C (TRANSLATE C)))
			(KWOTE O)
			(KWOTE (SETQ B (TRANSLATE B)))
			(MAKE-ALIST V C B)))))
   ((NOT (= (LENGTH (CDR X)) (ARITY (CAR X))))
    (ER SOFT (X (N (ARITY (CAR X)))) |the| |function| |symbol|
	(!PPR (CAR X) NIL) |takes| |exactly| (@ N)
	(PLURAL? N |arguments| |argument|) |.| |In| (!PPR X NIL) |it| |is|
	|given| |the| |wrong| |number| |of| |arguments| |.|))
   ((MEMBER-EQ (CAR X) BOOT-STRAP-MACRO-FNS)
    (SUB-PAIR-VAR (CADR (GET (CAR X) (QUOTE SDEFN)))
		  (ITERATE FOR ARG IN (CDR X) COLLECT (TRANSLATE ARG))
		  (CADDR (GET (CAR X) (QUOTE SDEFN)))))
   (T (CONS-TERM (CAR X)
		 (ITERATE FOR ARG IN (CDR X) COLLECT (TRANSLATE ARG))))))

(DEFUN PROPERTYLESS-SYMBOLP (X)
  ;; to be modified again in defn-sk.lisp
  (OR (CAR-CDRP X)
      (MEMBER-EQ X (QUOTE (NIL QUOTE LIST LET T F*)))))

(format t "~%Done -- now we may proceed.~%")

(format-when-ready "")

)

   (load (demo-file "adder.ev")
	    :verbose nil)
   (setq *just-did-one* t)
)

