;;; -*-  Mode: Lisp; Package: USER; Syntax: Common-Lisp; Base: 10 -*- ;;;;

;  Copyright (C) 1994 by Matt Kaufmann and Computational Logic, Inc.
;  All Rights Reserved.

;  Copying of this file is authorized to those who have read and
;  agreed with the terms in the "Pc-Nqthm-1992 GENERAL PUBLIC SOFTWARE
;  LICENSE" at the beginning of the Pc-Nqthm file "basis-pc.lisp".

;  Pc-Nqthm Version 1992 (for use with Nqthm-1992)

(IN-PACKAGE "USER")

(EVAL-WHEN (LOAD EVAL COMPILE)
	   (CHK-BASE-AND-PACKAGE-1992 10 *PACKAGE*))

; This file contains the state-changing functions and other
; functions of a "top-level" nature.


; Let's begin with the three functions called by
; INTERACTIVELY-REWRITE to carry out the user's instruction.

(DEFUN RETURN-NEW-STATE (INSTR STATE)
  
; This function is called from INTERACTIVELY-REWRITE to return
; a new state (or NIL).  It always returns a state in which
; the current goal is one which has not already been proved
; (if there is one).  Exception:  If the new state was created
; by a CHANGE-GOAL, COMMENT, or BOOKMARK command, then that
; is the state which is returned.
;    Note that if the current goal is already proved , the
; instruction is ignored (unless it is one of the above instructions).
; If I find other instructions for which it is both useful and
; sound to operate on a proved goal, I may wish to add them as
; additional such exceptions.

  ;; Because of computations such as finding all free variables in a
  ;; given state (as in CHANGE-STATE-WITH-REWRITE, for example), I
  ;; think it would be most efficient to drop all hypotheses when a
  ;; goal has been proved.
  
  (COND ((AND (EQUAL (ACCESS GOAL CONC (ACCESS STATE GOAL STATE))
		     TRUE)
	      (NOT (MEMBER-EQ (INSTR-NAME INSTR)
			      (QUOTE (CHANGE-GOAL COMMENT BOOKMARK)))))
	 (BLURB NIL CR |***NO| |CHANGE***| |--| |The| |current| |goal| |is|
		|already| |proved| EXCL CR))
	(T
	 (LET* ((NEW-STATE
		 (FUNCALL (PACK (LIST (QUOTE CHANGE-STATE-WITH-)
				      (INSTR-NAME INSTR)))
			  INSTR STATE))
		NEW-GOAL)
	       (COND
		(NEW-STATE
		 (SETQ NEW-GOAL (ACCESS STATE GOAL NEW-STATE))
		 (WHEN (AND (ACCESS GOAL HYPS NEW-GOAL)
			    (EQUAL (ACCESS GOAL CONC NEW-GOAL) TRUE))
		       (SETQ NEW-STATE
			     (MAKE-NEW-RECORD
			      (QUOTE STATE)
			      (LIST (CONS (QUOTE GOAL)
					  (MAKE-NEW-RECORD
					   (QUOTE GOAL)
					   (LIST (CONS (QUOTE HYPS) NIL))
					   NEW-GOAL)))
			      NEW-STATE)))
		 (IF (EQ (INSTR-NAME INSTR)
			 (QUOTE CHANGE-GOAL))
		     NEW-STATE
		     (CHANGE-TO-UNPROVED-GOAL NEW-STATE))))))))

; Here is the analogous function for printing helpful
; information.

(DEFUN PRINT-HELP (INSTR)

; This function prints helpful information.  It is evaluated
; only for side-effect (i.e., for this printing).

  (FUNCALL (PACK (LIST (QUOTE PRINT-HELP-WITH-)
		       (INSTR-NAME INSTR)))
	   INSTR))

; The next function auxiliary to the main loop is one
; which changes the STATE-STACK according to the given
; meta-instruction.  New meta instructions might be added
; in expanded versions of this function definition.

(DEFUN CHANGE-STATE-STACK (INSTR)
  (FUNCALL (PACK (LIST (QUOTE CHANGE-STATE-STACK-WITH-)
		       (INSTR-NAME INSTR)))
	   INSTR))

; Now for the main (interactive) function.

(DEFUN INTERACTIVELY-REWRITE (instrs
			      &aux (saved-blurb-flg blurb-flg)
			      (saved-io-fn io-fn))
  (unwind-protect
      (let (name args
		 (arity-alist arity-alist)
		 (start-state (car (our-last state-stack))))
	(cond
	 ((and (match (access state instruction start-state)
		      (start (defn name args & &)))
	       (member-eq name (all-fnnames (access state current-term start-state))))
	  (setq arity-alist
		(if (assoc-eq name arity-alist)
		    arity-alist
		  (CONS (CONS NAME (LENGTH ARGS)) ARITY-ALIST)))))
	(ITERATE WITH INPUT AND OUTPUT AND NAME
		 UNTIL (OR (EQ OUTPUT (QUOTE *EXIT*))
			   (MATCH OUTPUT (*EXIT* NAME)))
		 DO
		 (PROGN
		  (IPRINC *pc-prompt* NIL) ;prompt for next instr
		  (cond
		   (instrs 
		    (setq input (car instrs))
		    (setq instrs (cdr instrs))
		    (iprinc input nil)
		    (iterpri nil))
		   (t
		    (SETQ INPUT (READ NIL))))
		  (cond ((and (MACRO-COMMAND-DEF 'USER)
			      (neq input 'exit)
			      (not (and (consp input) (eq (car input) 'exit))))
			 (setq input (list 'user input))))
		  (BLURB NIL CR NIL)
		  (SETQ OUTPUT
			(LET ((PRINT-INSTR-FLG
			       (AND (NOT (INTEGERP INPUT))
				    (MACRO-COMMAND-DEF (INSTR-NAME INPUT)))))
					; Notice the wrap of USER around the instruction.  USER
					; could reset PRINT-INSTR-FLG to the appropriate value
					; if that's desired by the sophisticated user; otherwise,
					; it's on.  It makes sense to have it on, since USER could
					; change the instruction.
			     (INTERACTIVE-SINGLE-STEP INPUT))))
		 FINALLY (RETURN NAME)))
    (when (not (eq blurb-flg saved-blurb-flg))
	  (setq blurb-flg saved-blurb-flg)
	  (format t "~&Proof-checker printing status reset to ~S.~&" blurb-flg))
    (when (not (eq io-fn saved-io-fn))
	  (setq io-fn saved-io-fn)
	  (format t "~&Theorem prover printing status restored to function ~S.~&" io-fn))))

(DEFUN INTERACTIVE-SINGLE-STEP (INPUT)
  ;; After macro-expanding INPUT until the input is no longer a macro command,
  ;; execute the instruction.  The value returned is NIL for any instruction
  ;; other than a "successful" change or meta or EXIT instruction.
  ;; Macro-expansion errors are reported upon attempted expansion.
  ;; Note:  only the EXIT instruction should return a value of the form
  ;; (*EXIT* name).
  ;;    NOTE:  This function is not called by PROVE-LEMMA because
  ;; it calls READ-INSTR, which uses Lisp to
  ;; evaluate arguments prefixed by an exclamation point (!).
  (LET (INSTR-AND-TYPE NEW-STATE INSTR INSTR-TYPE
		       EVENT-NAME LEMMA-TYPES** BAD-DEPENDENTS
		       (ERROR1-SET-FLG NO-DEBUG-FLG)
		       exit-do-it-flag)
       (CATCH (QUOTE ERROR1-SET)
	      (COND
	       ((SETQ INPUT (READ-INSTR INPUT))
		(COND
		 ((SETQ INSTR-AND-TYPE (MAKE-INSTR-AND-TYPE INPUT))
		  (SETQ INSTR (CAR INSTR-AND-TYPE))
		  (SETQ INSTR-TYPE (CDR INSTR-AND-TYPE))
		  (COND			;carry out instruction
		   ((EQ INSTR-TYPE (QUOTE C)) ;state-change instruction
		    (COND (PRINT-INSTR-FLG
			   (BLURB (INSTR) | | (!PPR INSTR NIL) CR)))
		    (COND
		     ((SETQ NEW-STATE (RETURN-NEW-STATE INSTR (CAR STATE-STACK)))
		      (SETQ STATE-STACK (CONS NEW-STATE STATE-STACK)))))
		   ((EQ INSTR-TYPE (QUOTE H)) ;help instruction
		    (PRINT-HELP INSTR)
		    NIL)
		   ((EQ INSTR-TYPE (QUOTE M)) ;meta-instruction
		    (LET ((PRINT-INSTR-FLG T))
			 (CHANGE-STATE-STACK INSTR)))
		   ((EQ INSTR (QUOTE EXIT))
		    (BLURB NIL CR |Quitting| |the| |interactive| |proof| |checker| |.|
			   |Submit| |(VERIFY)| |to| |get| |back| |in| |at| |this|
			   |state| |.| |**NOTE**| |--| |No| |event| |has| |been| |stored| |.| CR)
		    (QUOTE *EXIT*))
		   ((SETQ BAD-DEPENDENTS
			  (NCONC
			   (ITERATE FOR X IN (ACCESS STATE CUMULATIVE-LEMMAS-USED (CAR STATE-STACK))
				    WHEN (NOT (OR (GET X 'EVENT)
						  (GET X 'MAIN-EVENT)))
				    COLLECT X)
			   (ITERATE FOR X IN (ALL-FNNAMES
					      (ACCESS GOAL CONC
						      (ACCESS STATE GOAL
							      (CAR (OUR-LAST STATE-STACK)))))
				    WHEN (NOT (OR (GET X 'EVENT)
						  (GET X 'MAIN-EVENT)))
				    COLLECT X)))
		    (ER SOFT (BAD-DEPENDENTS INPUT) |The| |following|
			(PLURAL? BAD-DEPENDENTS |"events" are| |"event" is|)
			|not| |currently| |in| |the| |data-base| |:| (!LIST BAD-DEPENDENTS) |.|
			|Apparently| |you|
			|did| |a| UBT |or| |an| |UNDO-NAME| |in| |the| |middle| |of| |this|
			|interactive| |session| |.| |You| |might| |wish| |to| EXIT |now|
			|rather| |than| (!PPR INPUT (QUOTE |.|))))
		   ((or (MATCH INSTR (EXIT 'T))
			(MATCH INSTR (EXIT 'T exit-do-it-flag)))
		    (COND
		     ((NOT (MATCH (ACCESS STATE INSTRUCTION (CAR (OUR-LAST STATE-STACK)))
				  (START (CONS 'DEFN &))))
		      (BLURB (exit-do-it-flag) |**NO| |CHANGE**| |--| |The| |command| |(EXIT T)|
			     (COND (EXIT-DO-IT-FLAG |(which| |in| |this| |case|
						    |has| |an| |extra| |argument|
						    (!ppr exit-do-it-flag nil)
						    |telling| |the| |system| |not| |to| |prompt|
						    |you| |upon| |exit)|))
			     |may| |only|
			     |be| |used| |to| |exit| |from| |an| |interactive| |verification| |for|
			     |a| DEFN |event| |,| |not| |for| |a| PROVE-LEMMA |.| CR))
		     ((SETQ EVENT-NAME (FINISH-INTERACTIVELY-REWRITE-FOR-DEFN exit-do-it-flag))
		      (LIST (QUOTE *EXIT*) EVENT-NAME))
		     (T (QUOTE *EXIT*))))
		   ((MATCH INSTR (EXIT &))
		    (BLURB NIL |**NO| |CHANGE**| |--| |The| |one-argument| |version| |of|
			   EXIT |,| |which| |indicates| |exiting| |from| |an| |interactive|
			   |verification| |for| |a| DEFN |event| |,| |must| |be| |(EXIT T)| |.| CR))
		   ((or (MATCH INSTR (EXIT EVENT-NAME LEMMA-TYPES**))
			(MATCH INSTR (EXIT EVENT-NAME LEMMA-TYPES** exit-do-it-flag)))
		    (COND 
		     ((MATCH (ACCESS STATE INSTRUCTION (CAR (OUR-LAST STATE-STACK)))
			     (START (CONS 'DEFN &)))
		      (COND
		       ((ASSOC-EQ (CADR (CADR (ACCESS STATE INSTRUCTION (CAR (OUR-LAST STATE-STACK)))))
				  ARITY-ALIST)
			(ER SOFT ((X (CADR (CADR (ACCESS STATE INSTRUCTION (CAR (OUR-LAST STATE-STACK)))))))
			    |This| |session| |began| |as| |a| |proof| |of| |termination| |of| |the|
			    |definition| |of| (!PPR X NIL) |,| |and| |the| |goal| |that| |was| |generated|
			    |for| |that| |purpose| |mentions| (!PPR X (QUOTE |.|)) |It| |is| |therefore|
			    |not| |allowed| |to| |make| |a| PROVE-LEMMA |event| |out| |of| |that| |goal| |.|))
		       (T
			(BLURB ((X (CADR (CADR (ACCESS STATE INSTRUCTION (CAR (OUR-LAST STATE-STACK)))))))
			    CR |This| |session| |began| |as| |a| |proof| |of| |termination| |for| |the|
			    |function| (!PPR X NIL) |,| |but| |you| |are| |making| |a| |lemma| |out| |of|
			    |the| |goal| |which| |guarantees| |termination| |.| |This| |is| |allowed| |,|
			    |but| |is| |this| |really| |what| |you| |want| |to| |do| |?| |If| |so| |,|
			    |confirm| |by| |answering| YES |;| |otherwise| |,| |answer| |NO:  |)
			(COND ((EQ (READ NIL) 'YES)
			       (BLURB NIL CR |**NOTE**| |--| |Now| |transforming| |this| |interactive|
				      |session| |into| |one| |for| |PROVE-LEMMA| |rather| |than| |for|
				      DEFN |.| CR)
			       (SETF (CAR (OUR-LAST STATE-STACK))
				     (MAKE-NEW-RECORD
				      (QUOTE STATE)
				      (LIST (CONS (QUOTE INSTRUCTION) (QUOTE START)))
				      (CAR (OUR-LAST STATE-STACK)))))
			      (T (ER SOFT NIL |Changed| |your| |mind| |,| |I| |guess| |.|
				     |Back| |into| |interactive| |verification| |of|
				     |termination| |.|)))))))
		    (COND
		     ((FINISH-INTERACTIVELY-REWRITE EVENT-NAME LEMMA-TYPES** exit-do-it-flag)
		      (LIST (QUOTE *EXIT*) EVENT-NAME))
		     (T (QUOTE *EXIT*))))
		   (T (ER HARD NIL |Unexpected| |situation| |in|
			  INTERACTIVE-SINGLE-STEP |.|))))))))))

(DEFUN FINISH-INTERACTIVELY-REWRITE (EVENT-NAME LEMMA-TYPES** exit-do-it-flag)

; Remember that the event should specify a START-GOAL when no PREVIOUS-EVENT
; is given and a GOAL-NAME is given.

; This function returns a non-NIL value if and only if the MAIN goal has been
; proved in the current STATE-STACK context.

  (LET (INSTRS PREVIOUS-EVENT STATEMENT ans
	(ALL-GOALS (ALL-GOALS (CAR STATE-STACK))))
    ; Collect instructions, in the correct order, noting PREVIOUS-EVENT
    ; if appropriate.  Only the instructions since the PREVIOUS-EVENT
    ; (if there is such) should be collected.
    (ITERATE FOR S IN STATE-STACK
	  UNTIL
	    (OR
	      (MATCH (ACCESS STATE INSTRUCTION S) (START PREVIOUS-EVENT))
	      (EQ (ACCESS STATE INSTRUCTION S) (QUOTE START)))
	  DO (SETQ INSTRS
		   (CONS (ACCESS STATE INSTRUCTION S) INSTRS)))
    (ITERPRI NIL)
    (COND
      ((ITERATE FOR GL IN ALL-GOALS
		ALWAYS (EQUAL (ACCESS GOAL CONC GL) TRUE))
       (SETQ STATEMENT
	     (UNTRANSLATE (ACCESS GOAL CONC
				  (ACCESS STATE GOAL (CAR (OUR-LAST STATE-STACK))))))
       (IPRINC "The indicated goal has been proved.  Here is the desired event:" NIL)
       (ITERPRI NIL)
       (PPR
	 (SETQ EV0
	       `(PROVE-LEMMA
		  ,EVENT-NAME
		  ,LEMMA-TYPES**
		  ,STATEMENT
		  ,(APPEND
		     (COND
		       (PREVIOUS-EVENT
			`((PREVIOUS-EVENT ,PREVIOUS-EVENT)))
		       (T NIL))
		     `((INSTRUCTIONS ,@INSTRS)))))
	 NIL))
      (PREVIOUS-EVENT
       (IPRINC "WARNING:  The appropriate goal has not been proved." NIL)
       (ITERPRI NIL)
       (IPRINC "The following event notes progress made since " NIL)
       (IPRINC PREVIOUS-EVENT NIL)
       (ITERPRI NIL)
       (PPR
	 (SETQ EV0
	       `(PROVE-LEMMA
		  ,EVENT-NAME
		  NIL
		  T
		  ((PREVIOUS-EVENT ,PREVIOUS-EVENT)
		   (INSTRUCTIONS ,@INSTRS))))
	 NIL))
      (T
       (IPRINC "WARNING:  The appropriate goal has not been proved." NIL)
       (ITERPRI NIL)
       (IPRINC "The following event notes progress made during this (fresh) session." NIL)
       (ITERPRI NIL)
       (PPR
	 (SETQ EV0
	       `(PROVE-LEMMA
		  ,EVENT-NAME
		  NIL
		  T
		  ((START-GOAL
		    ,(UNTRANSLATE
		      (ACCESS GOAL
			      CONC
			      (ACCESS STATE GOAL (CAR (OUR-LAST STATE-STACK))))))
		   (INSTRUCTIONS ,@INSTRS))))
	 NIL)))
    (iterpri nil)
    (when (null exit-do-it-flag)
	  (BLURB NIL |Do| |you| |want| |to| |submit| |this| |event?| CR
		 Y |(Yes)| |,| R |(Yes| |and| |replay| |commands)| |,| |or|
		 N |(No)| |? |)
	  (setq ANS (READ NIL))
	  (ITERATE WHILE (NOT (MEMBER-EQ ANS
					 '(Y N R YES NO REPLAY)))
		   as i from 1
		   DO (cond
		       ((> I 5)
			(IPRINC "That's it -- I'm quitting!!" NIL)
			(setq ans 'n))
		       ((> i 2)
			(iprinc "If you answer wrong again I may crash the machine: "
				NIL)
			(setq ans (read nil)))
		       (t
			(IPRINC "Please answer Y, N, or R:  " NIL)
			(SETQ ANS (READ NIL))))))
    (COND
     ((OR exit-do-it-flag
	  (EQUAL ANS 'Y)
	  (EQUAL ANS 'YES))
      (LET (REPLAY-FLG) (EVAL EV0)))
     ((OR (EQUAL ANS 'R)
	  (EQUAL ANS 'REPLAY))
      (ITERPRI NIL)
      (EVAL EV0))
     (T
      NIL))))

(DEFUN FINISH-INTERACTIVELY-REWRITE-FOR-DEFN (exit-do-it-flag)

; This function returns a non-NIL value if and only if the MAIN goal has been
; proved in the current STATE-STACK context.

  (LET ((INSTRS (ITERATE FOR S IN (CDR (REVERSE STATE-STACK))
			 COLLECT (ACCESS STATE INSTRUCTION S)))
	ans
	(ALL-GOALS (ALL-GOALS (CAR STATE-STACK))))
    (ITERPRI NIL)
    (COND
      ((ITERATE FOR GL IN ALL-GOALS
		ALWAYS (EQUAL (ACCESS GOAL CONC GL) TRUE))
       (IPRINC "The indicated goal has been proved.  Here is the desired event:" NIL)
       (ITERPRI NIL)
       (MATCH (ACCESS STATE INSTRUCTION (CAR (OUR-LAST STATE-STACK)))
	      (START EV0)) 
       (SETQ EV0 (APPEND EV0 (LIST (CONS 'INSTRUCTIONS INSTRS))))
       (PPR EV0 NIL)
       (iterpri nil)
       (when (null exit-do-it-flag)
	     (BLURB NIL |Do| |you| |want| |to| |submit| |this| |event?| CR
		    Y |(Yes)| |,| R |(Yes| |and| |replay| |commands)| |,| |or|
		    N |(No)| |? |)
	     (setq ANS (READ NIL))
	     (ITERATE WHILE (NOT (MEMBER-EQ ANS
					    '(Y N R YES NO REPLAY)))
		      as i from 1
		      DO (cond
			  ((> I 5)
			   (IPRINC "That's it -- I'm quitting!!" NIL)
			   (setq ans 'n))
			  ((> i 2)
			   (iprinc "If you answer wrong again I may crash the machine: "
				   NIL)
			   (setq ans (read nil)))
			  (t
			   (IPRINC "Please answer Y, N, or R:  " NIL)
			   (SETQ ANS (READ NIL))))))
       (COND
	((OR exit-do-it-flag
	     (EQUAL ANS 'Y)
	     (EQUAL ANS 'YES))
	 (LET (REPLAY-FLG) (EVAL EV0)))
	((OR (EQUAL ANS 'R)
	     (EQUAL ANS 'REPLAY))
	 (ITERPRI NIL)
	 (EVAL EV0))
	(T
	 NIL)))
      (T
       (IPRINC "WARNING:  The appropriate goal has not been proved." NIL)
       (ITERPRI NIL)
       (IPRINC "However, here are the instructions created so far:" NIL)
       (ITERPRI NIL)
       (ppr instrs nil)
       nil))
    ))

(DEFUN NONINTERACTIVELY-REWRITE-FOR-PROVE-LEMMA (INSTRS)
; Runs the given instructions in the context of the current
; state stack, printing each instruction as it is tried.  Note
; that only change instructions are run here, and that we
; don't evaluate arguments of instructions.
  (LET (INPUT INSTR-AND-TYPE NEW-STATE INSTR INSTR-TYPE)
    (SETQ UNDONE-INSTRS INSTRS)
    (COND
      ((NULL INSTRS)
       NIL)
      (T
       (SETQ INPUT (READ-INSTR-for-prove-lemma (CAR INSTRS)))	;fixed here for printing
       (BLURB (INPUT) | | (!PPR INPUT NIL) CR)
       (SETQ NEW-STATE NIL)
       (COND
	((SETQ INSTR-AND-TYPE (INSTR-AND-TYPE INPUT))
	 (SETQ INSTR (CAR INSTR-AND-TYPE))
	 (SETQ INSTR-TYPE (CDR INSTR-AND-TYPE))
	 (COND				;carry out instruction
	  ((EQ INSTR-TYPE (QUOTE C))	;state-change instruction
	   (COND
	    ((SETQ NEW-STATE (RETURN-NEW-STATE INSTR (CAR STATE-STACK)))
	     (SETQ STATE-STACK (CONS NEW-STATE STATE-STACK)))))
	  (t (blurb nil CR |In| |this| |context| |,| |instructions| |must| |be|
		    |legal| |change| |instructions| |,| |but| (!ppr instr nil) |is| |not|
		    |.| CR)))))
       (COND
	 (NEW-STATE
	  (SETQ UNDONE-INSTRS (CDR UNDONE-INSTRS))
	  (NONINTERACTIVELY-REWRITE-FOR-PROVE-LEMMA (CDR INSTRS)))
	 (T
	  (BLURB (INPUT) CR |Failed| |to| |create| |a| |new| |state| |with|
		 |the| |command| (!PPR INPUT (QUOTE |.|)) |Invoke|
		 |(UNDONE-INSTRUCTIONS)| |to| |see| |the| |failed| |commands| |.| CR)))))))

(defun replay-for-defn (term instrs)
  (let ((chronology (cdr chronology)))
    (INITIALIZE-FREWRITE))		;since the new event is not one which has been dealt with,
					;and we've already done the MAKE-EVENT
  (SETQ STATE-STACK
	(LIST
	 (MAKE STATE 
	       START-MARKER		;set in DEFN
	       TERM
	       NIL NIL
	       (MAKE GOAL
		     TERM
		     NIL
		     NIL (QUOTE MAIN))
	       NIL NIL (INITIAL-FREWRITE-DISABLED-RULES) NIL NIL)))
  (BLURB NIL CR |***| |Entering| |proof-checker| |***| CR)
  (NONINTERACTIVELY-REWRITE-FOR-PROVE-LEMMA instrs))

(DEFUN REPLAY-FOR-PROVE-LEMMA (TERM HINTS)
  (INITIALIZE-FREWRITE)
  (COND
    ((ASSOC-EQ (QUOTE PREVIOUS-EVENT) HINTS)
     ;Start in the middle
     (SETQ STATE-STACK
	   (CDR
	     (ASSOC-EQ (CADR (ASSOC-EQ (QUOTE PREVIOUS-EVENT) HINTS))
		   STATE-STACK-LIST))))
    (T
     ;Start at the beginning of the session, with the START-GOAL
     ;as the main goal, if such is given in the HINTS
     (COND
       ((ASSOC-EQ (QUOTE START-GOAL) HINTS)
	(SETQ TERM (CADR (ASSOC-EQ (QUOTE START-GOAL) HINTS)))))
     (SETQ STATE-STACK
	   (LIST
	     (MAKE STATE (QUOTE START) TERM
		   NIL NIL
		   (MAKE GOAL
			 TERM
			 NIL
			 NIL (QUOTE MAIN))
		   NIL NIL (INITIAL-FREWRITE-DISABLED-RULES) NIL NIL)))))
  (BLURB NIL CR |***| |Entering| |proof-checker| |***| CR)
  (NONINTERACTIVELY-REWRITE-FOR-PROVE-LEMMA
   (CDR (ASSOC-EQ (QUOTE INSTRUCTIONS) HINTS))))

; Before getting to the functions that are FUNCALLed by RETURN-NEW-STATE
; and PRINT-HELP, let's define the top-level functions called by
; VERIFY and RE-ENTER.

(defun protect-interactively-rewrite
  (instrs &optional (stack-depth restart-stack-depth))
  (my-unwind-protect
   (interactively-rewrite instrs)
   (restart-proof-checker-from-error stack-depth)))

(DEFUN START-PROOF-CHECKER (X instrs)
  ;; This function is NOT called by PROVE-LEMMA, so it is safe to
  ;; put an UNWIND-PROTECT around it.
  (LET ((X X))
       (INITIALIZE-FREWRITE)
       (SETQ OLD-SS NIL)
       (SETQ STATE-STACK
	     (LIST
	      (MAKE STATE (QUOTE START) X
		    NIL NIL
		    (MAKE GOAL X NIL NIL (QUOTE MAIN))
		    NIL NIL (INITIAL-FREWRITE-DISABLED-RULES) NIL NIL)))
       (ITERPRI NIL)
       (protect-INTERACTIVELY-REWRITE instrs)))

(DEFUN START-PROOF-CHECKER-DEFN (X START-MARKER instrs)
  ;; This function is NOT called by PROVE-LEMMA or DEFN, so it is safe to
  ;; put an UNWIND-PROTECT around it.
  (LET ((X X) (START-MARKER START-MARKER))
       (COND
	((EQUAL X TRUE)
	 (BLURB NIL CR |Did| |you| |really| |want| |to| |do| |this| |interactively?| |There|
		|is| |nothing| |to| |prove| EXCL |If| |you| |really| |want| |to| |exit|
		|here| |,| |put| |in| |a| COMMENT |first| |.| CR)))
       (INITIALIZE-FREWRITE)
       (SETQ OLD-SS NIL)
       (SETQ STATE-STACK
	     (LIST
	      (MAKE STATE
		    START-MARKER
		    X
		    NIL NIL
		    (MAKE GOAL X NIL NIL (QUOTE MAIN))
		    NIL NIL (INITIAL-FREWRITE-DISABLED-RULES) NIL NIL)))
       (ITERPRI NIL)
       (protect-INTERACTIVELY-REWRITE instrs)))

(DEFUN RESTART-PROOF-CHECKER (PREVIOUS-EVENT instrs)
  ;; the top-level function for re-entering the proof-checker from
  ;; an interactive proof which did not complete, except that when
  ;; PREVIOUS-EVENT is NIL then its effect is simply to put one
  ;; back in at the current STATE-STACK after re-initialization.
  (LET ((PREVIOUS-EVENT PREVIOUS-EVENT))
       (COND
	((ERROR1-SET
	  (AND PREVIOUS-EVENT		;error situations
	       (COND
		((NOT (MEMBER-EQ PREVIOUS-EVENT CHRONOLOGY))
		 (ER SOFT (PREVIOUS-EVENT) |the| |"event"| (!PPR PREVIOUS-EVENT NIL) |does|
		     |not| |belong| |to| |the| |chronology| |.|))
		((NULL (CDR (ASSOC-EQ PREVIOUS-EVENT STATE-STACK-LIST)))
		 (COND
		  ((AND (MATCH (GET PREVIOUS-EVENT 'EVENT)
			       (PROVE-LEMMA & & & &))
			(ASSOC-EQ 'INSTRUCTIONS (NTH 4 (GET PREVIOUS-EVENT 'EVENT))))
		   (COND
		    ((EQUAL (NTH 3 (GET PREVIOUS-EVENT 'EVENT)) TRUE)
		     (COND
		      ((NULL (ASSOC-EQ PREVIOUS-EVENT STATE-STACK-LIST))
		       (ER HARD (PREVIOUS-EVENT) (!PPR PREVIOUS-EVENT NIL) |is| |missing|
			   |from| STATE-STACK-LIST EXCL)))
		     (ER SOFT (PREVIOUS-EVENT) |Although| |the| |event| (!PPR PREVIOUS-EVENT NIL)
			 |is| |an| |interactive| |PROVE-LEMMA| |event| |with| |a| |partial| |proof| |,|
			 |,| |its| |partial| |proof|
			 |is| |not| |around| |.| |That| |is| |because| |this| |session| |began| |by|
			 |NOTE-LIBing| |a| |library| |which| |someone| |created| |with| |the|
			 |request| |not| |to| |save| |this| |partial| |proof| |.|))
		    (T
		     (ER SOFT (PREVIOUS-EVENT) |Although| |the| |event| (!PPR PREVIOUS-EVENT NIL)
			 |is| |an| |interactive| |PROVE-LEMMA| |event| |,| |its| |proof| |is|
			 |complete| |and| |hence| |it| |is| |not| |allowed| |to| |re-enter| |its|
			 |proof| |.| |If| |its| |proof| |was| |the| |most| |recent| |one| |worked|
			 |on| |,| |then| |the| |command| |(VERIFY)| |should| |work| |here| |.|))))
		  (T
		   (ER SOFT (PREVIOUS-EVENT) |the| |event| (!PPR PREVIOUS-EVENT NIL) |is|
		       |not| |an| |interactive| PROVE-LEMMA |event| |.|)))))))
	 (INITIALIZE-FREWRITE)
	 (COND
	  (PREVIOUS-EVENT
	   (SETQ OLD-SS NIL)
	   (SETQ STATE-STACK (CDR (ASSOC-EQ PREVIOUS-EVENT STATE-STACK-LIST)))))
	 (ITERPRI NIL)
	 (COMPARE-LOCAL-AND-GLOBAL-DISABLED-RULES)
	 (protect-INTERACTIVELY-REWRITE instrs)))))

(DEFUN RESTART-PROOF-CHECKER-FROM-ERROR (STACK-DEPTH) 
  (LET ((STACK-DEPTH STACK-DEPTH))
       #+kcl (si::reset-stack-limits)
       (COND
	((AND STACK-DEPTH (INT= STACK-DEPTH 0))
	 (COND ((INT= RESTART-STACK-DEPTH 0)
		(BLURB NIL CR |Now| |leaving| |proof| |checker| |.| |Submit| |(VERIFY)|
		       |to| |get| |back| |in| |where| |you| |left| |off| |.| CR))
	       (T
		(BLURB NIL CR |Stack| |depth| |exceeded| |for| |re-entering| |proof-checker| |.|
		       |But| |you| |can| |get| |right| |back| |in| |by| |submitting|
		       |(VERIFY)| |.| CR))))
	(T
	 (BLURB NIL CR |Now| |going| |back| |in| |to| |the|
		|proof-checker| |.| |Type| EXIT |to| |exit| |.| CR)
	 (protect-INTERACTIVELY-REWRITE nil (AND STACK-DEPTH (1- STACK-DEPTH)))))))

;;************************************************************;;
;;                 STATE CHANGING FUNCTIONS                   ;;
;;************************************************************;;

;;; The paradigm here is that most CHANGE-STATE-WITH- functions
;;; are associated with a "-CHK" function (defined just above them).
;;; The point is simply to increase readability of the code by
;;; isolating as much error-checking as is convenient.  The
;;; "-CHK" functions always return T or NIL.

(DEFUN CHANGE-STATE-WITH-ADD-ABBREVIATION-CHK (VAR ABBREVIATIONS)
  (COND
   ((NOT (AND (SYMBOLP VAR)
	      (EQUAL (CHAR (STRING VAR) 0) #\@)))
    (BLURB (VAR) |**NO| |CHANGE**| |--| |The| |first| |argument| |to|
	   ADD-ABBREVIATION |must| |be| |a| |symbol| |starting| |with|
	   |the| |character| |@| |,| |but| (!PPR VAR NIL) |is| |not| |.| CR))
   ((ITERATE FOR X IN ABBREVIATIONS
	     THEREIS
	     (AND
	      (EQ (CAR X) VAR)
	      (PROGN
	       (BLURB (VAR (TERM (CDR X)))
		      |**NO| |CHANGE**| |--| |The| |variable|
		      (!PPR VAR NIL) |is| |already| |an| |abbreviation|
		      |for| (!TERM TERM (QUOTE |.|)) CR)
	       T)))
    NIL)
   (T T)))

(DEFUN CHANGE-STATE-WITH-ADD-ABBREVIATION (INSTR STATE)
  (LET-FIELDS
    (LET ((VAR (cadr instr)) TERM)
      (COND
	((CHANGE-STATE-WITH-ADD-ABBREVIATION-CHK VAR ABBREVIATIONS)
	 (or (and (MATCH INSTR (ADD-ABBREVIATION & term))
		  (SETQ TERM (TRANSLATE-AND-CHK-ABB TERM ABBREVIATIONS)))
	     (setq term current-term))
	 (ITERATE FOR PAIR IN ABBREVIATIONS
	       WHEN
		 (EQUAL TERM (CDR PAIR))
		 DO (BLURB ((VAR-ABB (CAR PAIR))
			    (OLDTERM (CADDR INSTR)))
			   WARNING |:| |the| |term| |to| |be| |abbreviated| |,| |namely|
			   (cond (oldterm (!ppr OLDTERM (QUOTE |,|)))
				 (t |the| |current| |term| |,|))
			   |already| |has| |the| |abbreviation|
			   (!PPR VAR-ABB NIL) |,| |which| |will| |be|
			   |overridden| |during| |printing| |by| |the| |new| |abbreviation| |.|
			   |Submit| |the| |command| UNDO |if| |that|
			   |is| |not| |what| |you| |intended| |.| CR))
	 (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		       (UNION-EQ
			(ALL-FNNAMES TERM)
			CUMULATIVE-LEMMAS-USED))
		 (CONS (QUOTE ABBREVIATIONS)
		       (CONS (CONS VAR TERM)
			     ABBREVIATIONS)))
	   STATE))))))

(DEFUN CHANGE-STATE-WITH-bash-CHK (INSTR CURRENT-ADDR-R HYPS CONC ABBREVIATIONS
					 raw-hints full-prover-flg)
  (let (hints
	(goal-formula (MAKE-IMPLICATION (ASSUMPTIONS HYPS) CONC)))
    (SETQ ALL-LEMMAS-USED NIL)
    (COND
     (CURRENT-ADDR-R
      (BLURB NIL |**NO| |CHANGE**| |--| |The| BASH |command| |should| |only|
	     |be| |used| |at| |the| |top| |.| |use| |(= T)| |if| |that|
	     |is| |what| |you| |want| |,| |or| |perhaps| |use| PUSH
	     |first| |.| CR))
     ((and (consp instr)
	   (null (error1-set (match! (chk-acceptable-hints-abb raw-hints abbreviations)
				     (list hints)))))
      nil)
     (t
      (cond
       ;; Do not use LET here to share the COND terms below, because APPLY-HINTS
       ;; needs to be inside PROVE1 or BASH1 here for the appropriate UNWIND-PROTECT.
       (full-prover-flg
	(prove1 (cond ((atom instr) goal-formula)
		      (t (apply-hints hints goal-formula)))
		t))
       (t
	(bash1 (cond ((atom instr) goal-formula)
		     (t (apply-hints hints goal-formula))))
	(setq bash-clauses process-clauses)))
      (cond
       ((iterate for cl in bash-clauses
		 thereis (null cl))
	(blurb nil |!!!! WARNING !!!!| CR |#| |A| |subgoal| |has| |simplified|
	       |to| F |,| |which| |suggests| |that| |the| |goal| |probably|
	       |is| |not| |provable| |.| CR |**NO| |CHANGE**| CR))
       ((or (assoc-eq 'induct hints)
	    (assoc-eq 'use hints)
	    (not (INT= (length bash-clauses) 1))
	    (not (equal goal-formula
			(flatten-antecedents
			 (make-formula-from-clause (car bash-clauses))))))
	(COND (HINTS
	       (SETQ ALL-LEMMAS-USED
		     (UNION-EQ (EXTRACT-DEPENDENCIES-FROM-HINTS
				HINTS)
			       ALL-LEMMAS-USED))))
	t)
       (t (BLURB NIL CR |**NO| |CHANGE**| |--| |No| |simplification| |took|
		 |place| |.| CR)))))))

(DEFUN CHANGE-STATE-WITH-bash (INSTR STATE)
  (LET-FIELDS
   (LET*
    ((hints-return (split-return-clauses-from-hints (and (consp instr) (cdr instr))))
     (raw-hints (cdr hints-return))
     (full-prover-flg (car hints-return))
     (otf-flg (eq full-prover-flg t))
     NEW-GOAL-NAMES NEW-FORMULAS bad-free-variables)
    (cond
      ((CHANGE-STATE-WITH-bash-CHK INSTR CURRENT-ADDR-R HYPS CONC ABBREVIATIONS
				   raw-hints full-prover-flg)
       (when (null bash-clauses)
	     (blurb nil cr |The| |current| |goal| |has| |been| |proved| |,|
		    |without| |the| |creation| |of| |new| |subgoals| |.|))
       (if all-lemmas-used
	   (blurb (all-lemmas-used) cr |The| |goal| |has| |been| |simplified| |using| 
		  (!ppr-list all-lemmas-used (quote |.|)) cr)
	 (blurb nil cr |Note| |that| |simplification| |took| |place| |without| |particular| |use|
		|of| |rewrite| |rules| |or| |definitions| |.| cr))
       (SETQ NEW-FORMULAS
	     (iterate for cl in bash-clauses
		      collect (flatten-antecedents
			       (make-formula-from-clause cl)))) 
       (SETQ NEW-GOAL-NAMES
	     (NEW-GOAL-NAMES GOAL-NAME OTHER-GOALS
			     (LENGTH NEW-FORMULAS)))
       (when (setq bad-free-variables
		   (intersection-eq free-variables
				    (all-vars-lst new-formulas)))
	     (blurb (bad-free-variables) |#| WARNING |:| |The|
		    (PLURAL? bad-free-variables |variables| |variable|)
		    (!PPR-LIST bad-free-variables NIL) (PLURAL? bad-free-variables |are| |is|)
		    |being| |removed| |from| |the| |list| |of| |free| |variables| |of|
		    |the| |state| |,| |for| |soundness| |reasons| |.|
		    |The| |problem| |is| |that| |we| |cannot| |guarantee| |that|
		    |the| |conjunction| |of| |the| |new| |goals| |logically| |implies| |the| 
		    |given| |goal| |,| |but| |only| |that| |their| |theoremhood| |implies|
		    |its| |theoremhood| |.| |However| |,| |this| |is| |not| |likely| |to|
		    |affect| |your| |situation| |noticeably| |,| |except| |for| |the| |printing|
		    |of| |this| |message| excl CR))
       (MAKE-NEW-RECORD
	(QUOTE STATE)
	(LIST
	 (CONS (QUOTE INSTRUCTION) INSTR)
	 (CONS (QUOTE CURRENT-TERM) TRUE)
	 (CONS (QUOTE GOAL)
	       (MAKE-NEW-RECORD
		(QUOTE GOAL)
		(LIST (CONS (QUOTE CONC) TRUE)
		      (CONS (QUOTE DEPENDS-ON)
			    (APPEND NEW-GOAL-NAMES DEPENDS-ON)))
		GOAL))
	 (CONS (QUOTE OTHER-GOALS)
	       (if new-formulas
		   (UPDATE-OTHER-GOALS
		    OTHER-GOALS
		    GOAL-NAME
		    (ITERATE FOR X IN NEW-FORMULAS
			     COLLECT (CONS X NIL))
		    NEW-GOAL-NAMES)
		   other-goals))
	 ;; the following isn't documented in the March 1990 version
	 (cons (quote free-variables)
	       (if bad-free-variables
		   (set-diff-eq free-variables
				bad-free-variables)
		 free-variables))
	 (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
	       (UNION-EQ all-lemmas-used
			 CUMULATIVE-LEMMAS-USED)))
	STATE))))))

(DEFUN CHANGE-STATE-WITH-BK-CHK (CURRENT-ADDR-R)
  (COND
    ((NULL CURRENT-ADDR-R)
     (BLURB NIL |**NO| |CHANGE**| |--| |Already| |at| |the| |top| |of|
	    |the| |term| |.| CR))
    ((INT= (CAR CURRENT-ADDR-R) 1)
     (BLURB NIL |**NO| |CHANGE**| |--| |Can't| |go| |backward| |from| |first|
	    |argument| |.| CR))
    (T T)))

(DEFUN CHANGE-STATE-WITH-BK (INSTR STATE)
  (LET-FIELDS
    (LET (TEMP NEW-ADDR-R)
      (COND
	((CHANGE-STATE-WITH-BK-CHK CURRENT-ADDR-R)
	 (SETQ NEW-ADDR-R
	       (CONS (1- (CAR CURRENT-ADDR-R))
		     (CDR CURRENT-ADDR-R)))
	 (SETQ TEMP
	       (FETCH-TERM-AND-CL CONC
				  (REVERSE NEW-ADDR-R)
				  NIL
				  (ITERATE FOR PAIR IN GOVERNORS
					COLLECT (CDR PAIR))))
	 (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE CURRENT-TERM) (CAR TEMP))
		 (CONS (QUOTE GOVERNORS) (CDR TEMP))
		 (CONS (QUOTE CURRENT-ADDR-R) NEW-ADDR-R))
	   STATE))))))

(DEFUN CHANGE-STATE-WITH-BOOKMARK (INSTR STATE)
  (COND
    ((INTEGERP (CADR INSTR))
     (BLURB NIL |**NO| |CHANGE**| |--| |Numbers| |may| |not| |be|
	    |used| |as| |bookmarks| |.| CR))
    (T
     (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR))
	   STATE))))

(DEFUN CHANGE-STATE-WITH-CHANGE-GOAL-CHK (INSTR GOAL-NAME OTHER-GOALS)
; guarantees that the CHANGE-GOAL instruction will fire by ensuring that
; a relevant other goal exists
  (LET (FOUND-GOAL)
    (COND
      ((NULL OTHER-GOALS)
       (BLURB NIL |**NO| |CHANGE**| |--| |there| |are| |no| |other| |goals| |.| CR))
      ((CONSP INSTR)
       (COND
	 ((SETQ FOUND-GOAL
		(FIND-GOAL (CADR INSTR)
			   (ITERATE FOR GL-INFO IN OTHER-GOALS
				 COLLECT (CAR GL-INFO))))
	  ;print warning message if new goal is already proved
	  (COND ((EQUAL (ACCESS GOAL CONC FOUND-GOAL) TRUE)
		 (BLURB ((NAME (CADR INSTR))) WARNING |--| |the| |given| |goal| |,|
			(!PPR NAME (QUOTE |,|)) |has| |already| |been| |proved| |,|
			|and| |hence| |you| |may| |wish| |to| |undo| |this| |command|
			|.| CR)))
	  T)
	 ((EQUAL (CADR INSTR) GOAL-NAME)
	  (BLURB NIL |**NO| |CHANGE**| |--| |You| |gave| |the| |name| |of|
		 |the| |current| |goal| |.| CR))
	 (T
	  (BLURB ((NAME (CADR INSTR))) |**NO| |CHANGE**| |--| |There| |is| |no|
		 |goal| |named| (!PPR NAME (QUOTE |.|)) CR))))
      ;so, instruction is simply CHANGE-GOAL
      ((FIND-UNPROVED-GOAL-AND-INFO OTHER-GOALS)
       T)
      (T
       (BLURB NIL |**NO| |CHANGE**| |--| |All| |other| |goals| |have|
	      |true| |conclusions| |.| CR)))))

(DEFUN CHANGE-STATE-WITH-CHANGE-GOAL (INSTR STATE)
  (LET-FIELDS
   (COND
    ((CHANGE-STATE-WITH-CHANGE-GOAL-CHK INSTR GOAL-NAME OTHER-GOALS)
     (LET ((NEW-GOAL-AND-INFO
	    (COND ((CONSP INSTR)
		   (ITERATE FOR GL-INFO IN OTHER-GOALS
			    WHEN (EQUAL (ACCESS GOAL GOAL-NAME (CAR GL-INFO))
					(CADR INSTR))
			    DO (RETURN GL-INFO)))
		  (T
		   (FIND-UNPROVED-GOAL-AND-INFO OTHER-GOALS)))))
	  (BLURB ((NAME (ACCESS GOAL GOAL-NAME (CAR NEW-GOAL-AND-INFO))))
		 |Now| |proving| (!PPR NAME (QUOTE |.|)) CR)
	  (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR))
	   (LET ((STATE1
		  (SWAP-GOALS STATE
			      (ACCESS GOAL GOAL-NAME (CAR NEW-GOAL-AND-INFO))
			      NEW-GOAL-AND-INFO)))
		(IF (AND (CONSP INSTR) (CDDR INSTR))
		    (PUSH-GOAL-TO-END GOAL-NAME STATE1)
		    STATE1))))))))

(DEFUN CHANGE-STATE-WITH-CLAIM-CHK (INSTR ASSUMPTIONS CLAIM CONC ABBREVIATIONS)
  (SETQ ALL-LEMMAS-USED NIL)
  (LET ((TO-PROVE
	 (MAKE-IMPLICATION ASSUMPTIONS CLAIM))
	HINTS)
       (COND
	((NOT (MATCH INSTR (CLAIM & HINTS)))
	 (COND
	  ((PROVE1 TO-PROVE nil)
	   (SETQ ALL-LEMMAS-USED (UNION-EQ (ALL-FNNAMES CLAIM) ALL-LEMMAS-USED))
	   T)
	  (T
	   (BLURB NIL |**NO| |CHANGE**| |--| |Proof| |attempt| |unsuccessful| |.| CR))))
	((EQ HINTS (QUOTE TAUT))
	 (COND
	  ((TAUTOLOGYP
	    (FCONS-TERM* (QUOTE IMPLIES)
			 (DUMB-NEGATE-LIT CONC)
			 TO-PROVE))
	   (SETQ ALL-LEMMAS-USED (ALL-FNNAMES CLAIM))
	   T)
	  (T
	   (BLURB NIL |**NO| |CHANGE**| |--| |The| |active| |hypotheses| |and| |negated|
		  |conclusion| |do| |not| |imply| |the| |claimed| |term| |.| CR))))
	((ATOM HINTS)			;I'll note the new function names when making the new record
	 T)
	((NULL (ERROR1-SET (MATCH! (CHK-ACCEPTABLE-HINTS-ABB HINTS ABBREVIATIONS) (LIST HINTS))))
	 NIL)
	((PROVE1 (APPLY-HINTS HINTS TO-PROVE) nil) 
	 (SETQ ALL-LEMMAS-USED
	       (UNION-EQ ALL-LEMMAS-USED
			 (UNION-EQ
			  (EXTRACT-DEPENDENCIES-FROM-HINTS
			   HINTS)
			  (ALL-FNNAMES CLAIM))))
	 T)
	(T
	 (BLURB NIL |**NO| |CHANGE**| |--| |Proof| |attempt| |unsuccessful| |.| CR)))))

(DEFUN CHANGE-STATE-WITH-CLAIM (INSTR STATE)
  (LET-FIELDS
   (LET (NEW-GOAL-NAME (CLAIM (TRANSLATE-AND-CHK-ABB (CADR INSTR) ABBREVIATIONS)) CLAIMED-HYPS
		       (ASSUMPTIONS (ASSUMPTIONS HYPS)) HINTS
		       NEWGOAL
		       new-free-vars)
	(COND
	 ((CHANGE-STATE-WITH-CLAIM-CHK INSTR ASSUMPTIONS CLAIM CONC ABBREVIATIONS)
	  (setq new-free-vars
		(set-diff-eq (all-vars claim)
			     (all-vars-state goal other-goals)))
	  (WHEN NEW-FREE-VARS
		(BLURB (NEW-FREE-VARS) NOTE |:| |The|
		       (PLURAL? NEW-FREE-VARS |variables| |variable|)
		       (!PPR-LIST NEW-FREE-VARS NIL)
		       (PLURAL? NEW-FREE-VARS |are| |is|) |being| |introduced|
		       |into| |the| |proof| |state| |and|
		       (PLURAL? NEW-FREE-VARS |are| |is|)
		       |therefore| |being| |added| |to| |the|
		       |current| |list| |of| |free| |variables| |.| CR))
	  (SETQ CLAIMED-HYPS (MAKE-HYP-FLGS (IMMEDIATE-CONJUNCTS-OF CLAIM)))
	  (COND
	   ((NOT (AND (CDDR INSTR) (ATOM (CADDR INSTR)) (NEQ (CADDR INSTR) 'TAUT)))
	    (MAKE-NEW-RECORD
	     (QUOTE STATE)
	     (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		   (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
			 (UNION-EQ ALL-LEMMAS-USED CUMULATIVE-LEMMAS-USED))
		   (CONS (QUOTE GOAL)
			 (MAKE-NEW-RECORD
			  (QUOTE GOAL)
			  (LIST (CONS (QUOTE HYPS) (APPEND HYPS CLAIMED-HYPS)))
			  GOAL))
		   (cons (quote free-variables)
			 (union-eq free-variables new-free-vars)))
	     STATE))
	   (T				;so there is a new goal -- not using UPDATE-OTHER-GOALS
					;because I want the same CURRENT-ADDR-R at the new goal
	    (SETQ NEW-GOAL-NAME (CAR (NEW-GOAL-NAMES GOAL-NAME OTHER-GOALS 1)))
	    (SETQ NEWGOAL
		  (MAKE-NEW-RECORD
		   (QUOTE GOAL)
		   (LIST (CONS (QUOTE HYPS)
			       (APPEND (MAKE-HYP-FLGS ASSUMPTIONS)
				       (LIST (CONS
					      (DUMB-NEGATE-LIT CLAIM)
					      (QUOTE A)))))
			 (CONS (QUOTE DEPENDS-ON) NIL)
			 (CONS (QUOTE GOAL-NAME) NEW-GOAL-NAME))
		   GOAL))
	    (BLURB (NEW-GOAL-NAME) |Creating| |one| |new| |subgoal| |,|
		   (!PPR NEW-GOAL-NAME (QUOTE |.|)) CR)
	    (MAKE-NEW-RECORD
	     (QUOTE STATE)
	     (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		   (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
			 (UNION-EQ (ALL-FNNAMES CLAIM) CUMULATIVE-LEMMAS-USED))
		   (CONS (QUOTE GOAL)
			 (MAKE-NEW-RECORD
			  (QUOTE GOAL)
			  (LIST (CONS (QUOTE HYPS)
				      (APPEND HYPS CLAIMED-HYPS))
				(CONS (QUOTE DEPENDS-ON)
				      (CONS NEW-GOAL-NAME DEPENDS-ON)))
			  GOAL))
		   (CONS (QUOTE OTHER-GOALS)
			 (CONS
			  (LIST NEWGOAL CURRENT-TERM GOVERNORS CURRENT-ADDR-R)
			  (CLOSE-ONCE (LIST NEW-GOAL-NAME) GOAL-NAME OTHER-GOALS)))
		   (cons (quote free-variables)
			 (union-eq free-variables new-free-vars)))
	     STATE))))))))

(DEFUN CHANGE-STATE-WITH-COMMENT (INSTR STATE)
  (LET-FIELDS
    (MAKE-NEW-RECORD
      (QUOTE STATE)
      (LIST (CONS (QUOTE INSTRUCTION)
		  (COND
		    ((ATOM INSTR)
		     `(COMMENT ***CURRENT-TERM***
			       ,(UNTRANSLATE-ABB CURRENT-TERM
						 ABBREVIATIONS)))
		    (T
		     INSTR))))
      STATE)))

(DEFUN CHANGE-STATE-WITH-CONTRADICT-CHK (INSTR HYPS CURRENT-ADDR-R)
  (COND
    ((NOT (AND (INTEGERP (CADR INSTR))
	       (> (CADR INSTR) 0)))
     (BLURB ((N (CADR INSTR))) |**NO| |CHANGE**| |--| |The| |argument| |to|
	    CONTRADICT |must| |be| |a| |positive| |integer| |,| |and| |thus|
	    (!PPR N NIL) |is| |an| |inappropriate| |argument| |.| CR))
    ((NOT (<= (CADR INSTR) (LENGTH HYPS)))
     (BLURB ((N (CADR INSTR)) (K (LENGTH HYPS)))
	    |**NO| |CHANGE**| |--| |The| |argument| |to| CONTRADICT |may| |not|
	    |exceed| |the| |number| |of| |hypotheses| |,| |namely| (!PPR K (QUOTE |,|))
	    |and| |thus| (!PPR N NIL) |is| |an| |inappropriate| |argument| |.| CR))
    (CURRENT-ADDR-R
     (BLURB NIL |**NO| |CHANGE**| |--| |The| |command| CONTRADICT |may| |not| |be|
	    |given| |when| |one| |is| |inside| |the| |current| |conclusion| |.|
	    |Perhaps| |you| |should| |try| TOP |first| |.| CR))
    ((EQ (CDR (NTH (1- (CADR INSTR)) HYPS)) (QUOTE H))
     (BLURB ((N (CADR INSTR)) (CMD (LIST (QUOTE SHOW-HYPS) (CADR INSTR))))
	    |**NO| |CHANGE**| |--| |The| |hypothesis| |numbered| (!PPR N NIL)
	    |is| |currently| |hidden| |.| |First| |submit| |the| |command|
	    (!PPR CMD NIL) |if| |you| |wish| |.| CR))
    (T T)))

(DEFUN CHANGE-STATE-WITH-CONTRADICT (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-CONTRADICT-CHK INSTR HYPS CURRENT-ADDR-R)
       (LET ((NEWCONC (DUMB-NEGATE-LIT
			(CAR (NTH (1- (CADR INSTR)) HYPS)))))
	 (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE CURRENT-TERM) NEWCONC)
		 (CONS (QUOTE GOAL)
		       (MAKE-NEW-RECORD
			 (QUOTE GOAL)
			 (LIST (CONS (QUOTE CONC) NEWCONC)
			       (CONS (QUOTE HYPS)
				     (ITERATE FOR HYP-FLG IN HYPS
					   AS INDEX FROM 1
					   COLLECT
					     (COND
					       ((INT= INDEX (CADR INSTR))
						(CONS (DUMB-NEGATE-LIT CONC) (QUOTE A)))
					       (T HYP-FLG)))))
			 GOAL)))
	   STATE))))))

(DEFUN CHANGE-STATE-WITH-COPY (INSTR STATE &AUX TEMP)
  (LET-FIELDS
   (COND
    ((AND (CONSP INSTR)
	  (SETQ TEMP
		(ITERATE FOR N IN (CDR INSTR)
			 WHEN
			 (NOT (AND (INTEGERP N)
				   (> N 0)
				   (<= N (LENGTH HYPS))
				   (EQ (CDR (NTH (1- N) HYPS)) 'A)))
			 COLLECT N)))
     (BLURB (TEMP) |**NO| |CHANGE**| |--| |The| |arguments| |to| COPY
	    |must| |be| |indices| |of| |active| |top-level|
	    |hypotheses| |,| |but| (!PPR-LIST TEMP NIL)
	    (PLURAL? TEMP |are| |is|) |not| |.| CR)) 
    (T 
     (LET* ((OTHER-VARS
	     (UNION-EQ-LIST
	      (ITERATE FOR GL-INFO IN OTHER-GOALS
		       COLLECT (ALL-VARS-GOAL (CAR GL-INFO)))))
	    (COPY-HYPS
	     (IF (ATOM INSTR)
		 (ITERATE FOR HYP-FLG IN HYPS
			  AS I FROM 1
			  WHEN (EQ (CDR HYP-FLG) 'A)
			  COLLECT I)
		 (CDR INSTR))) 
	    (FREE-HYP-VARS
	     (INTERSECTION-EQ FREE-VARIABLES
			      (ALL-VARS-LST
			       (ITERATE FOR HYP-FLG IN HYPS
					AS I FROM 1
					WHEN (MEMBER-EQUAL I COPY-HYPS)
					COLLECT (CAR HYP-FLG)))))
	    (STATE-VARS (UNION-EQ (ALL-VARS-GOAL GOAL) OTHER-VARS))
	    (COPY-ALIST (MAKE-COPY-ALIST
			 (SET-DIFF-EQ FREE-HYP-VARS OTHER-VARS)
			 STATE-VARS))
	    (UNCOPIED-VARS (INTERSECTION-EQ FREE-HYP-VARS OTHER-VARS)))
	   (WHEN UNCOPIED-VARS
		 (BLURB (UNCOPIED-VARS) NOTE |--| |Not| |copying| |the|
			(PLURAL? UNCOPIED-VARS |variables| |variable|)
			(!PPR-LIST UNCOPIED-VARS NIL)  |since|
			(PLURAL? UNCOPIED-VARS
				 (PROGN |they| |occur|)
				 (PROGN |it| |occurs|))
			|outside| |the| |current| |goal's| |hypotheses| |:| CR)
		 (SHOW-VARIABLES-WITH-GOALS UNCOPIED-VARS STATE)
		 (BLURB NIL CR))
	   (COND
	    ((AND (ATOM INSTR)
		  (NULL (SETQ COPY-HYPS
			      (ITERATE FOR HYP-FLG IN HYPS
				       WITH COPY-VARS = (MAPCAR #'CAR COPY-ALIST)
				       AS I FROM 1
				       WHEN (AND (MEMBER-EQUAL I COPY-HYPS)
						 (INTERSECTP-EQ COPY-VARS
								(ALL-VARS (CAR HYP-FLG))))
				       COLLECT I))))
	     (BLURB NIL |**NO| |CHANGE**| |--| |No| |active| |hypotheses| |have|
		    |any| |free| |variables| |that| |may| |be| |copied| |.| CR))
	    (T
	     (IF COPY-ALIST
		 (BLURB (COPY-HYPS
			 (X (ITERATE FOR I FROM 1 TO (LENGTH COPY-HYPS)
				     WITH LENGTH = (LENGTH HYPS)
				     COLLECT (+ I LENGTH)))
			 (CA (ITERATE FOR X IN COPY-ALIST COLLECT (LIST (CAR X) (CDR X)))))
			|#| |Copying| (PLURAL? COPY-HYPS |hypotheses| |hypothesis|)
			(!PPR-LIST COPY-HYPS NIL) |to|
			(PLURAL? COPY-HYPS |hypotheses| |hypothesis|)
			(!PPR-LIST X NIL)
			|by| |renaming| |free| |variables|
			|according| |to| |the| |substitution| (!PPR CA (QUOTE |.|)) CR)
		 (BLURB (COPY-HYPS
			 (X (ITERATE FOR I FROM 1 TO (LENGTH COPY-HYPS)
				     WITH LENGTH = (LENGTH HYPS)
				     COLLECT (+ I LENGTH))))
			|#| WARNING |--| |No| |free| |variables| |have| |been| |renamed|
			|during| |the| COPY |of|
			(PLURAL? COPY-HYPS |hypotheses| |hypothesis|)
			(!PPR-LIST COPY-HYPS NIL)
			|to| (PLURAL? COPY-HYPS |hypotheses| |hypothesis|)
			(!PPR-LIST X (QUOTE |.|)) CR))
	     (MAKE-NEW-RECORD
	      (QUOTE STATE)
	      (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		    (CONS (QUOTE GOAL)
			  (MAKE-NEW-RECORD
			   (QUOTE GOAL)
			   (LIST
			    (CONS
			     (QUOTE HYPS)
			     (APPEND HYPS
				     (ITERATE FOR HYP-FLG IN HYPS
					      AS I FROM 1
					      WHEN (MEMBER-EQUAL I COPY-HYPS)
					      COLLECT
					      (CONS (SUBLIS-VAR COPY-ALIST
								(CAR HYP-FLG))
						    'A)))))
			   GOAL))
		    (CONS (QUOTE FREE-VARIABLES)
			  (APPEND FREE-VARIABLES (MAPCAR #'CDR COPY-ALIST))))
	      STATE))))))))

(DEFUN CHANGE-STATE-WITH-DEMOTE-CHK (INSTR CURRENT-ADDR-R HYPS)
  (LET ((ASSUMPTIONS (ASSUMPTIONS HYPS))
	BADINDICES)
    (COND
      (CURRENT-ADDR-R
       (BLURB NIL |**NO| |CHANGE**| |--| |You| |must| |be| |at| |the| |top|
	      |of| |the| |current| |conclusion| |in| |order| |to| |use| |the|
	      DEMOTE |command| |.| |Try| TOP |first| |.| CR))
      ((NULL ASSUMPTIONS)
       (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |no| |active|
	      |hypotheses| |.| CR))
      ((ATOM INSTR)
       T)
      ((SETQ BADINDICES
	     (ITERATE FOR X IN (CDR INSTR)
		   WITH LEN = (LENGTH HYPS)
		   WHEN (OR (NOT (INTEGERP X))
			    (<= X 0)
			    (> X LEN)
			    (EQ (CDR (NTH (1- X) HYPS)) (QUOTE H)))
		     COLLECT X))
       (BLURB (BADINDICES) |**NO| |CHANGE**| |--| |The| |arguments| |to| DEMOTE
	      |must| |be| |indices| |of| |active| |top-level| |hypotheses| |,|
	      |but| |the| |following| |are| |not| |:|
	      (!PPR-LIST BADINDICES (QUOTE |.|)) CR))
      (T T))))

(DEFUN CHANGE-STATE-WITH-DEMOTE (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-DEMOTE-CHK INSTR CURRENT-ADDR-R HYPS)
       (LET ((INDICES
	       (COND
		 ((ATOM INSTR)
		  (ITERATE FOR X IN HYPS AS I FROM 1
			WHEN (EQ (CDR X) (QUOTE A))
			  COLLECT I))
		 (T (CDR INSTR))))
	     NEWCONC)
	 (SETQ NEWCONC
	       (MAKE-IMPLICATION
		 (ITERATE FOR I IN INDICES
		       COLLECT (CAR (NTH (1- I) HYPS)))
		 CONC))
	 (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE CURRENT-TERM) NEWCONC)
		 (CONS (QUOTE GOAL)
		       (MAKE-NEW-RECORD
			 (QUOTE GOAL)
			 (LIST (CONS (QUOTE CONC) NEWCONC)
			       (CONS (QUOTE HYPS)
				     (ITERATE FOR HYP IN HYPS
					   AS I FROM 1
					   WHEN (NOT (MEMBER-EQUAL I INDICES))
					     COLLECT HYP)))
			 GOAL)))
	   STATE))))))

(DEFUN CHANGE-STATE-WITH-DISABLE (INSTR STATE)
  (LET-FIELDS
   (LET ((NEW-INSTR-GUTS (SAFECDR INSTR))
	 (NEW-FREWRITE-DISABLED-RULES FREWRITE-DISABLED-RULES)
	 (ALL-NAMES (APPEND ALL-FREWRITE-RULE-NAMES
			    ALL-DEFINED-FUNCTION-NAMES)))
	(COND
	 ((ATOM INSTR)
	  (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE FREWRITE-DISABLED-RULES) ALL-NAMES))
	   STATE))
	 ((ITERATE FOR X IN (CDR INSTR)
		   THEREIS (NOT (SYMBOLP X)))
	  (BLURB NIL |**NO| |CHANGE**| |--| |The| DISABLE |command| |must| |be|
		 |either| |the| |atom| DISABLE |or| |else| |of| |the| |form|
		 |(DISABLE <name-1> ... <name-n>)| |,| |where| |each| |<name-i>| |is|
		 |a| |symbol| |.| CR))
	 (T
	  (ITERATE FOR NAME IN (CDR INSTR)
		   DO
		   (COND
		    ((MEMBER-EQ NAME FREWRITE-DISABLED-RULES)
		     (SETQ NEW-INSTR-GUTS (REMOVE1 NAME NEW-INSTR-GUTS))
		     (BLURB (NAME) WARNING |:| (!PPR NAME NIL) |is| |already|
			    |disabled| |.| CR))
		    ((MEMBER-EQ NAME NEW-FREWRITE-DISABLED-RULES)
		     (SETQ NEW-INSTR-GUTS (REMOVE1 NAME NEW-INSTR-GUTS))
		     (BLURB (NAME) NOTE |:| |You| |specified| |the| |name|
			    (!PPR NAME NIL) |more| |than| |once| EXCL CR))
		    ((AND (NOT (GET NAME 'EVENT))
			  (NOT (GET NAME 'MAIN-EVENT)))
		     (SETQ NEW-INSTR-GUTS (REMOVE1 NAME NEW-INSTR-GUTS))
		     (BLURB (NAME) WARNING |:| (!PPR NAME NIL) |is| |not| |the|
			    |name| |of| |an| |event| |or| |of| |a| |satellite| |of|
			    |an| |event| |.| CR))
		    (T
		     (SETQ NEW-FREWRITE-DISABLED-RULES
			   (CONS NAME NEW-FREWRITE-DISABLED-RULES)))))
	  (COND
	   ((NULL NEW-INSTR-GUTS)
	    (BLURB NIL |**NO| |CHANGE**| CR))
	   (T
	    (MAKE-NEW-RECORD
	     (QUOTE STATE)
	     (LIST (CONS (QUOTE INSTRUCTION) (CONS (QUOTE DISABLE) NEW-INSTR-GUTS))
		   (CONS (QUOTE FREWRITE-DISABLED-RULES) NEW-FREWRITE-DISABLED-RULES))
	     ;; Notice that we don't update dependencies, which could make UNDO-NAME
	     ;; inadequate in case one undoes all arguments to the instruction.  However,
	     ;; we choose to take our chances on this; it's legal (for soundness) to do
	     ;; so, since UNDO-NAME is not supposed to be "entirely sound", and it seems
	     ;; desirable to do so, since we don't want to lose an event to UNDO-NAME simply
	     ;; because of ENABLE or DISABLE commands.  If necessary I could rewrite PROVE-LEMMA
	     ;; to be happy with ENABLE or DISABLE when nothing happens.  Probably in many cases,
	     ;; anything that's enabled or disabled (except when the ENABLE is generated by a macro command,
	     ;; the very case that makes us not want to add the dependency!) is perhaps made
	     ;; a dependent by some other action anyhow.
	     STATE))))))))

(DEFUN CHANGE-STATE-WITH-DISABLE-THEORY (INSTR STATE)
  (LET-FIELDS
   (COND
    ((ATOM INSTR)
     (BLURB NIL |**NO| |CHANGE**| |--| |The| DISABLE-THEORY |command| |requires|
	    |at| |least| |one| |argument| |.| |Perhaps| |you| |meant|
	    DISABLE |.| CR))
    ((ITERATE FOR NAME IN (CDR INSTR)
	      WITH TEMP AND NON-SYMBOLS
	      WHEN (NOT (THEORYP NAME))
	      COLLECT NAME INTO TEMP
	      FINALLY
	      (COND (TEMP
		     (IF
		      (SETQ NON-SYMBOLS (ITERATE FOR X IN TEMP WHEN (NOT (SYMBOLP X)) COLLECT X))
		      (BLURB (NON-SYMBOLS) |**NO| |CHANGE**| |--| |The| DISABLE-THEORY |command|
			     |expects| |a| |list| |of| |symbols| |,| |but| |the|
			     |following| (PLURAL? NON-SYMBOLS |are| |is|) |not| |:|
			     (!PPR-LIST NON-SYMBOLS (quote |.|)) CR)
		      (BLURB (TEMP) |**NO| |CHANGE**| |--| |The| DISABLE-THEORY |command|
			     |expects| |a| |list| |of| |theory| |names| |,| |but| |the|
			     |following| (PLURAL? TEMP |are| |is|) |not| |:|
			     (!PPR-LIST TEMP (quote |.|)) CR))
		     (RETURN T))
		    (T (RETURN NIL))))
     NIL)
    (T
     (MAKE-NEW-RECORD
      (QUOTE STATE)
      (LIST (CONS (QUOTE INSTRUCTION) INSTR)
	    (CONS (QUOTE FREWRITE-DISABLED-RULES) 
		  (UNION-EQ (FORM-THEORY (CDR INSTR))
			    FREWRITE-DISABLED-RULES))
	    (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		  (UNION-EQ (CDR INSTR)
			    CUMULATIVE-LEMMAS-USED)))
      STATE)))))

(DEFUN CHANGE-STATE-WITH-DIVE-CHK (INSTR CURRENT-TERM GOVERNORS)
  (LET (TEMP)
    (COND
      ((ATOM INSTR)
       T)
      ;INSTR is (DIVE N1 ... Nk)
      ((SETQ TEMP
	     (ITERATE FOR X IN (CDR INSTR)
		   WHEN
		     (OR (NOT (INTEGERP X))
			 (< X 1))
		     COLLECT X))
       (BLURB (TEMP) |**NO| |CHANGE**| |--| |All| |the| |arguments| |to| DIVE
	      |must| |be| |positive| |integers| |,| |and| (!PPR-LIST TEMP NIL)
	      (PLURAL? TEMP |are| |is|) |not| |.| CR))
      ((NULL (FETCH-TERM-AND-CL CURRENT-TERM (CDR INSTR) GOVERNORS NIL))
       (BLURB ((LST (CDR INSTR))) |**NO| |CHANGE**| |--| |Can't| DIVE |into|
	      |the| |current| |subterm| |according| |to| |the| |list| |of|
	      |addresses| |you| |have| |given| |,| (!PPR LST (QUOTE |.|)) CR))
      (T T))))

(DEFUN CHANGE-STATE-WITH-DIVE (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-DIVE-CHK INSTR CURRENT-TERM GOVERNORS)
       (LET (LST TEMP)
	 (COND
	   ((ATOM INSTR)
	    (COND
	      ((OR (NULL CURRENT-ADDR-R)
		   (VARIABLEP CURRENT-TERM)
		   (FQUOTEP CURRENT-TERM)
		   (NEQ (FFN-SYMB CURRENT-TERM) (QUOTE IF)))
	       ;Then dive from the top, if possible
	       (COND
		 ((SETQ LST (FIND-NEXT-BRANCH CONC))
		  ;so there is some branch which is not a non-F e.v.
		  (COND
		    ((CAR LST)
		     ;so there is a non-explicit value (or F) to dive to
		     (COND
		       ((EQUAL (REVERSE (CAR LST)) CURRENT-ADDR-R)
			(BLURB NIL |**NO| |CHANGE**| |--| |An| |attempt| |to|
			       DIVE |to| |the| |leftmost| |appropriate| |branch|
			       |showed| |that| |we| |are| |already| |there| EXCL CR))
		       (T
			(SETQ TEMP
			      (FETCH-TERM-AND-CL CONC (CAR LST) NIL NIL))
			(MAKE-NEW-RECORD
			  (QUOTE STATE)
			  (LIST (CONS (QUOTE INSTRUCTION) INSTR)
				(CONS (QUOTE CURRENT-TERM) (CAR TEMP))
				(CONS (QUOTE GOVERNORS) (CDR TEMP))
				(CONS (QUOTE CURRENT-ADDR-R) (REVERSE (CAR LST))))
			  STATE))))
		    (T		;so CONC can't be dived into
		     (BLURB NIL |**NO| |CHANGE**| |--| |Type| |(HELP DIVE)| |if|
			    |you| |wish| |to| |review| |the| |rules| |for| |diving|
			    |into| |an| IF |expression| |.| CR))))
		 (T
		  (BLURB NIL |Goal| |reduces| |to| TRUE |since| |every|
			 |branch| |is| |a| |non-FALSE| |explicit| |value| |.| CR)
		  (MAKE-NEW-RECORD
		    (QUOTE STATE)
		    (LIST (CONS (QUOTE INSTRUCTION) INSTR)
			  (CONS (QUOTE CURRENT-TERM) TRUE)
			  (CONS (QUOTE GOVERNORS) NIL)
			  (CONS (QUOTE CURRENT-ADDR-R) NIL)
			  (CONS (QUOTE GOAL)
				(MAKE-NEW-RECORD
				  (QUOTE GOAL)
				  (LIST (CONS (QUOTE CONC) TRUE))
				  GOAL)))
		    STATE))))
	      ;so current term has function symbol IF, and we are
	      ;not at the top.  Notice that the following LST is
	      ;either NIL or is of the form (CONS (CONS a b) c).
	      ((SETQ LST (FIND-NEXT-BRANCH CURRENT-TERM))
	       ;we know then that there is a suitable branch to dive to
	       (SETQ TEMP
		     (FETCH-TERM-AND-CL CURRENT-TERM (CAR LST) GOVERNORS NIL))
	       (MAKE-NEW-RECORD
		 (QUOTE STATE)
		 (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		       (CONS (QUOTE CURRENT-TERM) (CAR TEMP))
		       (CONS (QUOTE GOVERNORS) (CDR TEMP))
		       (CONS (QUOTE CURRENT-ADDR-R)
			     (APPEND (REVERSE (CAR LST)) CURRENT-ADDR-R)))
		 STATE))
	      ;so every branch under the current subterm is
	      ;a non-F explicit value term
	      ((EQ (TERM-ID-IFF CONC (REVERSE CURRENT-ADDR-R) (QUOTE IFF))
		   (QUOTE IFF))
	       (BLURB NIL |The| |current| |term| |reduces| |to| TRUE |since|
		      |every| |branch| |is| |a| |non-FALSE| |explicit| |value|
		      |and| |only| |propositional| |equivalence| |matters| |at|
		      |the| |current| |term| |.| CR)
	       (MAKE-NEW-RECORD
		 (QUOTE STATE)
		 (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		       (CONS (QUOTE CURRENT-TERM) TRUE)
		       (CONS (QUOTE GOAL)
			     (MAKE-NEW-RECORD
			       (QUOTE GOAL)
			       (LIST (CONS (QUOTE CONC)
					   (DEPOSIT-TERM-R
					     CONC
					     CURRENT-ADDR-R
					     TRUE)))
			       GOAL)))
		 STATE))
	      (T
	       (BLURB NIL |**NO| |CHANGE**| |--| |The| |current| |subterm|
		      |,| |which| |is| |not| |at| |the| |top| |of| |the| |goal| |,|
		      |is| |an| IF |expression| |all| |of| |whose| |branches| |are|
		      |non-FALSE| |explicit| |value| |terms| |.| CR))))
	   (T			;so INSTR is (DIVE ...)
	    (SETQ TEMP (FETCH-TERM-AND-CL CURRENT-TERM (CDR INSTR) GOVERNORS NIL))
	    (MAKE-NEW-RECORD
	      (QUOTE STATE)
	      (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		    (CONS (QUOTE CURRENT-TERM) (CAR TEMP))
		    (CONS (QUOTE GOVERNORS) (CDR TEMP))
		    (CONS (QUOTE CURRENT-ADDR-R)
			  (APPEND (REVERSE (CDR INSTR)) CURRENT-ADDR-R)))
	      STATE))))))))

(DEFUN CHANGE-STATE-WITH-DROP-CHK (INSTR HYPS)
  (LET (TEMP)
    (COND
      ((NULL HYPS)
       (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |no| |top-level|
	      |hypotheses| |to| |drop| |.| CR))
      ((AND (EQUAL INSTR (QUOTE (DROP HIDDEN)))
	    (NOT (ITERATE FOR HYP-FLG IN HYPS
		       THEREIS
			 (EQ (CDR HYP-FLG) (QUOTE H)))))
       (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |no| |hidden|
	      |hypotheses| |.| CR))
      ((AND (CONSP INSTR)
	    (NOT (EQUAL INSTR (QUOTE (DROP HIDDEN))))
	    (SETQ TEMP
		  (ITERATE FOR X IN (CDR INSTR)
			WHEN
			  (NOT (AND (INTEGERP X)
				    (< 0 X)
				    (<= X (LENGTH HYPS))))
			  COLLECT X)))
       (BLURB (TEMP) |**NO| |CHANGE**| |--| |The| |arguments| |to| DROP |must|
	      |either| |be| |the| |single| |key| |word| HIDDEN |or| |else|
	      |must| |all| |be| |integers| |from| | 1 | |to| |the| |number|
	      |of| |hypotheses| |,| |and| (!PPR-LIST TEMP NIL)
	      (PLURAL? TEMP |are| |is|) |not| |.| CR))
      (T T))))

(DEFUN CHANGE-STATE-WITH-DROP (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-DROP-CHK INSTR HYPS)
       (COND
	 ((ATOM INSTR)
	  (MAKE-NEW-RECORD
	    (QUOTE STATE)
	    (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		  (CONS (QUOTE GOAL)
			(MAKE-NEW-RECORD (QUOTE GOAL)
					 (LIST (CONS (QUOTE HYPS) NIL))
					 GOAL)))
	    STATE))
	 ((EQUAL INSTR (QUOTE (DROP HIDDEN)))
	  (ITERATE FOR HYP-FLG IN HYPS AS INDEX FROM 1
		WITH DROPPED-INDICES
		WHEN (EQ (CDR HYP-FLG) (QUOTE H))
		  DO (SETQ DROPPED-INDICES (CONS INDEX DROPPED-INDICES))
		FINALLY
		  (BLURB ((L (REVERSE DROPPED-INDICES))) |Dropping|
			 (PLURAL? L |hypotheses| |hypothesis|)
			 |numbered| (!PPR-LIST L (QUOTE |.|)) CR))
	  (MAKE-NEW-RECORD
	    (QUOTE STATE)
	    (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		  (CONS (QUOTE GOAL)
			(MAKE-NEW-RECORD
			  (QUOTE GOAL)
			  (LIST (CONS (QUOTE HYPS)
				      (ITERATE FOR HYP-FLG IN HYPS
					    WHEN (EQ (CDR HYP-FLG) (QUOTE A))
					      COLLECT HYP-FLG)))
			  GOAL)))
	    STATE))
	 (T
	  (BLURB ((L (CDR INSTR))) |Dropping| (PLURAL? L |hypotheses| |hypothesis|)
		 (!PPR-LIST L (QUOTE |.|)) CR)
	  (MAKE-NEW-RECORD
	    (QUOTE STATE)
	    (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		  (CONS (QUOTE GOAL)
			(MAKE-NEW-RECORD
			  (QUOTE GOAL)
			  (LIST (CONS (QUOTE HYPS)
				      (ITERATE FOR HYP IN HYPS
					    AS I FROM 1
					    WITH DROPPED-INDICES = (CDR INSTR)
					    WHEN (NOT (MEMBER-EQUAL I DROPPED-INDICES)) 
					      COLLECT HYP)))
			  GOAL)))
	    STATE)))))))

(DEFUN CHANGE-STATE-WITH-ENABLE (INSTR STATE)
  (LET-FIELDS
   (LET ((NEW-INSTR-GUTS (SAFECDR INSTR)))
	(COND
	 ((ATOM INSTR)
	  (COND
	   (FREWRITE-DISABLED-RULES
	    (MAKE-NEW-RECORD
	     (QUOTE STATE)
	     (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		   (CONS (QUOTE FREWRITE-DISABLED-RULES) NIL))
	     STATE))
	   (T
	    (BLURB NIL |**NO| |CHANGE**| |--| |All| |fast| |rewrite| |rules|
		   |and| |defined| |functions| |are| |already| |enabled| |.| CR))))
	 ((ITERATE FOR X IN (CDR INSTR)
		   THEREIS (NOT (SYMBOLP X)))
	  (BLURB NIL |**NO| |CHANGE**| |--| |The| ENABLE |command| |must| |be|
		 |either| |the| |atom| ENABLE |or| |else| |of| |the| |form|
		 |(ENABLE <name-1> ... <name-n>)| |,| |where| |each| |<name-i>| |is|
		 |a| |symbol| |.| CR))
	 (T
	  (ITERATE FOR NAME IN (CDR INSTR)
		   WHEN (AND (NOT (GET NAME 'EVENT))
			     (NOT (GET NAME 'MAIN-EVENT)))
		   DO
		   (PROGN
		    (SETQ NEW-INSTR-GUTS (REMOVE1 NAME NEW-INSTR-GUTS))
		    (BLURB (NAME) WARNING |:| (!PPR NAME NIL) |is| |not| |the|
			    |name| |of| |an| |event| |or| |of| |a| |satellite| |of|
			    |an| |event| |.| CR)))
	  (COND
	   (NEW-INSTR-GUTS
	    (MAKE-NEW-RECORD
	     (QUOTE STATE)
	     (LIST (CONS (QUOTE INSTRUCTION) (CONS (QUOTE ENABLE) NEW-INSTR-GUTS))
		   (CONS (QUOTE FREWRITE-DISABLED-RULES)
			 (ITERATE FOR NAME IN FREWRITE-DISABLED-RULES
				  WHEN (NOT (MEMBER-EQ NAME NEW-INSTR-GUTS))
				  COLLECT NAME)))
	     ;; See comment regarding dependencies in code for CHANGE-STATE-WITH-DISABLE.
	     STATE))
	   (T
	    (BLURB NIL |**NO| |CHANGE**| CR))))))))

(DEFUN CHANGE-STATE-WITH-ENABLE-THEORY (INSTR STATE)
  (LET-FIELDS
   (COND
    ((ATOM INSTR)
     (BLURB NIL |**NO| |CHANGE**| |--| |The| ENABLE-THEORY |command| |requires|
	    |at| |least| |one| |argument| |.| |Perhaps| |you| |meant|
	    ENABLE |.| CR))
    ((ITERATE FOR NAME IN (CDR INSTR)
	      WITH TEMP AND NON-SYMBOLS
	      WHEN (NOT (THEORYP NAME))
	      COLLECT NAME INTO TEMP
	      FINALLY
	      (COND (TEMP
		     (IF
		      (SETQ NON-SYMBOLS (ITERATE FOR X IN TEMP WHEN (NOT (SYMBOLP X)) COLLECT X))
		      (BLURB (NON-SYMBOLS) |**NO| |CHANGE**| |--| |The| ENABLE-THEORY |command|
			     |expects| |a| |list| |of| |symbols| |,| |but| |the|
			     |following| (PLURAL? NON-SYMBOLS |are| |is|) |not| |:|
			     (!PPR-LIST NON-SYMBOLS (quote |.|)) CR)
		      (BLURB (TEMP) |**NO| |CHANGE**| |--| |The| ENABLE-THEORY |command|
			     |expects| |a| |list| |of| |theory| |names| |,| |but| |the|
			     |following| (PLURAL? TEMP |are| |is|) |not| |:|
			     (!PPR-LIST TEMP (quote |.|)) CR))
		     (RETURN T))
		    (T (RETURN NIL))))
     NIL)
    (T
     (MAKE-NEW-RECORD
      (QUOTE STATE)
      (LIST (CONS (QUOTE INSTRUCTION) INSTR)
	    (CONS (QUOTE FREWRITE-DISABLED-RULES)
		  (ITERATE FOR NAME IN FREWRITE-DISABLED-RULES
			   WITH THEORY = (FORM-THEORY (CDR INSTR))
			   WHEN (NOT (MEMBER-EQ NAME THEORY))
			   COLLECT NAME))
	    (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		  (UNION-EQ (CDR INSTR)
			    CUMULATIVE-LEMMAS-USED)))
      STATE))
    (T
     (BLURB NIL |**NO| |CHANGE**| CR)))))

(DEFUN CHANGE-STATE-WITH-=-CHK (INSTR TO-PROVE CURRENT-TERM SUBTERM ABBREVIATIONS)
  (LET (HINTS)
       (SETQ ALL-LEMMAS-USED NIL)
       (COND
	((and (neq subterm current-term)
	      (NOT (OCCUR SUBTERM CURRENT-TERM)))
	 (BLURB (SUBTERM) |**NO| |CHANGE**| |--| |The| |term|
		(!TERM SUBTERM NIL) |does| |not| |occur| |in| |the| |current|
		|term| |.| CR))
	((ATOM INSTR) T)
	((AND (CDDR INSTR)
	      (QUOTEP SUBTERM))
	 (BLURB (SUBTERM) |**ILLEGAL| |INSTRUCTION**| |--| |Recall|
		|that| |only| |the| |0-| |and| |1-| |argument| |versions| |of| |the|
		|=| |command| |are| |allowed| |when| |substituting| |for| |an| |explicit|
		|value| |term| |,| |which| |in| |your| |case| |is| |the| |term|
		(!TERM SUBTERM (QUOTE |.|)) CR))
	((MATCH INSTR (= & & HINTS))
	 (COND
	  ((ATOM HINTS)
	   T)
	  ((NULL (ERROR1-SET (MATCH! (CHK-ACCEPTABLE-HINTS-ABB HINTS ABBREVIATIONS)
				     (LIST HINTS))))
	   NIL)
	  ((PROVE1 (APPLY-HINTS HINTS TO-PROVE) nil) 
	   (SETQ ALL-LEMMAS-USED
		 (UNION-EQ ALL-LEMMAS-USED
			   (UNION-EQ
			    (EXTRACT-DEPENDENCIES-FROM-HINTS
			     HINTS)
			    (ALL-FNNAMES TO-PROVE))))
	   T)
	  (T
	   (BLURB NIL |**NO| |CHANGE**| |--| |Equality| |not| |proved| |.| CR))))
	((PROVE1 TO-PROVE nil)
	 (SETQ ALL-LEMMAS-USED (UNION-EQ ALL-LEMMAS-USED (ALL-FNNAMES TO-PROVE)))
	 T)
	(T
	 (BLURB NIL |**NO| |CHANGE**| |--| |Equality| |not| |proved| |.| CR)))))

(DEFUN CHANGE-STATE-WITH-= (INSTR STATE)
  (LET-FIELDS
   (LET (NEWTERM SUBTERM TO-PROVE NEW-CURRENT-TERM NEW-GOAL-NAME REQUIRED-EQUALITY
		 (ASSUMPTIONS (ASSUMPTIONS2 HYPS GOVERNORS)))
	(COND				;define SUBTERM and NEWTERM
	 ((ATOM INSTR)
	  (SETQ SUBTERM CURRENT-TERM)
	  (SETQ NEWTERM			;could be NIL
		(ITERATE FOR HYP IN ASSUMPTIONS
			 WITH X AND Y
			 WHEN
			 (OR (AND (MATCH HYP (EQUAL X Y))
				  (EQUAL X CURRENT-TERM))
			     (AND (MATCH HYP (EQUAL Y X))
				  (EQUAL X CURRENT-TERM)))
			 DO (RETURN Y))))
	 ((CDDR INSTR)
	  (SETQ SUBTERM
		(IF (EQ (CADR INSTR) (QUOTE *))
		    CURRENT-TERM
		    (TRANSLATE-AND-CHK-ABB (CADR INSTR) ABBREVIATIONS)))
	  (SETQ NEWTERM (TRANSLATE-AND-CHK-ABB (CADDR INSTR) ABBREVIATIONS)))
	 (T
	  (SETQ SUBTERM CURRENT-TERM)
	  (SETQ NEWTERM (TRANSLATE-AND-CHK-ABB (CADR INSTR) ABBREVIATIONS))))
	(COND
	 ((NULL NEWTERM)		;so command was the atom = and no match was found
	  (BLURB NIL |**NO| |CHANGE**| |--| |There| |is| |no| |hypothesis|
		 |which| |equates| |the| |current| |term| |with| |a| |term| |.| CR))
	 (T
	  (cond ((not (atom instr))
		 (setq required-equality
		       (if (and (not (eq (cadddr instr) (quote id)))
				(EQUAL subterm current-term)
				(eq (term-id-iff conc
						 (reverse current-addr-r)
						 (quote iff))
				    (quote iff))
				;; so it's legal to replace EQUAL by IFF
				(or (eq (cadddr instr) (quote iff))
				    (not
				     (let (type-alist)
				       (and (our-boolean subterm)
					    (our-boolean newterm)))))) 
			   (progn
			     (blurb nil |Checking| |propositional| |equality| excl cr)
			     (fcons-term* (quote iff) subterm newterm))
			 (progn
			   (when (eq (cadddr instr) (quote iff))
				 (blurb nil |***WARNING***:| |Not| |checking|
					|propositional| |equality| |,| |even|
					|though| IFF |flag| |was| |supplied| |.|
					|See| |(HELP-LONG =).| cr))
			   (FCONS-TERM* (QUOTE EQUAL) SUBTERM NEWTERM))))
		 (SETQ TO-PROVE
		       (MAKE-IMPLICATION
			(IF (AND (CONSP INSTR) (CDDR INSTR) (CDDDR INSTR)
				 (ATOM (CADDDR INSTR)))
			    (FLESH-OUT-ASSUMPTIONS ASSUMPTIONS)
			    ASSUMPTIONS)
			required-equality))))
	  (COND
	   ((CHANGE-STATE-WITH-=-CHK INSTR TO-PROVE CURRENT-TERM SUBTERM ABBREVIATIONS)
	    (SETQ NEW-CURRENT-TERM
		  (MY-SUBST-EXPR NEWTERM SUBTERM CURRENT-TERM))
	    (COND
	     ((AND (CONSP INSTR) (CDDR INSTR) (CDDDR INSTR)
		   (ATOM (CADDDR INSTR))
		   (not (iterate for x in assumptions
				 with term1 and term2
				 thereis
				 (and (match x (equal term1 term2))
				      (or (and (equal term1 subterm)
					       (equal term2 newterm))
					  (and (equal term1 newterm)
					       (equal term2 subterm)))))))
	      (SETQ NEW-GOAL-NAME
		    (CAR (NEW-GOAL-NAMES GOAL-NAME OTHER-GOALS 1)))
	      (MAKE-NEW-RECORD
	       (QUOTE STATE)
	       (LIST
		(CONS (QUOTE INSTRUCTION) INSTR)
		(CONS (QUOTE CURRENT-TERM) NEW-CURRENT-TERM)
		(CONS (QUOTE GOAL)
		      (MAKE-NEW-RECORD
		       (QUOTE GOAL)
		       (LIST (CONS (QUOTE CONC)
				   (DEPOSIT-TERM-R
				    CONC CURRENT-ADDR-R NEW-CURRENT-TERM))
			     (CONS (QUOTE DEPENDS-ON)
				   (CONS NEW-GOAL-NAME DEPENDS-ON)))
		       GOAL))
		(CONS (QUOTE OTHER-GOALS)
		      (UPDATE-OTHER-GOALS
		       OTHER-GOALS
		       GOAL-NAME
		       (LIST (CONS REQUIRED-EQUALITY ASSUMPTIONS))
		       (LIST NEW-GOAL-NAME)))
		(CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		      (UNION-EQ (ALL-FNNAMES NEWTERM) CUMULATIVE-LEMMAS-USED)))
	       STATE))
	     (T
	      (MAKE-NEW-RECORD
	       (QUOTE STATE)
	       (LIST
		(CONS (QUOTE INSTRUCTION) INSTR)
		(CONS (QUOTE CURRENT-TERM) NEW-CURRENT-TERM)
		(CONS (QUOTE GOAL)
		      (MAKE-NEW-RECORD
		       (QUOTE GOAL)
		       (LIST (CONS (QUOTE CONC)
				   (DEPOSIT-TERM-R
				    CONC CURRENT-ADDR-R NEW-CURRENT-TERM)))
		       GOAL))
		(CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		      (UNION-EQ ALL-LEMMAS-USED CUMULATIVE-LEMMAS-USED)))
	       STATE))))))))))

(DEFUN CHANGE-STATE-WITH-GENERALIZE-CHK
  (INSTR current-addr-r all-goal-names vars-of-state state &aux bad-pairs)
  (COND
    (CURRENT-ADDR-R
     ;; It would be slightly awkward to deal with the situation where one is strictly inside
     ;; a term that is being generalized away.
     (BLURB NIL |**NO| |CHANGE**| |--| |Generalization| |may| |only| |be|
	    |applied| |at| |the| |top| |of| |the| |current| |goal| |.|
	    |Try| TOP |first| |.| CR))
    ((AND (CDDR INSTR)
	  (MEMBER-EQUAL (CADDR INSTR) all-goal-names))
     (BLURB ((NAME (CADDR INSTR))) |**NO| |CHANGE**| |--| |The| |proposed|
	    |new| |goal| |name| |,| (!PPR NAME NIL) |is| |already| |taken| |.|
	    |You| |may| |however| |choose| |to| |leave| |out| |the| |third|
	    |argument| |,| |in| |which| |case| |a| |new| |goal| |name| |will| |be|
	    |chosen| |by| |the| |system| |.| CR))
    ((not (consp (cadr instr)))
     (blurb nil |**NO| |CHANGE**| |--| |The| |first| |argument| |to| GENERALIZE
	    |must| |be| |a| |non-empty| |list| |.| CR))
    ((cdr (our-last (cadr instr)))
     (blurb nil |**NO| |CHANGE**| |--| |Your| |first| |argument| |to| GENERALIZE
	    |has| |a| |non-NIL| |last| CDR |.| CR))
    ((setq bad-pairs
	   (iterate for x in (cadr instr) when
		    (not (and (consp x)
			      (consp (cdr x))
			      (null (cddr x))))
		    collect x))
     (BLURB (bad-pairs) |**NO| |CHANGE**| |--| |Each| |element| |of| |the|
	    |first| |argument| |to| GENERALIZE |is| |supposed| |to|
	    |be| |a| |list| |of| |length| |2| |,| |but| (!PPR-list bad-pairs NIL)
	    (plural? bad-pairs |are| |is|) |not| |.| CR))
    (t
     (let ((range (iterate for x in (cadr instr)
			   when (not (eq (cadr x) (quote ?)))
			   collect (cadr x)))
	   bad)
       (cond
	((not (no-duplicatesp range))
	 (blurb nil |**NO| |CHANGE**| |--| |The| |substitution|
		|supplied| |to| GENERALIZE |must| |have| |distinct|
		|variables| |supplied| |in| |the| |second| |positions| |.|
		|That| |is| |,| |the| |generalization| |variables| |that|
		|you| |specify| |must| |be| |distinct| |.| CR))
	((setq bad
	       (iterate for v in range
			when (not (variablep (translate v)))
			collect v))
	 (blurb (bad)
		|**NO| |CHANGE**| |--| |The| |second| |components|
		|of| |the| |substitution| |supplied| |to| GENERALIZE
		|must| |all| |be| |variables| |,| |which| |isn't| |the|
		|case| |for| (!ppr-list bad (quote |.|)) CR))
	((setq bad (intersection-eq range vars-of-state))
	 (blurb nil |**NO| |CHANGE**| |--| |The| |second| |components|
		|of| |the| |substitution| |supplied| |to| GENERALIZE
		|must| |not| |occur| |in| |the| |current| |state| |,|
		|but| |we| |have| |:| CR)
	 (iterate for var in bad
		  do
		  (blurb (var (names
			       (iterate for goal
					in (cons (access state goal state)
						 (mapcar #'car
							 (access state other-goals state)))
					for vars = (all-vars-goal goal)
					when (member-eq var vars)
					collect (access goal goal-name goal))))
			 |#| (!ppr var nil) |occurs| |in| (!ppr-list names nil) CR)))
	(t t))))))

(DEFUN CHANGE-STATE-WITH-GENERALIZE (INSTR STATE)
; I use ELIM-VARIABLE-NAMES simply because it's a long list.  It's ironic
; that in the prover, this list is required to be disjoint from
; GEN-VARIABLE-NAMES.
  (LET-FIELDS
    (LET (GEN-PAIRS NEW-GOAL-NAME NEWCONC NEW-ASSUMPTIONS OTHER-NEW-VARS
	  (OTHER-GOAL-NAMES
	    (ITERATE FOR GL-INFO IN OTHER-GOALS
		  COLLECT (ACCESS GOAL GOAL-NAME (CAR GL-INFO))))
	  (VARS-OF-state (all-vars-state goal other-goals))
	  FREE-VARS-TO-BIND)
      (COND
	((CHANGE-STATE-WITH-GENERALIZE-CHK INSTR current-addr-r
					   (cons goal-name other-goal-names)
					   vars-of-state state)
	 ;; If we get this far, we know that no generalization variable occurs
	 ;; in the current state.  At this point we modify the free-variables
	 ;; to remove any variables that are supposed to be "new".
	 (let ((temp (iterate for x in (cadr instr)
			      when (and (neq (cadr x) (quote ?))
					(member-eq (cadr x) free-variables))
			      collect (cadr x))))
	   (when temp
		 (blurb (temp) |#| WARNING |:| |The|
			(PLURAL? temp |variables| |variable|)
			(!PPR-LIST temp NIL) (PLURAL? temp |are| |is|)
			|being| |removed| |from| |the| |list| |of| |free| |variables| |of|
			|the| |state| |,| |for| |soundness| |reasons| |,|
			|because| |you| |have| |specified| (PLURAL? temp |them| |it|)
			|as| |new| |generalization|
			(PLURAL? temp |variables| |variable|) |.| |This| |should| |be|
			|harmless| |,| |however| |,| |since|
			(plural? temp
				 (progn |none| |of| |these| |variables| |occur|)
				 (progn |this| |variable| |does| |not| |occur|))
			|in| |the| |current| |proof| |state| |.| CR))
	   (setq free-variables (set-diff-eq free-variables temp)))
	 (SETQ OTHER-NEW-VARS
	       (ITERATE FOR X IN (CADR INSTR)
		     WHEN (NEQ (CADR X) (QUOTE ?))
		       COLLECT (CADR X)))
	 (SETQ GEN-PAIRS
	       (COND
		 ((NULL (CADR INSTR))
		  (LIST (CONS CURRENT-TERM
			      (CAR (GEN-VARS1 ELIM-VARIABLE-NAMES VARS-OF-state 1)))))
		 (T
		  (ITERATE FOR X IN (CADR INSTR)
			WITH NEWVARS = (SET-DIFF ELIM-VARIABLE-NAMES
						 (APPEND free-variables OTHER-NEW-VARS VARS-OF-state))
			AND VAR
			for temp = (TRANSLATE-AND-CHK-ABB (CAR X) ABBREVIATIONS) 
			COLLECT
			  (COND
			    ((EQ (CADR X) (QUOTE ?))
			     (SETQ VAR (CAR NEWVARS))
			     (SETQ NEWVARS (CDR NEWVARS))
			     (CONS temp VAR))
			    (T
			     (CONS temp (CADR X))))))))
	 (SETQ NEWCONC (SUBLIS-EXPR1 GEN-PAIRS CONC))
	 (SETQ GENERALIZE-LEMMA-NAMES NIL) 
	 (SETQ NEW-ASSUMPTIONS
	       (APPEND 
		(ITERATE FOR HYP-FLG IN HYPS
			 WHEN (EQ (CDR HYP-FLG) (QUOTE A))
			 COLLECT
			 (SUBLIS-EXPR1 GEN-PAIRS (CAR HYP-FLG)))
		(union-equal
		 (ITERATE FOR PAIR IN GEN-PAIRS	;type set hypotheses on new variables
			  APPEND
			  (TYPE-SET-HYPS (CAR PAIR) (CDR PAIR)))
		 (my-GENERALIZE2 gen-pairs))))
	 (SETQ NEW-GOAL-NAME
	       (IF (CDDR INSTR)
		   (CADDR INSTR)
		   (CONS GOAL-NAME
			 (NEW-GOAL-SUFFIX GOAL-NAME OTHER-GOAL-NAMES))))
	 (BLURB (NEW-GOAL-NAME GOAL-NAME) |The| |goal| (!PPR GOAL-NAME NIL)
		|has| |been| |generalized| |to| |the| |new| |goal|
		(!PPR NEW-GOAL-NAME (QUOTE |,|)) |which| |is| |now| |the| |current|
		|goal| |.| CR)
	 (WHEN (AND FREE-VARIABLES	;check this as an optimization
		    (SETQ FREE-VARS-TO-BIND
			  (free-vars-to-bind
			   (intersection-eq free-variables
					    (union-eq (all-vars newconc)
						      (all-vars-lst new-assumptions)))
			   other-goals
			   free-variables
			   (all-vars-lst (mapcar #'car gen-pairs)))))
	       (BLURB (FREE-VARS-TO-BIND)
		      |#| WARNING |:| |The|
		      (PLURAL? FREE-VARS-TO-BIND |variables| |variable|)
		      (!PPR-LIST FREE-VARS-TO-BIND NIL)
		      (PLURAL? FREE-VARS-TO-BIND |are| |is|)
		      |being| |removed| |from| |the| |list| |of| |free| |variables| |of|
		      |the| |current| |proof| |state| |,| |for| |soundness| |reasons| |,|
		      |because|
		      (plural? FREE-VARS-TO-BIND (progn |they| |have|) (progn |it| |has|))
		      |been| |generalized| |away| |but|
		      |still| (plural? FREE-VARS-TO-BIND |"interact"| |"interacts"|)
		      |with| |free| |variables| |in|
		      |the| |resulting| |goal| |.| |(Use| |(HELP-LONG GENERALIZE)| |for|
		      |a| |further| |explanation.)| CR))
	 (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE CURRENT-TERM) NEWCONC)
		 (CONS (QUOTE GOAL)
		       (MAKE-NEW-RECORD
			(QUOTE GOAL)
			(LIST (CONS (QUOTE CONC) TRUE)
			      (CONS (QUOTE DEPENDS-ON)
				    (CONS NEW-GOAL-NAME DEPENDS-ON)))
			GOAL))
		 (CONS (QUOTE OTHER-GOALS)
		       (CONS
			 (LIST
			   (MAKE-NEW-RECORD
			    (QUOTE GOAL)
			    (LIST
			     (CONS (QUOTE CONC) NEWCONC)
			     (CONS (QUOTE HYPS) (MAKE-HYP-FLGS NEW-ASSUMPTIONS))
			     (CONS (QUOTE DEPENDS-ON) NIL)
			     (CONS (QUOTE GOAL-NAME) NEW-GOAL-NAME))
			    GOAL)
			   NEWCONC NIL NIL)
			 (CLOSE-ONCE (LIST NEW-GOAL-NAME) GOAL-NAME OTHER-GOALS)))
		 (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		       (UNION-EQ GENERALIZE-LEMMA-NAMES CUMULATIVE-LEMMAS-USED))
		 (CONS (QUOTE FREE-VARIABLES)
		       (IF FREE-VARS-TO-BIND ;an optimization
			   (SET-DIFF-EQ FREE-VARIABLES FREE-VARS-TO-BIND)
			   FREE-VARIABLES)))
	   STATE))))))

(DEFUN CHANGE-STATE-WITH-HIDE-GOVS-CHK (INSTR GOVERNORS)
  (LET (INDICES)
    (COND
     ((NULL GOVERNORS)
      (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |currently| |no|
	     |governors| |.| CR))
     ((ITERATE FOR HYP-FLG IN GOVERNORS
	       NEVER (EQ (CDR HYP-FLG) (QUOTE A)))
      (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |currently| |no|
	     |active| |governors| |.| CR))
     ((ATOM INSTR)
      T)
     ((SETQ INDICES
	    (ITERATE FOR X IN (CDR INSTR)
		     WITH LENGTH = (LENGTH GOVERNORS)
		     WHEN (NOT (AND (INTEGERP X)
				    (< 0 X)
				    (<= X LENGTH)))
		     COLLECT X))
      (BLURB (INDICES) |**NO| |CHANGE**| |--| (!PPR-LIST INDICES NIL)
	     (PLURAL? INDICES |are not| |is not a|) |valid|
	     (PLURAL? INDICES |indices| |index|) |.| CR))
     (T T))))

(DEFUN CHANGE-STATE-WITH-HIDE-GOVS (INSTR STATE)
  (LET-FIELDS
   (COND
    ((CHANGE-STATE-WITH-HIDE-GOVS-CHK INSTR GOVERNORS)
     (COND
      ((ATOM INSTR)
       (MAKE-NEW-RECORD
	(QUOTE STATE)
	(LIST (CONS (QUOTE INSTRUCTION) INSTR)
	      (CONS (QUOTE GOVERNORS)
		    (ITERATE FOR HYP-FLG IN GOVERNORS
			     COLLECT (CONS (CAR HYP-FLG) (QUOTE H)))))
	STATE))
      (T
       (LET ((NEWGOVS
	      (ITERATE FOR HYP-FLG IN GOVERNORS
		       AS I FROM 1
		       WITH FLG
		       COLLECT
		       (COND
			((MEMBER-EQUAL I (CDR INSTR))
			 (COND
			  ((EQ (CDR HYP-FLG) (QUOTE A))
			   (SETQ FLG T)
			   (CONS (CAR HYP-FLG) (QUOTE H)))
			  (T
			   (BLURB (I) CR |**WARNING**| |--| |Governor|
				  |number| (!PPR I NIL) |is| |already|
				  |hidden| EXCL)
			   HYP-FLG)))
			(T HYP-FLG))
		       FINALLY (COND ((NULL FLG)
				      (RETURN (QUOTE *FAILED*)))))))
	 (COND
	  ((EQ NEWGOVS (QUOTE *FAILED*))
	   (BLURB NIL |**NO| |CHANGE**| |--| |none| |of| |the|
		  |indices| |specified| |is| |the| |index| |of| |an|
		  |active| |governor| |.| CR))
	  (T
	   (MAKE-NEW-RECORD
	    (QUOTE STATE)
	    (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		  (CONS (QUOTE GOVERNORS) NEWGOVS))
	    STATE))))))))))

(DEFUN CHANGE-STATE-WITH-HIDE-HYPS-CHK (INSTR HYPS)
  (LET (INDICES)
    (COND
     ((NULL HYPS)
      (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |currently| |no|
	     |top-level| |hypotheses| |.| CR))
     ((ITERATE FOR HYP-FLG IN HYPS
	       NEVER (EQ (CDR HYP-FLG) (QUOTE A)))
      (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |currently| |no|
	     |active| |top-level| |hypotheses| |.| CR))
     ((ATOM INSTR)
      T)
     ((SETQ INDICES
	    (ITERATE FOR X IN (CDR INSTR)
		     WITH LENGTH = (LENGTH HYPS)
		     WHEN (NOT (AND (INTEGERP X)
				    (< 0 X)
				    (<= X LENGTH)))
		     COLLECT X))
      (BLURB (INDICES) |**NO| |CHANGE**| |--| (!PPR-LIST INDICES NIL)
	     (PLURAL? INDICES |are not| |is not a|) |valid|
	     (PLURAL? INDICES |indices| |index|) |.| CR))
     (T T))))

(DEFUN CHANGE-STATE-WITH-HIDE-HYPS (INSTR STATE)
  (LET-FIELDS
   (COND
    ((CHANGE-STATE-WITH-HIDE-HYPS-CHK INSTR HYPS)
     (COND
      ((ATOM INSTR)
       (MAKE-NEW-RECORD
	(QUOTE STATE)
	(LIST (CONS (QUOTE INSTRUCTION) INSTR)
	      (CONS (QUOTE GOAL)
		    (MAKE-NEW-RECORD
		     (QUOTE GOAL)
		     (LIST (CONS (QUOTE HYPS)
				 (ITERATE FOR HYP-FLG IN HYPS
					  COLLECT (CONS (CAR HYP-FLG) (QUOTE H)))))
		     GOAL)))
	STATE))
      (T
       (LET ((NEWHYPS
	      (ITERATE FOR HYP-FLG IN HYPS
		       AS I FROM 1
		       WITH FLG
		       COLLECT
		       (COND
			((MEMBER-EQUAL I (CDR INSTR))
			 (COND
			  ((EQ (CDR HYP-FLG) (QUOTE A))
			   (SETQ FLG T)
			   (CONS (CAR HYP-FLG) (QUOTE H)))
			  (T
			   (BLURB (I) CR |**WARNING**| |--| |Hypothesis|
				  |number| (!PPR I NIL) |is| |already|
				  |hidden| EXCL)
			   HYP-FLG)))
			(T HYP-FLG))
		       FINALLY (COND ((NULL FLG)
				      (RETURN (QUOTE *FAILED*)))))))
	 (COND
	  ((EQ NEWHYPS (QUOTE *FAILED*))
	   (BLURB NIL |**NO| |CHANGE**| |--| |none| |of| |the|
		  |indices| |specified| |is| |the| |index| |of| |an|
		  |active| |hypothesis| |.| CR))
	  (T
	   (MAKE-NEW-RECORD
	    (QUOTE STATE)
	    (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		  (CONS (QUOTE GOAL)
			(MAKE-NEW-RECORD
			 (QUOTE GOAL)
			 (LIST (CONS (QUOTE HYPS) NEWHYPS))
			 GOAL)))
	    STATE))))))))))

(DEFUN CHANGE-STATE-WITH-INDUCT-CHK (X CURRENT-ADDR-R CONC)
  (COND
    (CURRENT-ADDR-R
     (BLURB NIL |**NO| |CHANGE**| |--| |You| |must| |be| |at| |the| |top|
	    |of| |the| |goal| |to| INDUCT |.| CR))
    ((INTEGERP X)
     (COND
       ((> X 0)
	T)
       (T
	(BLURB (X) |**NO| |CHANGE**| |--| |An| |integer| |argument| |to|
	       INDUCT |must| |be| |positive| |.| CR))))
    ((AND (ATOM X)
	  (GET X (QUOTE JUSTIFICATIONS))
	  (GET X (QUOTE INDUCTION-MACHINE))	;*** I have wondered if this is necessary.
	  (GET X (QUOTE SDEFN)))
     (LET ((INDUCTION-TERM (FIND-INDUCTION-TERM X CONC)))
       (COND
	 (INDUCTION-TERM
	  (BLURB (INDUCTION-TERM) |Inducting| |according| |to| |the| |scheme|
		 |generated| |from| |the| |definition| |of|
		 (!TERM INDUCTION-TERM (QUOTE |.|)) CR)
	  T)
	 (T
	  (BLURB (X) |**NO| |CHANGE**| |--| |No| |subterm| |of| |the| |conclusion|
		 |with| |function| |symbol| (!PPR X NIL) |has| |distinct| |variable|
		 |arguments| |.| CR)))))
    ((AND			; Pretty much lifted from the INDUCT case of
				; the BMTP definition for CHK-ACCEPTABLE-HINTS
       (EQUAL X (TRANSLATE X))
       (NVARIABLEP X)
       (NOT (FQUOTEP X))
       (GET (FFN-SYMB X) (QUOTE JUSTIFICATIONS))
       (GET (FFN-SYMB X) (QUOTE INDUCTION-MACHINE))
       (GET (FFN-SYMB X) (QUOTE SDEFN))
       (ITERATE FOR V IN (FARGS X) ALWAYS (VARIABLEP V))
       (NO-DUPLICATESP (FARGS X)))
     T)
    (T
     (BLURB ((H (QUOTE (INDUCT (|fn| |v1| |...| |vn|)))))
	    |**NO| |CHANGE**| |--| |The| INDUCT |instruction| |must| |have|
	    |the| |form| (!PPR H NIL) |where|
	    |fn| |is| |a| |recursively| |defined| |function| |and|
	    |the| |vi| |are| |distinct| |variables| |of| |a| |number|
	    |matching| |the| |arity| |of| |fn| |.| |Alternately| |,|
	    |the| |instruction| |may| |be| |of| |the| |form| |(INDUCT g)|
	    |where| |there| |is| |a| |subterm| |of| |the| |conclusion| |of|
	    |the| |form| |(g v1 ... vn)| |with| |the| |vi| |being| |distinct|
	    |variables| |.| CR))))

(DEFUN CHANGE-STATE-WITH-INDUCT (INSTR STATE)
  (LET-FIELDS
    (LET (NEW-GOAL-NAMES NEW-FORMULAS CL-SET
	  INDUCTION-CANDIDATES IND-CAND
	  (SAVED-INSTR INSTR)
	  (ASSUMPTIONS (ASSUMPTIONS2 HYPS GOVERNORS))
	  FREE-VARS-TO-BIND)
      (COND
	((ATOM INSTR)
	 (SETQ INSTR (QUOTE (INDUCT 1)))))
      (COND
	((CHANGE-STATE-WITH-INDUCT-CHK (CADR INSTR) CURRENT-ADDR-R CONC)
	 (COND
	   ((INTEGERP (CADR INSTR))
	    (SETQ INDUCTION-CANDIDATES
		  (MY-INDUCTION-CANDIDATES HYPS CONC))
	    (COND
	      ((NULL INDUCTION-CANDIDATES)
	       (BLURB NIL |**NO| |CHANGE**| |--| |There| |is| |no|
		      |heuristically| |selected| |induction| |scheme| |available|
		      |.| CR))
	      ((< (LENGTH INDUCTION-CANDIDATES) (CADR INSTR))
	       (BLURB ((N (LENGTH INDUCTION-CANDIDATES)))
		      |**NO| |CHANGE**| |--| |There| |are| |only| (!PPR N NIL)
		      |heuristically| |selected| |induction| |candidates| |.| CR))
	      (T
	       (SETQ IND-CAND
		     (NTH (1- (CADR INSTR))
			      INDUCTION-CANDIDATES))
	       (SETQ CL-SET (LIST (LIST (MAKE-IMPLICATION ASSUMPTIONS CONC))))
	       (SETQ NEW-FORMULAS
		     (MY-INDUCTION-FORMULA-OF-CANDIDATE IND-CAND CL-SET))
	       (BLURB ((SCHEME
			 (CAR (GET-SCHEMA-MEASURE-RELATION
				IND-CAND
				CL-SET))))
		      CR |Inducting| |according| |to| |the| |scheme| |:| CR
		      (!PPR SCHEME NIL))
	       (WHEN (AND FREE-VARIABLES ;check this as an optimization
			  (SETQ FREE-VARS-TO-BIND
				(INTERSECTION-EQ FREE-VARIABLES (ALL-VARS-GOAL GOAL))))
		     (BLURB (FREE-VARS-TO-BIND) WARNING |:| |Removing| |the|
			    (PLURAL? FREE-VARS-TO-BIND |variables| |variable|)
			    |from| |the| |list| |of| |free| |variables| |of|
			    |the| |state| |.| |(Use| |(HELP-LONG INDUCT)| |for|
			    |an| |explanation| |.)| CR))
	       (SETQ NEW-GOAL-NAMES
		     (NEW-GOAL-NAMES GOAL-NAME OTHER-GOALS
				     (LENGTH NEW-FORMULAS)))
	       (MAKE-NEW-RECORD
		 (QUOTE STATE)
		 (LIST
		  (CONS (QUOTE INSTRUCTION) SAVED-INSTR)
		  (CONS (QUOTE CURRENT-TERM) TRUE)
		  (CONS (QUOTE GOAL)
			(MAKE-NEW-RECORD
			 (QUOTE GOAL)
			 (LIST (CONS (QUOTE CONC) TRUE)
			       (CONS (QUOTE DEPENDS-ON)
				     (APPEND NEW-GOAL-NAMES DEPENDS-ON)))
			 GOAL))
		  (CONS (QUOTE OTHER-GOALS)
			(UPDATE-OTHER-GOALS
			 OTHER-GOALS
			 GOAL-NAME
			 (ITERATE FOR X IN NEW-FORMULAS
				  COLLECT (CONS X NIL))
			 NEW-GOAL-NAMES))
		  (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
			(UNION-EQ (MY-LEMMAS-USED-OF-CANDIDATE IND-CAND)
				  CUMULATIVE-LEMMAS-USED))
		  (CONS (QUOTE FREE-VARIABLES)
			(IF FREE-VARS-TO-BIND ;an optimization
			    (SET-DIFF-EQ FREE-VARIABLES FREE-VARS-TO-BIND)
			    FREE-VARIABLES)))
		 STATE))))
	   (T
	    (COND
	      ((ATOM (CADR INSTR))
	       (SETQ INSTR
		     (LIST (QUOTE INDUCT)
			   (FIND-INDUCTION-TERM (CADR INSTR) CONC)))))
	    (WHEN (AND FREE-VARIABLES	;check this as an optimization
		       (SETQ FREE-VARS-TO-BIND
			     (INTERSECTION-EQ FREE-VARIABLES
					      (UNION-EQ (FARGS (CADR INSTR))
							(ALL-VARS-GOAL GOAL)))))
		  (BLURB (FREE-VARS-TO-BIND) WARNING |:| |Removing| |the|
			 (PLURAL? FREE-VARS-TO-BIND |variables| |variable|)
			 (!PPR-LIST FREE-VARS-TO-BIND NIL)
			 |from| |the| |list| |of| |free| |variables| |of|
			 |the| |state| |.| |(Use| |(HELP-LONG INDUCT)| |for|
			 |an| |explanation.)| CR))
	    (SETQ NEW-GOAL-NAMES
		  (NEW-GOAL-NAMES GOAL-NAME OTHER-GOALS
				  (LENGTH (GET (FFN-SYMB (CADR INSTR))
						(QUOTE INDUCTION-MACHINE)))))
	    (MAKE-NEW-RECORD
	      (QUOTE STATE)
	      (LIST
		(CONS (QUOTE INSTRUCTION) INSTR)
		(CONS (QUOTE CURRENT-TERM) TRUE)
		(CONS (QUOTE GOAL)
		      (MAKE-NEW-RECORD
			(QUOTE GOAL)
			(LIST (CONS (QUOTE CONC) TRUE)
			      (CONS (QUOTE DEPENDS-ON)
				    (APPEND NEW-GOAL-NAMES DEPENDS-ON)))
			GOAL))
		(CONS (QUOTE OTHER-GOALS)
		      (UPDATE-OTHER-GOALS
			OTHER-GOALS
			GOAL-NAME
			(ITERATE FOR TAC IN (GET (FFN-SYMB (CADR INSTR))
					       (QUOTE INDUCTION-MACHINE))
			      WITH IND-FORMULA = (MAKE-IMPLICATION ASSUMPTIONS CONC)
			      AND FORMALS =
				(CADR (GET (FFN-SYMB (CADR INSTR)) (QUOTE SDEFN)))
			      AND VARS = (FARGS (CADR INSTR))
			      COLLECT
				(CONS (MAKE-NEW-IND-INSTANCE
					IND-FORMULA TAC FORMALS VARS)
				      NIL))
			NEW-GOAL-NAMES))
		(CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		      (ADD-TO-SET (FFN-SYMB (CADR INSTR)) CUMULATIVE-LEMMAS-USED))
		(CONS (QUOTE FREE-VARIABLES)
		      (IF FREE-VARS-TO-BIND ;an optimization
			  (SET-DIFF-EQ FREE-VARIABLES FREE-VARS-TO-BIND)
			  FREE-VARIABLES)))
	      STATE))))))))

(DEFUN CHANGE-STATE-WITH-NX (INSTR STATE)
  (LET-FIELDS
    (LET (TEMP)
      (COND
	((NULL CURRENT-ADDR-R)
	 (BLURB NIL |**NO| |CHANGE**| |--| |You| |are| |at| |the| |top| |of|
		|the| |conclusion| |.| CR))
	((NULL (SETQ TEMP
		     (FETCH-TERM-AND-CL CONC
					(REVERSE (CONS (1+ (CAR CURRENT-ADDR-R))
						       (CDR CURRENT-ADDR-R)))
					NIL
					(ITERATE FOR PAIR IN GOVERNORS
					      COLLECT (CDR PAIR)))))
	 (BLURB NIL |**NO| |CHANGE**| |--| |Already| |at| |the| |last| |argument|
		|of| |the| |current| |subterm| |.| CR))
	(T
	 (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE CURRENT-TERM) (CAR TEMP))
		 (CONS (QUOTE GOVERNORS) (CDR TEMP))
		 (CONS (QUOTE CURRENT-ADDR-R)
		       (CONS (1+ (CAR CURRENT-ADDR-R))
			     (CDR CURRENT-ADDR-R))))
	   STATE))))))

(DEFUN CHANGE-STATE-WITH-PROMOTE-CHK (CURRENT-ADDR-R CONC INSTR)
  (COND
    (CURRENT-ADDR-R
     (BLURB NIL |**NO| |CHANGE**| |--| |You| |must| |be| |at| |the| |top|
	    |of| |the| |goal| |in| |order| |to| |promote| |the| |antecedents|
	    |of| |an||implication| |.| |Try| TOP |first| |.| CR))
    ((AND (CONSP INSTR) (NOT (INTEGERP (CADR INSTR))))
     (BLURB NIL |**NO| |CHANGE**| |--| |The| |optional| |argument|
	    |to| |the| PROMOTE |command| |must| |be| |an| |integer| |.| CR))
    ((MATCH CONC (IMPLIES & &))
     T)
    (T
     (BLURB NIL |**NO| |CHANGE**| |--| |The| |goal| |must| |be| |of|
	    |the| |form| |(IMPLIES P Q)| |.| CR))))

(DEFUN CHANGE-STATE-WITH-PROMOTE (INSTR STATE)
  (LET-FIELDS
   (COND
      ((CHANGE-STATE-WITH-PROMOTE-CHK CURRENT-ADDR-R CONC INSTR)
       (MAKE-NEW-RECORD
	 (QUOTE STATE)
	 (LIST (CONS (QUOTE INSTRUCTION) INSTR)
	       (CONS (QUOTE CURRENT-TERM) (FARGN CONC 2))
	       (CONS (QUOTE GOAL)
		     (MAKE-NEW-RECORD
		       (QUOTE GOAL)
		       (LIST (CONS (QUOTE HYPS)
				   (MY-UNION-EQUAL
				     HYPS
				     (ITERATE FOR X IN
					      (IF (CONSP INSTR)
						  (IMMEDIATE-CONJUNCTS-OF (FARGN CONC 1) (CADR INSTR))
						(CONJUNCTS-OF (FARGN CONC 1)))
					   COLLECT (CONS X (QUOTE A)))))
			     (CONS (QUOTE CONC) (FARGN CONC 2)))
		       GOAL)))
	 STATE)))))

(DEFUN CHANGE-STATE-WITH-PROVE-CHK (INSTR CURRENT-ADDR-R HYPS CONC ABBREVIATIONS)
  (LET ((TERM-TO-PROVE (MAKE-IMPLICATION (ASSUMPTIONS HYPS) CONC)) HINTS)
       (SETQ ALL-LEMMAS-USED NIL)
       (COND
	(CURRENT-ADDR-R
	 (BLURB NIL |**NO| |CHANGE**| |--| |The| PROVE |command| |should| |only|
		|be| |used| |at| |the| |top| |.| |use| |(= T)| |if| |that| |is| |what|
		|you| |want| |.| CR))
	((ATOM INSTR)
	 (COND
	  ((PROVE1 TERM-TO-PROVE nil)
	   T)
	  (T
	   (BLURB NIL |**NO| |CHANGE**| |--| |Proof| |attempt| |unsuccessful| |.| CR))))
	((CAR (ERROR1-SET (MATCH! (CHK-ACCEPTABLE-HINTS-ABB (CDR INSTR) ABBREVIATIONS)
				  (LIST HINTS))))
	 (COND
	  ((PROVE1 (APPLY-HINTS HINTS TERM-TO-PROVE) nil)
	   (SETQ ALL-LEMMAS-USED
		 (UNION-EQ (EXTRACT-DEPENDENCIES-FROM-HINTS
			    HINTS)
			   ALL-LEMMAS-USED))
	   T)
	  (T
	   (BLURB NIL |**NO| |CHANGE**| |--| |Proof| |attempt| |unsuccessful| |.| CR)))))))

(DEFUN CHANGE-STATE-WITH-PROVE (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-PROVE-CHK INSTR CURRENT-ADDR-R HYPS CONC ABBREVIATIONS)
       (MAKE-NEW-RECORD
	 (QUOTE STATE)
	 (LIST (CONS (QUOTE INSTRUCTION) INSTR)
	       (CONS (QUOTE CURRENT-TERM) TRUE)
	       (CONS (QUOTE GOAL)
		     (MAKE-NEW-RECORD
		       (QUOTE GOAL)
		       (LIST (CONS (QUOTE CONC) TRUE))
		       GOAL))
	       (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		     (UNION-EQ ALL-LEMMAS-USED CUMULATIVE-LEMMAS-USED)))
	 STATE)))))

(DEFUN CHANGE-STATE-WITH-PUSH-CHK
       (INSTR GOAL-NAME OTHER-GOALS CURRENT-TERM CURRENT-ADDR-R)
  (COND
    ((AND (CONSP INSTR)
	  (OR (EQUAL GOAL-NAME (CADR INSTR))
	      (ITERATE FOR GL-INFO IN OTHER-GOALS
		    THEREIS (EQUAL (ACCESS GOAL GOAL-NAME (CAR GL-INFO))
				   (CADR INSTR)))))
     (BLURB ((NAME (CADR INSTR))) |**NO| |CHANGE**| |--| |The| |proposed| |new|
	    |goal| |name| |,| (!PPR NAME (QUOTE |,|)) |is| |already| |taken| |.| CR))
    ((EQUAL CURRENT-TERM TRUE)
     (BLURB NIL |**NO| |CHANGE**| |--| |The| |current| |term| |is|
	    |already| |(TRUE)| |.| CR))
    ((NULL CURRENT-ADDR-R)
     (BLURB NIL WARNING |:| |You| |are| |already| |at| |the| |top| |of|
	    |the| |conclusion| |!| |So| |,| |you| |may| |wish| |to| UNDO
	    |this| |command| |.| CR)
     T)
    (T T)))

(DEFUN CHANGE-STATE-WITH-PUSH (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-PUSH-CHK
	 INSTR GOAL-NAME OTHER-GOALS CURRENT-TERM CURRENT-ADDR-R)
       (LET ((ASSUMPTIONS (ASSUMPTIONS2 HYPS GOVERNORS))
	     (NEW-GOAL-NAME
	       (IF (CONSP INSTR)
		   (CADR INSTR)
		   (CAR (NEW-GOAL-NAMES GOAL-NAME OTHER-GOALS 1))))
	     (NEWTERM
	       (IF (EQ (TERM-ID-IFF CONC (REVERSE CURRENT-ADDR-R) (QUOTE IFF))
		       (QUOTE IFF))
		   TRUE
		   (ASSERT-TERM CURRENT-TERM NIL))))
	 (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE CURRENT-TERM) NEWTERM)
		 (CONS (QUOTE GOAL)
		       (MAKE-NEW-RECORD
			 (QUOTE GOAL)
			 (LIST (CONS (QUOTE CONC)
				     (DEPOSIT-TERM-R CONC CURRENT-ADDR-R NEWTERM))
			       (CONS (QUOTE DEPENDS-ON)
				     (CONS NEW-GOAL-NAME DEPENDS-ON)))
			 GOAL))
		 (CONS (QUOTE OTHER-GOALS)
		       (UPDATE-OTHER-GOALS
			 OTHER-GOALS
			 GOAL-NAME
			 (LIST (CONS CURRENT-TERM ASSUMPTIONS))
			 (LIST NEW-GOAL-NAME))))
	   STATE))))))

(DEFUN CHANGE-STATE-WITH-PUT-CHK (INSTR ABBREVIATIONS FREE-VARIABLES
				       &AUX TEMP)
  (COND
   ((ATOM INSTR)
    (BLURB NIL |**NO| |CHANGE**| |--| |PUT| |requires| |at| |least|
	   |one| |argument| |.| CR))
   ((NOT (SUBSTP (CDR INSTR) ABBREVIATIONS))
    (BLURB ((SUB (CDR INSTR))) |**NO| |CHANGE**| |--| |The|
	   |argument| |list| |for| PUT |must| |be| |a| |list| |of| |pairs| |of|
	   |the| |form| |(var term)| |,| |where| | var | |is| |a| |variable|
	   |and| | term | |is| |a| |term| |(relative| |to| |the| |current|
	   |abbreviations)| |.| |However| |,| |the| |given| |substitution| |,|
	   (!PPR SUB (QUOTE |,|)) |is| |not| |in| |this| |form| |.| CR))
   ((SETQ TEMP (ITERATE FOR PAIR IN (CDR INSTR)
			WHEN (NOT (MEMBER-EQ (CAR PAIR) FREE-VARIABLES))
			COLLECT (CAR PAIR)))
    (BLURB (TEMP) |**NO| |CHANGE**| |--| |The|
	   (PLURAL? TEMP |variables| |variable|) (!PPR-LIST TEMP NIL)
	   (PLURAL? TEMP |are| |is|) |not| |in| |the| |list| |of|
	   |free| |variables| |of| |the| |current| |state| |.| CR))
   (T T)))

(DEFUN CHANGE-STATE-WITH-PUT (INSTR STATE)
  ;; **** I could consider making this more efficient by only substituting
  ;; into goals and then recomputing the governors and current-term in each
  ;; goal.  I might also want to use a version of SUBLIS-VAR which first
  ;; looks to see if the variable is free, to avoid excess consing.
  (LET-FIELDS
   (LET ((ALL-VARS-OTHER-GOALS
	  (UNION-EQ-LIST
	   (ITERATE FOR GL-INFO IN OTHER-GOALS
		    COLLECT (ALL-VARS-GOAL (CAR GL-INFO)))))
	 NEW-FREE-VARS SUBST)
	(COND
	 ((CHANGE-STATE-WITH-PUT-CHK INSTR ABBREVIATIONS FREE-VARIABLES)
	  (LET ((DANGER-VARS (INTERSECTION-EQ (MAPCAR #'CAR (CDR INSTR))
					      ALL-VARS-OTHER-GOALS)))
	       (WHEN DANGER-VARS
		     (BLURB (DANGER-VARS GOAL-NAME) WARNING |:| |The|
			    (PLURAL? DANGER-VARS |variables| |variable|)
			    (!PPR-LIST DANGER-VARS NIL)
			    (PLURAL? DANGER-VARS |occurs| |occurs|)
			    |in| |other| |than| |the| |current| |goal| |,|
			    (!PPR GOAL-NAME (QUOTE |:|)) CR)
		     (SHOW-VARIABLES-WITH-GOALS DANGER-VARS STATE)))
	  (SETQ SUBST
		(ITERATE FOR PAIR IN (CDR INSTR)
			 COLLECT (CONS (CAR PAIR)
				       (TRANSLATE-ABB (CADR PAIR) ABBREVIATIONS))))
	  (SETQ NEW-FREE-VARS
		(SET-DIFF-EQ (SET-DIFF-EQ (ALL-VARS-LST (MAPCAR #'CDR SUBST))
					  (MAPCAR #'CAR SUBST))
			     (UNION-EQ (ALL-VARS-GOAL GOAL)
				       ALL-VARS-OTHER-GOALS)))
	  (WHEN NEW-FREE-VARS
		(BLURB (NEW-FREE-VARS) NOTE |:| |The|
		       (PLURAL? NEW-FREE-VARS |variables| |variable|)
		       (!PPR-LIST NEW-FREE-VARS NIL)
		       (PLURAL? NEW-FREE-VARS |are| |is|) |being| |introduced|
		       |into| |the| |proof| |state| |and|
		       (PLURAL? NEW-FREE-VARS |are| |is|)
		       |therefore| |being| |added| |to| |the|
		       |current| |list| |of| |free| |variables| |.| CR))
	  (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE CURRENT-TERM)
		       (SUBLIS-VAR SUBST CURRENT-TERM))
		 (CONS (QUOTE GOVERNORS)
		       (SUBLIS-VAR-HYPS SUBST GOVERNORS))
		 (CONS (QUOTE GOAL)
		       (SUBLIS-VAR-GOAL SUBST GOAL))
		 (CONS (QUOTE OTHER-GOALS)
		       (ITERATE FOR X IN OTHER-GOALS
				WITH GL AND CUR AND GOVS AND CUR-ADDR-R
				COLLECT (PROGN (MATCH X (LIST GL CUR GOVS CUR-ADDR-R))
					       (LIST (SUBLIS-VAR-GOAL SUBST GL)
						     (SUBLIS-VAR SUBST CUR)
						     (SUBLIS-VAR-HYPS SUBST GOVS)
						     CUR-ADDR-R))))
		 (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		       (UNION-EQ (ALL-FNNAMES-LST (MAPCAR #'CDR SUBST))
				 CUMULATIVE-LEMMAS-USED))
		 (CONS (QUOTE FREE-VARIABLES)
		       (UNION-EQ (SET-DIFF-EQ FREE-VARIABLES (MAPCAR #'CAR SUBST))
				 NEW-FREE-VARS)))
	   STATE))))))

(DEFUN CHANGE-STATE-WITH-REMOVE-ABBREVIATIONS-CHK (INSTR ABBREVIATIONS)
  (LET (BADVARS)
    (COND
      ((ATOM INSTR)
       (COND
	 ((NULL ABBREVIATIONS)
	  (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |already|
		 |no| |abbreviations| |.| CR))
	 (T T)))
      ((SETQ BADVARS
	     (ITERATE FOR X IN (CDR INSTR)
		   WHEN
		     (NOT (ASSOC-EQ X ABBREVIATIONS))
		     COLLECT X))
       (BLURB (BADVARS) |**NO| |CHANGE**| |--| |The| |following| |entries| |are|
	      |not| |currently| |abbreviations| |:| (!PPR-LIST BADVARS (QUOTE |.|)) CR))
      (T T))))

(DEFUN CHANGE-STATE-WITH-REMOVE-ABBREVIATIONS (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-REMOVE-ABBREVIATIONS-CHK INSTR ABBREVIATIONS)
       (COND
	 ((ATOM INSTR)
	  (MAKE-NEW-RECORD
	    (QUOTE STATE)
	    (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		  (CONS (QUOTE ABBREVIATIONS) NIL))
	    STATE))
	 (T
	  (MAKE-NEW-RECORD
	    (QUOTE STATE)
	    (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		  (CONS (QUOTE ABBREVIATIONS)
			(ITERATE FOR PAIR IN ABBREVIATIONS
			      WHEN (NOT (MEMBER-EQ (CAR PAIR) (CDR INSTR)))
				COLLECT PAIR)))
	    STATE)))))))

(DEFUN CHANGE-STATE-WITH-REWRITE-CHK (INSTR CURRENT-TERM ABBREVIATIONS)
  (LET (BADPAIRS)
    (COND
      ((VARIABLEP CURRENT-TERM)
       (BLURB NIL |**NO| |CHANGE**| |--| |It| |is| |impossible| |to|
	      |rewrite| |a| |variable| |.| CR))
      ((ATOM INSTR)
       T)
      ((AND (NOT (INTEGERP (CADR INSTR)))
	    (NOT (SYMBOLP (CADR INSTR))))
       (BLURB NIL |**NO| |CHANGE**| |--| |The| |first| |argument| |to| REWRITE
	      |must| |be| |a| |positive| |integer| |or| |a| |symbol| |.| CR))
      ((AND (INTEGERP (CADR INSTR))
	    (< (CADR INSTR) 1))
       (BLURB NIL |**NO| |CHANGE**| |--| |The| |first| |argument| |to| REWRITE
	      |may| |not| |be| |zero| |or| |negative| |.| CR))
      ((AND (SYMBOLP (CADR INSTR))
	    (ITERATE FOR LEMMA IN (GET (FN-SYMB CURRENT-TERM) (QUOTE LEMMAS))
		  NEVER
		    (EQ (CADR INSTR)
			(ACCESS REWRITE-RULE NAME LEMMA))))
       (BLURB ((RULE-NAME (CADR INSTR)) (FN (FN-SYMB CURRENT-TERM)))
	      |**NO| |CHANGE**| |--| |There| |is|
	      |no| |rewrite| |rule| |named| (!PPR RULE-NAME NIL) |for| |terms|
	      |whose| |top| |function| |symbol| |is| (!PPR FN (QUOTE |.|)) CR))
      ((AND (CDDR INSTR)
	    (CADDR INSTR)
	    (OR (NOT (CONSP (CADDR INSTR)))
		(CDR (OUR-LAST (CADDR INSTR)))
		(SETQ BADPAIRS
		      (ITERATE FOR PAIR IN (CADDR INSTR)
			    WITH VAR AND TERM
			    WHEN
			      (NOT (AND (MATCH PAIR (LIST VAR TERM))
					(SYMBOLP VAR)
					(EQUAL (CHAR (STRING VAR) 0) #\$)
					(error1-set
					 (TRANSLATE-AND-CHK-ABB TERM ABBREVIATIONS))))
			      COLLECT PAIR))))
       (BLURB NIL |**NO| |CHANGE**| |--| |The| |second| |argument| |to|
	      REWRITE |,| |i.e.| |the| |substitution| |,| |must| |be| |a| |list|
	      |of| |pairs| |of| |the| |form| |($variable term)| |,| |where|
	      |$variable| |is| |a| |symbol| |starting| |with| |a| |dollar|
	      |sign| |.| CR)
       (COND
	 (BADPAIRS
	  (BLURB (BADPAIRS) |However| |,| |the| |following| |pairs| |are| |not| |of|
		 |this| |form| |:| (!PPR-LIST BADPAIRS (QUOTE |.|)) CR)))
       NIL)
      (T T))))

(DEFUN CHANGE-STATE-WITH-REWRITE (INSTR STATE)
  (LET-FIELDS
   (COND
    ((CHANGE-STATE-WITH-REWRITE-CHK INSTR CURRENT-TERM ABBREVIATIONS)
     (LET ((ASSUMPTIONS (ASSUMPTIONS2 HYPS GOVERNORS))
	   DOLLAR-SUBST TERM-ID-IFF SUBST-LEMMA-VARS-HYPS
	   RULE-NAME NEW-GOAL-NAMES UNSUBSTITUTED-FREE-VARS
	   NEW-INSTR NEW-CURRENT-TERM-DOLLAR NEW-CURRENT-TERM
	   NEW-FREE-VARS)
	  (SETQ DOLLAR-SUBST
		(COND
		 ((AND (CONSP INSTR) (CDDR INSTR))
		  (ITERATE FOR PAIR IN (CADDR INSTR)
			   COLLECT
			   (CONS (CAR PAIR)
				 (TRANSLATE-AND-CHK-ABB (CADR PAIR) ABBREVIATIONS))))))
	  (SETQ TERM-ID-IFF
		(TERM-ID-IFF CONC (REVERSE CURRENT-ADDR-R) (QUOTE IFF)))
	  (COND
	   ((SETQ SUBST-LEMMA-VARS-HYPS
		  (BEST-SUBST-LEMMA-VARS-HYPS
		   (APPLICABLE-REWRITES
		    CURRENT-TERM TERM-ID-IFF
		    (COND ((CONSP INSTR) (CADR INSTR)))
		    (COND ((AND (CONSP INSTR) (CADDDR INSTR))
			   (TRANSLATE (CADDDR INSTR)))))
		   ASSUMPTIONS DOLLAR-SUBST))
	    (COND 
	     ((EQ SUBST-LEMMA-VARS-HYPS 'CONTRADICTION)
	      (BLURB NIL |**NO| |CHANGE**| |--| |Contradiction| |in| |the| |current|
		     |hypotheses| |and| |governors| |.| |The| |command| S
		     |may| |be| |useful| |in| |this| |context| |.| CR))
	     (T
	      (SETQ UNSUBSTITUTED-FREE-VARS
		    (SET-DIFF (CADDR SUBST-LEMMA-VARS-HYPS)
			      (UNDOLLAR-FY-LST (DOMAIN DOLLAR-SUBST))))
	      (SETQ RULE-NAME
		    (ACCESS REWRITE-RULE NAME (CADR SUBST-LEMMA-VARS-HYPS)))
	      (SETQ NEW-CURRENT-TERM-DOLLAR
		    (SUBLIS-VAR (APPEND DOLLAR-SUBST (CAR SUBST-LEMMA-VARS-HYPS))
				(FARGN (ACCESS REWRITE-RULE CONCL
					       (CADR SUBST-LEMMA-VARS-HYPS))
				       2)))
	      (COND ((AND UNSUBSTITUTED-FREE-VARS
			  (OR (CDDDR SUBST-LEMMA-VARS-HYPS)
			      ;;			 (ITERATE FOR ATOM
			      ;;				  IN (ALL-ATOMS NEW-CURRENT-TERM-DOLLAR) 
			      ;;				  THEREIS
			      ;;				  (AND (EQUAL (CHAR (STRING ATOM) 0) #\$)
			      ;;				       (NOT (MEMBER-EQ 
			      ;;					     ATOM
			      ;;					     (DOMAIN DOLLAR-SUBST))))
			      ;;         )
			      (INTERSECTION-EQ UNSUBSTITUTED-FREE-VARS
					       (ALL-ATOMS
						NEW-CURRENT-TERM-DOLLAR))))
		     (BLURB (UNSUBSTITUTED-FREE-VARS RULE-NAME) |**WARNING**| |:|
			    |The| |following| |variables| |occur| |free| |in|
			    |the| |lemma| (!PPR RULE-NAME NIL) |but| |were|
			    |not| |explicitly| |substituted| |for| |:|
			    (!PPR-LIST UNSUBSTITUTED-FREE-VARS (QUOTE |.|)) CR)))
	      (SETQ NEW-INSTR
		    (COND
		     ((AND (NOT (ATOM INSTR))
			   (INTEGERP (CADR INSTR))
			   (CONSP (CDR (APPLICABLE-REWRITES
					CURRENT-TERM TERM-ID-IFF RULE-NAME))))
		      `(REWRITE ,RULE-NAME ,(CADDR INSTR)
				,(UNTRANSLATE
				  (UNDOLLAR-FY
				   (MAKE-IMPLICATION
				    (ACCESS REWRITE-RULE
					    HYPS (CADR SUBST-LEMMA-VARS-HYPS))
				    (ACCESS REWRITE-RULE
					    CONCL (CADR SUBST-LEMMA-VARS-HYPS)))))))
		     ((ATOM INSTR)
		      `(REWRITE ,RULE-NAME))
		     (T
		      `(REWRITE ,RULE-NAME ,@(CDDR INSTR)))))
	      (SETQ NEW-CURRENT-TERM (UNDOLLAR-FY NEW-CURRENT-TERM-DOLLAR))
	      (BLURB (RULE-NAME) |Rewriting| |with|
		     (!PPR RULE-NAME (QUOTE |.|)) CR)
	      (SETQ NEW-FREE-VARS
		    (SET-DIFF-EQ (ALL-VARS-LST
				  (CONS NEW-CURRENT-TERM
					(CDDDR SUBST-LEMMA-VARS-HYPS)))
				 (ALL-VARS-STATE GOAL OTHER-GOALS)))
	      (WHEN NEW-FREE-VARS
		    (BLURB (NEW-FREE-VARS) CR NOTE |:| |The|
			   (PLURAL? NEW-FREE-VARS |variables| |variable|)
			   (!PPR-LIST NEW-FREE-VARS NIL)
			   (PLURAL? NEW-FREE-VARS |are| |is|) |being| |introduced|
			   |into| |the| |proof| |state| |and|
			   (PLURAL? NEW-FREE-VARS |are| |is|)
			   |therefore| |being| |added| |to| |the|
			   |current| |list| |of| |free| |variables| |.| CR))
	      (COND
	       ((NULL (CDDDR SUBST-LEMMA-VARS-HYPS))
		(MAKE-NEW-RECORD
		 (QUOTE STATE)
		 (LIST (CONS (QUOTE INSTRUCTION) NEW-INSTR)
		       (CONS (QUOTE CURRENT-TERM) NEW-CURRENT-TERM)
		       (CONS (QUOTE GOAL)
			     (MAKE-NEW-RECORD
			      (QUOTE GOAL)
			      (LIST (CONS (QUOTE CONC)
					  (DEPOSIT-TERM-R
					   CONC CURRENT-ADDR-R NEW-CURRENT-TERM)))
			      GOAL))
		       (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
			     (ADD-TO-SET RULE-NAME CUMULATIVE-LEMMAS-USED)))
		 STATE))
	       (T
		(SETQ NEW-GOAL-NAMES
		      (NEW-GOAL-NAMES GOAL-NAME OTHER-GOALS
				      (LENGTH (CDDDR SUBST-LEMMA-VARS-HYPS))))
		(MAKE-NEW-RECORD
		 (QUOTE STATE)
		 (LIST (CONS (QUOTE INSTRUCTION) NEW-INSTR)
		       (CONS (QUOTE CURRENT-TERM) NEW-CURRENT-TERM)
		       (CONS (QUOTE GOAL)
			     (MAKE-NEW-RECORD
			      (QUOTE GOAL)
			      (LIST (CONS (QUOTE CONC)
					  (DEPOSIT-TERM-R
					   CONC CURRENT-ADDR-R NEW-CURRENT-TERM))
				    (CONS (QUOTE DEPENDS-ON)
					  (APPEND NEW-GOAL-NAMES DEPENDS-ON)))
			      GOAL))
		       (CONS (QUOTE OTHER-GOALS)
			     (UPDATE-OTHER-GOALS
			      OTHER-GOALS GOAL-NAME
			      (ITERATE FOR HYP IN (CDDDR SUBST-LEMMA-VARS-HYPS)
				       COLLECT (CONS HYP ASSUMPTIONS))
			      NEW-GOAL-NAMES))
		       (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
			     (ADD-TO-SET RULE-NAME CUMULATIVE-LEMMAS-USED))
		       (CONS (QUOTE FREE-VARIABLES)
			     (UNION-EQ FREE-VARIABLES NEW-FREE-VARS)))
		 STATE))))))
					;no applicable rewrites
	   ((AND (NOT (ATOM INSTR))
		 (INTEGERP (CADR INSTR)))
	    (BLURB ((N (CADR INSTR))) |No| |rewriting| |done| |,| |as|
		   (!PPR N NIL) |is| |not| |a| |valid| |index| |.|
		   |Try| SHOW-REWRITES |to| |see| |your| |choices| |,| |or|
		   |supply| |the| |name| |of| |a| |rule| |.| CR))
	   ((NOT (ATOM INSTR))
	    (BLURB ((NAME (CADR INSTR)) (x (cadddr instr)))
		   |**NO| |CHANGE**| |--| |The|
		   |rewrite| |rule| (!PPR NAME NIL) |does| |not| |apply|
		   (cond (x |as| |shown|))
		   |at| |the| |current| |subterm| |.| CR))
	   (T
	    (BLURB NIL |**NO| |CHANGE**| |--| |No| |rewrite| |rules| |apply|
		   |here| |.| CR))))))))

(DEFUN CHANGE-STATE-WITH-S (INSTR STATE)
  (LET-FIELDS
   (LET ((TYPE-ALIST (HYPS-TYPE-ALIST (APPEND HYPS GOVERNORS)))
	 EXTRACTED-FREWRITE-RULES)
	(COND
	 ((AND (CONSP INSTR)
	       (NOT (EQ (CADR INSTR) (QUOTE ALL)))
	       (NOT (EQ (CADR INSTR) (QUOTE LEMMAS)))
	       (NOT (AND (INTEGERP (CADR INSTR))
			 (>= (CADR INSTR) 0)))
	       (NOT (or (null (cadr instr))
			(AND (CONSP (CADR INSTR))
			     (NULL (CDR (OUR-LAST (CADR INSTR))))))))
	  (BLURB NIL |**NO| |CHANGE**| |--| |The| |only| |arguments| |allowed|
		 |for| S |are| ALL |,| LEMMAS |,| |a| |nonnegative| |integer| |,|
		 |or| |a| |list| |of| |names|
		 |of| |fast| |rewrite| |rules| |.| CR))
	 ((EQ TYPE-ALIST (QUOTE CONTRADICTION))
	  (COND
	   ((OR (NULL CURRENT-ADDR-R)
		(EQ (HYPS-TYPE-ALIST HYPS) (QUOTE CONTRADICTION)))
	    (BLURB NIL |Contradiction| |in| |the| |hypotheses|
		   |of| |the| |current| |goal| EXCL CR)
	    (MAKE-NEW-RECORD
	     (QUOTE STATE)
	     (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		   (CONS (QUOTE CURRENT-TERM) TRUE)
		   (CONS (QUOTE GOVERNORS) NIL)
		   (CONS (QUOTE CURRENT-ADDR-R) NIL)
		   (CONS (QUOTE GOAL)
			 (MAKE-NEW-RECORD
			  (QUOTE GOAL)
			  (LIST (CONS (QUOTE CONC) TRUE))
			  GOAL)))
	     STATE))
	   (T
	    (BLURB NIL |**NO| |CHANGE**| |--| |You| |have| |a| |contradiction| |in|
		   |the| |active| |hypotheses| |and| |governors| |but| |not| |in| |the|
		   |hypotheses| |alone| |,| |so| |you| |may| |wish| |to| |move| |up|
		   |in| |the| |term| |as| |far| |as| |you| |can| |while| |preserving|
		   |this| |property| |and| |then| |use| |(= * new-term T)| |to| |replace|
		   |that| |term| |with| |your| |favorite| |new-term| |while| |using| S
		   |to| |prove| |the| |new| |goal| |.| CR))))
	 (T
	  (SETQ MY-ALL-LEMMAS-USED NIL)
	  (COND
	   ((AND (CONSP INSTR)
		 (CONSP (CADR INSTR)))
	    ;; notice that EXTRACTED-FREWRITE-RULES stays NIL for the command (S NIL).
	    (SETQ EXTRACTED-FREWRITE-RULES
		  (EXTRACT-FREWRITE-RULES (CADR INSTR)))
	    (LET ((RULES-USED
		   (INTERSECTION-EQ 
		    (CADR INSTR)
		    (ITERATE FOR FN-RULES IN EXTRACTED-FREWRITE-RULES
			     APPEND
			     (ITERATE FOR RULE IN (CDR FN-RULES)
				      COLLECT (CADR RULE))))))
		 (COND ((NOT (EQUAL RULES-USED (CADR INSTR)))
			(BLURB ((X (SET-DIFF (CADR INSTR) RULES-USED)))
			       |**WARNING**| |--| |The| |following| |are| |not| |names| |of|
			       |fast| |rewrite| |rules| |:| (!PPR-LIST X (QUOTE |.|)) CR)
			(SETQ INSTR (LIST 'S RULES-USED)))))))
	  (LET
	   ((NEWTERM
	     (LET ((FNS-ALLOWED-FOR-FAST-REWRITE-FLG
		    (OR (EQ INSTR 'S)
			(EQUAL INSTR '(S ALL))))
		   (LEVEL (IF (AND (CONSP INSTR) (INTEGERP (CADR INSTR)))
			      (CADR INSTR)
			      0)))
		  (FAST-REWRITE
		   CURRENT-TERM
		   TYPE-ALIST
		   (TERM-ID-IFF CONC (REVERSE CURRENT-ADDR-R) (QUOTE IFF))
		   (COND
		    ((ATOM INSTR)
		     FREWRITE-RULES)
		    ((and (cadr instr) (ATOM (CADR INSTR)))
		     ALL-FREWRITE-RULES)
		    (T
		     EXTRACTED-FREWRITE-RULES))
		   (COND
		    ((AND (CONSP INSTR)
			  (or (null (cadr instr))
			      (CONSP (CADR INSTR))))
		     NIL)
		    (T FREWRITE-DISABLED-RULES))
		   LEVEL))))
	   (COND
	    ((EQUAL CURRENT-TERM NEWTERM)
	     (BLURB NIL |**NO| |CHANGE**| CR))
	    (T
	     (MAKE-NEW-RECORD
	      (QUOTE STATE)
	      (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		    (CONS (QUOTE CURRENT-TERM) NEWTERM)
		    (CONS (QUOTE GOAL)
			  (MAKE-NEW-RECORD
			   (QUOTE GOAL)
			   (LIST (CONS (QUOTE CONC)
				       (DEPOSIT-TERM-R CONC CURRENT-ADDR-R NEWTERM)))
			   GOAL))
		    (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
			  (UNION-EQ MY-ALL-LEMMAS-USED CUMULATIVE-LEMMAS-USED)))
	      STATE)))))))))

(DEFUN CHANGE-STATE-WITH-S-PROP-CHK (INSTR ABBREVIATIONS)
  (LET (BADHINTS)
       (COND
	((OR (ATOM INSTR)
	     (EQUAL INSTR (QUOTE (S-PROP NIL))))
	 T)
	((SETQ BADHINTS
	       (ITERATE FOR X IN (CDR INSTR)
			WHEN
			(COND
			 ((CONSP X)
			  (NOT (GET (FN-SYMB (TRANSLATE-AND-CHK-ABB X ABBREVIATIONS))
				    (QUOTE SDEFN))))
			 (T
			  (OR (NOT (SYMBOLP X))
			      (NOT (GET X (QUOTE SDEFN)))))) 
			COLLECT X))
	 (BLURB (BADHINTS) |**NO| |CHANGE**| |--| |The| |arguments| |to| |S-PROP|
		|must| |either| |be| |simply| | NIL | |or| |else| |must| |all| |be|
		|defined| |function| |symbols| |or| |terms| |with| |defined|
		|top-level| |function| |symbols| |.| |The|
		(PLURAL? BADHINTS |arguments| |argument|) (!PPR-LIST BADHINTS NIL)
		(PLURAL? BADHINTS |do| |does|) |not| |have| |these| |properties| |.| CR))
	(T T))))

(DEFUN CHANGE-STATE-WITH-S-PROP (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-S-PROP-CHK INSTR ABBREVIATIONS)
       (LET (NEWTERM TRUE-TERMS FALSE-TERMS
	     (FN-LIST
	       (COND
		 ((ATOM INSTR)
		  (QUOTE (IMPLIES AND OR NOT IFF)))
		 ((EQUAL INSTR (QUOTE (S-PROP NIL)))
		  NIL)
		 (T
		  (ITERATE FOR X IN (CDR INSTR)
			WHEN (NOT (CONSP X))
			  COLLECT X))))
	     (EXPAND-LIST
	       (COND
		 ((ATOM INSTR)
		  NIL)
		 (T
		  (ITERATE FOR X IN (CDR INSTR)
			WHEN (CONSP X)
			  COLLECT (TRANSLATE-ABB X ABBREVIATIONS)))))
	     (IFF-FLG
	       (IF (EQ (TERM-ID-IFF CONC (REVERSE CURRENT-ADDR-R) (QUOTE IFF)) (QUOTE IFF))
		   T
		   NIL)))
	 (SETQ MY-ALL-LEMMAS-USED NIL)
	 (ITERATE FOR HYP-FLG IN (APPEND HYPS GOVERNORS)
	       WHEN (EQ (CDR HYP-FLG) (QUOTE A))
	       WITH HYP
	       DO
	   (COND
	     ((MATCH (SETQ HYP (CAR HYP-FLG)) (NOT HYP))
	      (SETQ FALSE-TERMS (CONS HYP FALSE-TERMS)))
	     (T
	      (SETQ TRUE-TERMS (CONS HYP TRUE-TERMS)))))
	 (SETQ NEWTERM (NORMALIZE-IFS
			 (EXPAND-FNS CURRENT-TERM FN-LIST EXPAND-LIST)
			 TRUE-TERMS FALSE-TERMS IFF-FLG))
	 (COND
	   ((EQUAL CURRENT-TERM NEWTERM)
	    (BLURB NIL |**NO| |CHANGE**| CR))
	   (T
	    ; The following line is there because of weakness of NORMALIZE-IFS
	    (setq newterm
		  (normalize-ifs newterm true-terms false-terms IFF-FLG))
	    (MAKE-NEW-RECORD
	      (QUOTE STATE)
	      (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		    (CONS (QUOTE CURRENT-TERM) NEWTERM)
		    (CONS (QUOTE GOAL)
			  (MAKE-NEW-RECORD
			    (QUOTE GOAL)
			    (LIST (CONS (QUOTE CONC)
					(DEPOSIT-TERM-R CONC CURRENT-ADDR-R NEWTERM)))
			    GOAL))
		    (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
			  (UNION-EQ MY-ALL-LEMMAS-USED CUMULATIVE-LEMMAS-USED)))
	      STATE))))))))

(DEFUN CHANGE-STATE-WITH-SHOW-GOVS-CHK (INSTR GOVERNORS)
  (LET (BADINDICES)
    (COND
      ((NULL GOVERNORS)
       (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |no| |governors| |,|
	      |active| |or| |hidden| |.| CR))
      ((ITERATE FOR HYP-FLG IN GOVERNORS
	     NEVER (EQ (CDR HYP-FLG) (QUOTE H)))
       (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |no| |hidden|
	      |governors| |.| CR))
      ((AND (CONSP INSTR)
	    (SETQ BADINDICES
		  (ITERATE FOR INDEX IN (CDR INSTR)
			WITH LENGTH = (LENGTH GOVERNORS)
			WHEN (OR (NOT (INTEGERP INDEX))
				 (< INDEX 1)
				 (> INDEX LENGTH)
				 (NEQ (CDR (NTH (1- INDEX) GOVERNORS)) (QUOTE H)))
			  COLLECT INDEX)))
       (BLURB (BADINDICES) |**NO| |CHANGE**| |--| |The| |following|
	      (PLURAL? BADINDICES |arguments| |argument|)
	      |to| SHOW-GOVS (PLURAL? BADINDICES |are| |is|) |not|
	      (PLURAL? BADINDICES |indices of| |an index of a|)
	      |hidden| (PLURAL? BADINDICES |governors| |governor|) |:|
	      (!PPR-LIST BADINDICES (QUOTE |.|)) CR))
      (T T))))

(DEFUN CHANGE-STATE-WITH-SHOW-GOVS (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-SHOW-GOVS-CHK INSTR GOVERNORS)
       (MAKE-NEW-RECORD
	 (QUOTE STATE)
	 (LIST (CONS (QUOTE INSTRUCTION) INSTR)
	       (CONS (QUOTE GOVERNORS)
		     (COND
		       ((ATOM INSTR)
			(ITERATE FOR HYP-FLG IN GOVERNORS
			      COLLECT (CONS (CAR HYP-FLG) (QUOTE A))))
		       (T
			(ITERATE FOR HYP-FLG IN GOVERNORS
			      AS I FROM 1
			      COLLECT
				(COND
				  ((MEMBER-EQUAL I (CDR INSTR))
				   (CONS (CAR HYP-FLG) (QUOTE A)))
				  (T HYP-FLG)))))))
	 STATE)))))

(DEFUN CHANGE-STATE-WITH-SHOW-HYPS-CHK (INSTR HYPS)
  (LET (BADINDICES)
    (COND
      ((NULL HYPS)
       (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |no| |top-level|
	      |hypotheses| |,| |active| |or| |hidden| |.| CR))
      ((ITERATE FOR HYP-FLG IN HYPS
	     NEVER (EQ (CDR HYP-FLG) (QUOTE H)))
       (BLURB NIL |**NO| |CHANGE**| |--| |There| |are| |no| |hidden|
	      |top-level| |hypotheses| |.| CR))
      ((AND (CONSP INSTR)
	    (SETQ BADINDICES
		  (ITERATE FOR INDEX IN (CDR INSTR)
			WITH LENGTH = (LENGTH HYPS)
			WHEN (OR (NOT (INTEGERP INDEX))
				 (< INDEX 1)
				 (> INDEX LENGTH)
				 (NEQ (CDR (NTH (1- INDEX) HYPS)) (QUOTE H)))
			  COLLECT INDEX)))
       (BLURB (BADINDICES) |**NO| |CHANGE**| |--| |The| |following|
	      (PLURAL? BADINDICES |arguments| |argument|)
	      |to| SHOW-HYPS (PLURAL? BADINDICES |are| |is|) |not|
	      (PLURAL? BADINDICES |indices of| |an index of a|)
	      |hidden| |top-level| (PLURAL? BADINDICES |hypotheses| |hypothesis|) |:|
	      (!PPR-LIST BADINDICES (QUOTE |.|)) CR))
      (T T))))

(DEFUN CHANGE-STATE-WITH-SHOW-HYPS (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-SHOW-HYPS-CHK INSTR HYPS)
       (MAKE-NEW-RECORD
	 (QUOTE STATE)
	 (LIST (CONS (QUOTE INSTRUCTION) INSTR)
	       (CONS (QUOTE GOAL)
		     (MAKE-NEW-RECORD
		       (QUOTE GOAL)
		       (LIST (CONS (QUOTE HYPS)
				   (COND
				     ((ATOM INSTR)
				      (ITERATE FOR HYP-FLG IN HYPS
					    COLLECT (CONS (CAR HYP-FLG) (QUOTE A))))
				     (T
				      (ITERATE FOR HYP-FLG IN HYPS
					    AS I FROM 1
					    COLLECT
					      (COND
						((MEMBER-EQUAL I (CDR INSTR))
						 (CONS (CAR HYP-FLG) (QUOTE A)))
						(T HYP-FLG)))))))
		       GOAL)))
	 STATE)))))

(DEFUN CHANGE-STATE-WITH-SPLIT (INSTR STATE)
  (LET-FIELDS
    (COND
      (CURRENT-ADDR-R
       (BLURB NIL |**NO| |CHANGE**| |--| |You| |must| |be| |at| |the| |top|
	      |of| |the| |goal| |to| SPLIT |.| CR))
      (T
       (LET ((CLAUSES
	       (CNF (MAKE-IMPLICATION (ASSUMPTIONS HYPS) CONC)))
	     NEW-GOAL-NAMES)	 
	 (COND
	   ((NULL CLAUSES)
	    ;; was disjunction of (NULL CLAUSES) and
	    ;; (NULL (CNF (MAKE-IMPLICATION
	    ;;             (FLESH-OUT-ASSUMPTIONS (ASSUMPTIONS HYPS))
	    ;;             CONC)))
	    (BLURB NIL |The| |current| |goal| |is| |a| |tautology| EXCL CR)
	    (MAKE-NEW-RECORD
	      (QUOTE STATE)
	      (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		    (CONS (QUOTE CURRENT-TERM) TRUE)
		    (CONS (QUOTE GOAL)
			  (MAKE-NEW-RECORD
			    (QUOTE GOAL)
			    (LIST (CONS (QUOTE CONC) TRUE))
			    GOAL)))
	      STATE))
	   ((AND (NULL (CDR CLAUSES))
		 (EQUAL CONC
			(CAR (OUR-LAST (CAR CLAUSES))))
		 (INT= (LENGTH HYPS)
		    (1- (LENGTH (CAR CLAUSES)))))
	    (BLURB NIL |**NO| |CHANGE**| |--| |No| |case| |splitting| |has|
		   |occurred| |,| |as| |only| |one| |subgoal| |was| |generated|
		   |and| |it| |had| |the| |same| |conclusion| |and| |number|
		   |of| |hypotheses| |as| |the| |current| |goal| |.| CR))
	   (T
	    (SETQ NEW-GOAL-NAMES
		  (NEW-GOAL-NAMES GOAL-NAME OTHER-GOALS (LENGTH CLAUSES)))
	    (MAKE-NEW-RECORD
	      (QUOTE STATE)
	      (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		    (CONS (QUOTE CURRENT-TERM) TRUE)
		    (CONS (QUOTE GOAL)
			  (MAKE-NEW-RECORD
			    (QUOTE GOAL)
			    (LIST (CONS (QUOTE CONC) TRUE)
				  (CONS (QUOTE DEPENDS-ON)
					(APPEND NEW-GOAL-NAMES DEPENDS-ON)))
			    GOAL))
		    (CONS (QUOTE OTHER-GOALS)
			  (UPDATE-OTHER-GOALS
			    OTHER-GOALS GOAL-NAME
			    (ITERATE FOR CL IN CLAUSES
				  COLLECT
				    (CONS (CAR (OUR-LAST CL))
					  (ITERATE FOR TAIL ON CL
						WHILE (CDR TAIL)
						COLLECT (DUMB-NEGATE-LIT (CAR TAIL)))))
			    NEW-GOAL-NAMES)))
	      STATE))))))))

(DEFUN CHANGE-STATE-WITH-SUBV-CHK (INSTR ASSUMPTIONS current-addr-r ABBREVIATIONS)
  (LET (BADPAIRS)
    (COND
      (current-addr-r
       (BLURB nil |**NO| |CHANGE**| |--| SUBV |may| |only| |be| |used| |when|
	      |at| |the| |top| |of| |the| |conclusion| |.| |Try| |TOP| |first| |.| CR))
      ((ATOM INSTR)
       T)
      ((SETQ BADPAIRS
	     (ITERATE FOR PAIR IN (CDR INSTR)
		   WHEN (NOT (AND (MATCH PAIR (LIST & &))
				  (VARIABLEP (TRANSLATE (CAR PAIR)))
				  (TRANSLATE-AND-CHK-ABB (CADR PAIR) ABBREVIATIONS)))
		     COLLECT PAIR))
       (BLURB (BADPAIRS) |**NO| |CHANGE**| |--| |The| |following| |substitution|
	      (PLURAL? BADPAIRS |pairs are| |pair is|) |not| |in| |the| |correct|
	      |form| |:| (!PPR-LIST BADPAIRS (QUOTE |.|)) |A| |substitution| |pair|
	      |here| |must| |have| |the| |form| |(variable term)| |.| CR))
      ((SETQ BADPAIRS
	     (ITERATE FOR PAIR IN (CDR INSTR)
		      for x = (TRANSLATE-ABB (CADR PAIR) ABBREVIATIONS)
		   WHEN (NOT (or (MEMBER-EQUAL
				  (FCONS-TERM* (QUOTE EQUAL) (CAR PAIR) x)
				  ASSUMPTIONS)
				 (MEMBER-EQUAL
				  (FCONS-TERM* (QUOTE EQUAL) x (CAR PAIR))
				  ASSUMPTIONS)))
		     COLLECT PAIR))
       (BLURB (BADPAIRS) |**NO| |CHANGE**| |--| |The| |following|
	      (PLURAL? BADPAIRS |arguments are| |argument is|) |invalid| |because|
	      |the| (PLURAL? BADPAIRS |equalities| |equality|) |of| |the| |components|
	      (PLURAL? BADPAIRS |are| |is|)
	      |not| |among| |the|
	      |active| |hypotheses| |:| (!ppr-list badpairs (quote |.|)) CR))
      ((no-duplicatesp (mapcar #'car (cdr instr)))
       t)
      (t (BLURB () |**NO| |CHANGE**| |--| |There| |are| |duplicate| |variables| |in|
		|your| |substitution| |.| CR)))))

(DEFUN CHANGE-STATE-WITH-SUBV (INSTR STATE)
  (LET-FIELDS
    (LET ((ASSUMPTIONS (ASSUMPTIONS HYPS))
	  SUB-PAIRS NEWTERM)
      (COND
	((and (CHANGE-STATE-WITH-SUBV-CHK INSTR ASSUMPTIONS current-addr-r ABBREVIATIONS)
	      (or (consp instr)
		  (setq sub-pairs
			(all-subv-pairs assumptions))
		  (blurb () |**NO| |CHANGE**| |--| |There| |are|
			 |no| |allowable| |variable| |substitutions| |.| CR)))
	 (if (consp instr)
	     (setq sub-pairs
		   (iterate for pair in (cdr instr)
			    collect (cons (car pair)
					  (translate-abb (cadr pair) abbreviations))))
	   (blurb ((x (iterate for x in sub-pairs
			       collect (list (car x)
					     (untranslate-abb (cdr x) abbreviations)))))
		  |Using| |the| |substitution| (!ppr x (quote |.|)) cr))
	 (SETQ NEWTERM (SUBLIS-VAR SUB-PAIRS conc))
	 (MAKE-NEW-RECORD
	  (QUOTE STATE)
	  (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		(CONS (QUOTE CURRENT-TERM) NEWTERM)
		(CONS (QUOTE GOAL)
		      (MAKE-NEW-RECORD
		       (QUOTE GOAL)
		       (LIST (CONS (QUOTE CONC) NEWTERM)
			     (CONS (QUOTE HYPS)
				   (ITERATE FOR HYP-FLG IN HYPS 
					    WITH X AND Y
					    COLLECT
					    (COND
					     ((EQ (CDR HYP-FLG) (QUOTE H))
					      HYP-FLG)
					     ((and (match (car hyp-flg) (equal x y))
						   (or (member-equal (cons x y) sub-pairs)
						       (member-equal (cons y x) sub-pairs)))
					      (CONS (CAR HYP-FLG) (QUOTE H)))
					     (T
					      (CONS (SUBLIS-VAR SUB-PAIRS (CAR HYP-FLG))
						    (QUOTE A)))))))
		       GOAL)))
	  STATE))))))

(DEFUN CHANGE-STATE-WITH-TOP (INSTR STATE)
  (LET-FIELDS
    (COND
      (CURRENT-ADDR-R
       (MAKE-NEW-RECORD
	 (QUOTE STATE)
	 (LIST (CONS (QUOTE INSTRUCTION) INSTR)
	       (CONS (QUOTE CURRENT-TERM) CONC)
	       (CONS (QUOTE GOVERNORS) NIL)
	       (CONS (QUOTE CURRENT-ADDR-R) NIL))
	 STATE))
      (T
       (BLURB NIL |**NO| |CHANGE**| |--| |Already| |at| |the| |top| |.| CR)))))

(DEFUN CHANGE-STATE-WITH-UP (INSTR STATE)
  (LET-FIELDS
    (LET ((TEMP
	    (FETCH-TERM-AND-CL CONC (REVERSE (CDR CURRENT-ADDR-R))
			       NIL
			       (ITERATE FOR PAIR IN GOVERNORS
				     COLLECT (CDR PAIR)))))
      (COND
	(CURRENT-ADDR-R
	 (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE CURRENT-TERM) (CAR TEMP))
		 (CONS (QUOTE GOVERNORS) (CDR TEMP))
		 (CONS (QUOTE CURRENT-ADDR-R) (CDR CURRENT-ADDR-R)))
	   STATE))
	(T
	 (BLURB NIL |**NO| |CHANGE**| |--| |Already| |at| |the| |top| |of| |the|
		|goal's| |conclusion| |.| CR))))))

(DEFUN CHANGE-STATE-WITH-USE-LEMMA-CHK (INSTR ABBREVIATIONS)
  (COND
   ((AND (CDDR INSTR)
	 (NOT (SUBSTP (CADDR INSTR) ABBREVIATIONS)))
    (BLURB ((SUB (CADDR INSTR))) |**NO| |CHANGE**| |--| |The| |second|
	   |argument| |to| USE-LEMMA |must| |be| |a| |list| |of| |pairs| |of|
	   |the| |form| |(var term)| |,| |where| | var | |is| |a| |variable|
	   |and| | term | |is| |a| |term| |(relative| |to| |the| |current|
	   |abbreviations)| |.| |However| |,| |the| |given| |substitution| |,|
	   (!PPR SUB (QUOTE |,|)) |is| |not| |in| |this| |form| |.| CR))
   ((NOT (AND (SYMBOLP (CADR INSTR))
	      (FORMULA-OF (CADR INSTR))))
    (BLURB ((NAME (CADR INSTR))) |**NO| |CHANGE**| |--| |There| |is| |no|
	   |usable| |formula| |in| |the| |current| |database| |associated|
	   |with| |the| |name| (!PPR NAME (QUOTE |.|)) CR))
   (T T)))

(DEFUN CHANGE-STATE-WITH-USE-LEMMA (INSTR STATE)
  (LET-FIELDS
   (COND
    ((CHANGE-STATE-WITH-USE-LEMMA-CHK INSTR ABBREVIATIONS)
     (LET ((FORMULA (FORMULA-OF (CADR INSTR)))
	   (SUBST
	    (IF (CDDR INSTR)
		(ITERATE FOR X IN (CADDR INSTR)
			 COLLECT (CONS (CAR X)
				       (TRANSLATE-ABB (CADR X) ABBREVIATIONS)))
		NIL))
	   NEW-HYP NEW-FREE-VARS)
	  (COND
	   ((CDDR INSTR)
	    (BLURB ((NAME (CADR INSTR))
		    (SUB (ITERATE FOR X IN SUBST
				  COLLECT (LIST (CAR X)
						(UNTRANSLATE-ABB
						 (CDR X) ABBREVIATIONS)))))
		   |Using| (!PPR NAME NIL)
		   |with| |the| |substitution| (!PPR SUB (QUOTE |.|)) CR))
	   (T
	    (BLURB ((NAME (CADR INSTR))) |Using|
		   (!PPR NAME (QUOTE |.|)) CR)))
	  (SETQ NEW-HYP (SUBLIS-VAR SUBST FORMULA))
	  (SETQ NEW-FREE-VARS
		(SET-DIFF-EQ (ALL-VARS NEW-HYP)
			     (ALL-VARS-STATE GOAL OTHER-GOALS)))
	  (WHEN NEW-FREE-VARS
		(BLURB (NEW-FREE-VARS) CR NOTE |:| |The|
		       (PLURAL? NEW-FREE-VARS |variables| |variable|)
		       (!PPR-LIST NEW-FREE-VARS NIL)
		       (PLURAL? NEW-FREE-VARS |are| |is|) |being| |introduced|
		       |into| |the| |proof| |state| |and|
		       (PLURAL? NEW-FREE-VARS |are| |is|)
		       |therefore| |being| |added| |to| |the|
		       |current| |list| |of| |free| |variables| |.| CR))
	  (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE GOAL)
		       (MAKE-NEW-RECORD
			(QUOTE GOAL)
			(LIST (CONS (QUOTE HYPS)
				    (APPEND HYPS
					    (LIST (CONS NEW-HYP (QUOTE A))))))
			GOAL))
		 (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		       (UNION-EQ
			(ALL-FNNAMES-LST
			 (ITERATE FOR PAIR IN SUBST COLLECT (CDR PAIR)))
			(ADD-TO-SET (CADR INSTR) CUMULATIVE-LEMMAS-USED)))
		 (CONS (QUOTE FREE-VARIABLES)
		       (UNION-EQ FREE-VARIABLES NEW-FREE-VARS)))
	   STATE))))))

(DEFUN CHANGE-STATE-WITH-X-CHK (INSTR CURRENT-TERM HYPS-TYPE-ALIST CURRENT-ADDR-R HYPS)
  (COND
   ((AND (CONSP INSTR)
	 (NOT (EQ (CADR INSTR) (QUOTE ALL)))
	 (NOT (EQ (CADR INSTR) (QUOTE LEMMAS)))
	 (NOT (AND (INTEGERP (CADR INSTR))
		   (>= (CADR INSTR) 0))))
    (BLURB NIL |**NO| |CHANGE**| |--| |The| |only| |arguments| |allowed|
	   |for| X |are| ALL |,| LEMMAS |,| |and| |a| |positive| |integer| |.| CR))
   ((AND (EQ HYPS-TYPE-ALIST (QUOTE CONTRADICTION))
	 CURRENT-ADDR-R
	 (NEQ (HYPS-TYPE-ALIST HYPS) (QUOTE CONTRADICTION)))
    (BLURB NIL |**NO| |CHANGE**| |--| |You| |have| |a| |contradiction| |in|
	   |the| |active| |hypotheses| |and| |governors| |but| |not| |in| |the|
	   |hypotheses| |alone| |,| |so| |you| |may| |wish| |to| |move| |up|
	   |in| |the| |term| |as| |far| |as| |you| |can| |while| |preserving|
	   |this| |property| |and| |then| |use| |(= * new-term T)| |to| |replace|
	   |that| |term| |with| |your| |favorite| |new-term| |while| |using| S
	   |to| |prove| |the| |new| |goal| |.| |Or| |,| |use| X-DUMB |here| |.| CR))
   ((VARIABLEP CURRENT-TERM)
    (BLURB (CURRENT-TERM) |**NO| |CHANGE**| |--| |The| |current| |subterm| |,|
	   (!PPR CURRENT-TERM (QUOTE |,|)) |is| |a| |variable| |and| |thus|
	   |no| |expansion| |is| |possible| |.| CR))
   ((NOT (GET (FN-SYMB CURRENT-TERM) (QUOTE SDEFN)))
    (BLURB ((FN (FN-SYMB CURRENT-TERM))) |**NO| |CHANGE**| |--| |The| |current|
	   |term's| |function| |symbol| |,| (!PPR FN (QUOTE |,|)) |is| |not| |a|
	   |defined| |function| |symbol| |.| CR))
   (T T)))

(DEFUN CHANGE-STATE-WITH-X (INSTR STATE)
  (LET-FIELDS
    (LET ((HYPS-TYPE-ALIST
	    (HYPS-TYPE-ALIST (APPEND HYPS GOVERNORS)))
	  TEMP)
      (COND
	((CHANGE-STATE-WITH-X-CHK INSTR CURRENT-TERM HYPS-TYPE-ALIST CURRENT-ADDR-R HYPS)
	 (COND
	   ((EQ HYPS-TYPE-ALIST (QUOTE CONTRADICTION))
	    ;so the top-level hypotheses alone have a contradiction, by
	    ;definition of CHANGE-STATE-WITH-X-CHK
	    (BLURB NIL |Contradiction| |in| |the| |hypotheses|
		   |of| |the| |current| |goal| EXCL CR)
	    (MAKE-NEW-RECORD
	      (QUOTE STATE)
	      (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		    (CONS (QUOTE CURRENT-TERM) TRUE)
		    (CONS (QUOTE GOVERNORS) NIL)
		    (CONS (QUOTE CURRENT-ADDR-R) NIL)
		    (CONS (QUOTE GOAL)
			  (MAKE-NEW-RECORD
			    (QUOTE GOAL)
			    (LIST (CONS (QUOTE CONC) TRUE))
			    GOAL)))
	      STATE))
	   (T
	    (SETQ MY-ALL-LEMMAS-USED NIL) 
	    (SETQ TEMP
		  (LET ((FNS-ALLOWED-FOR-FAST-REWRITE-FLG
			 (OR (EQ INSTR 'X)
			     (EQUAL INSTR '(X ALL))))
			(LEVEL (IF (AND (CONSP INSTR) (INTEGERP (CADR INSTR)))
				   (CADR INSTR)
				   0)))
		       (EXPAND-FNCALL CURRENT-TERM HYPS-TYPE-ALIST
				      (TERM-ID-IFF CONC (REVERSE CURRENT-ADDR-R) (QUOTE IFF))
				      (COND
				       ((ATOM INSTR)
					FREWRITE-RULES)
				       (T
					ALL-FREWRITE-RULES))
				      FREWRITE-DISABLED-RULES
				      LEVEL)))
	    (COND
	      ((EQUAL TEMP CURRENT-TERM)
	       (BLURB NIL |**NO| |CHANGE**| CR))
	      (T
	       (MAKE-NEW-RECORD
		 (QUOTE STATE)
		 (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		       (CONS (QUOTE CURRENT-TERM) TEMP)
		       (CONS (QUOTE GOAL)
			     (MAKE-NEW-RECORD
			       (QUOTE GOAL)
			       (LIST (CONS (QUOTE CONC)
					   (DEPOSIT-TERM-R CONC CURRENT-ADDR-R TEMP)))
			       GOAL))
		       (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
			     (UNION-EQ MY-ALL-LEMMAS-USED CUMULATIVE-LEMMAS-USED)))
		 STATE))))))))))

(DEFUN CHANGE-STATE-WITH-X-DUMB-CHK (CURRENT-TERM)
  (COND
    ((VARIABLEP CURRENT-TERM)
     (BLURB (CURRENT-TERM) |**NO| |CHANGE**| |--| |The| |current| |subterm| |,|
	    (!PPR CURRENT-TERM (QUOTE |,|)) |is| |a| |variable| |and| |thus|
	    |no| |expansion| |is| |possible| |.| CR))
    ((NOT (GET (FN-SYMB CURRENT-TERM) (QUOTE SDEFN)))
     (BLURB ((FN (FN-SYMB CURRENT-TERM))) |**NO| |CHANGE**| |--| |The| |current|
	    |term's| |function| |symbol| |,| (!PPR FN (QUOTE |,|)) |is| |not| |a|
	    |defined| |function| |symbol| |.| CR))
    (T T)))

(DEFUN CHANGE-STATE-WITH-X-DUMB (INSTR STATE)
  (LET-FIELDS
    (COND
      ((CHANGE-STATE-WITH-X-DUMB-CHK CURRENT-TERM)
       (LET ((NEWTERM (EXPAND-FNCALL-TAUT CURRENT-TERM)))
	 (MAKE-NEW-RECORD
	   (QUOTE STATE)
	   (LIST (CONS (QUOTE INSTRUCTION) INSTR)
		 (CONS (QUOTE CURRENT-TERM) NEWTERM)
		 (CONS (QUOTE GOAL)
		       (MAKE-NEW-RECORD
			 (QUOTE GOAL)
			 (LIST (CONS (QUOTE CONC)
				     (DEPOSIT-TERM-R CONC CURRENT-ADDR-R NEWTERM)))
			 GOAL))
		 (CONS (QUOTE CUMULATIVE-LEMMAS-USED)
		       (ADD-TO-SET (FN-SYMB CURRENT-TERM)
				   CUMULATIVE-LEMMAS-USED)))
	   STATE))))))

;;************************************************************;;
;;              STATE-STACK CHANGING FUNCTIONS                ;;
;;************************************************************;;

(DEFUN CHANGE-STATE-STACK-WITH-BIND (INSTR)
  (IF (ATOM INSTR)
      (ER SOFT NIL |The| |command| BIND |is| |not| |allowed| |zero|
	  |arguments| |.|)
      (LET ((COMMAND-BINDINGS
	      (UPDATE-COMMAND-BINDINGS (CADR INSTR) COMMAND-BINDINGS)))
	(CHANGE-STATE-STACK-WITH-DO-ALL (CONS (QUOTE DO-ALL) (CDDR INSTR))))))

(DEFUN CHANGE-STATE-STACK-WITH-DO-ALL (INSTRS)
  (ITERATE FOR INSTR IN (SAFECDR INSTRS)
	WITH FLG = T
	DO
    (SETQ FLG (AND (INTERACTIVE-SINGLE-STEP INSTR) FLG))
	FINALLY (RETURN FLG)))

(DEFUN CHANGE-STATE-STACK-WITH-LISP (INSTR)
  (EVAL-LOCAL (CADR INSTR) COMMAND-BINDINGS))

(DEFUN CHANGE-STATE-STACK-WITH-NEGATE (X)
  (NULL (INTERACTIVE-SINGLE-STEP (CADR X))))

(DEFUN CHANGE-STATE-STACK-WITH-PROG2 (X)
  (COND
    ((INTERACTIVE-SINGLE-STEP (CADR X))
     (INTERACTIVE-SINGLE-STEP (CADDR X)))))

(DEFUN CHANGE-STATE-STACK-WITH-PROTECT (X)
  (LET ((SAVED-SS STATE-STACK) (SAVED-OLD-SS OLD-SS))
    (OR
      (INTERACTIVE-SINGLE-STEP (CADR X))
      (PROGN (SETQ STATE-STACK SAVED-SS)
	     (SETQ OLD-SS SAVED-OLD-SS)
	     NIL))))

(DEFUN CHANGE-STATE-STACK-WITH-RESTORE (INSTR)
  INSTR				; for the compiler
  (LET ((TEMP-SS STATE-STACK))
    (COND
      ((NULL OLD-SS)
       (BLURB NIL |**NO| |CHANGE**| |--| |nothing| |to| |restore| |.| CR))
      (T
       (SETQ STATE-STACK OLD-SS)
       (SETQ OLD-SS TEMP-SS)))))

(DEFUN CHANGE-STATE-STACK-WITH-SUCCEED (INSTR)
  (PROGN (INTERACTIVE-SINGLE-STEP (CADR INSTR)) T))

(DEFUN CHANGE-STATE-STACK-WITH-UNDO (INSTR)
  (LET (X START-MARKER LAST-INSTR)
    (COND
      ((EQ INSTR (QUOTE UNDO))
       (COND
	 ((NULL (CDR STATE-STACK))
	  (BLURB NIL |**NO| |CHANGE**| |--| |Already| |at| |the| |start| |.| CR))
	 ((MATCH (ACCESS STATE INSTRUCTION (CAR STATE-STACK))
		 (START START-MARKER))
	  (BLURB (START-MARKER) WARNING |:| |Now| |undoing| |back| |past| |the|
		 |start| |of| |the| |current| |session| |,| |which| |is| |marked|
		 |by| (!PPR START-MARKER (QUOTE |.|)) |Submit| RESTORE |if|
		 |that| |isn't| |what| |you| |intended| |.| CR)
	  (SETQ OLD-SS STATE-STACK)
	  (SETQ STATE-STACK (CDR STATE-STACK)))
	 (T
	  (BLURB ((INST (ACCESS STATE INSTRUCTION (CAR STATE-STACK))))
		 |Undoing| |:| (!PPR INST NIL) CR)
	  (SETQ OLD-SS STATE-STACK)
	  (SETQ STATE-STACK (CDR STATE-STACK)))))
      ((AND (MATCH INSTR (UNDO X))
	    (INTEGERP X))
       (COND
	 ((NOT (> X 0))
	  (BLURB NIL |**NO| |CHANGE**| |--| |A| |numeric| |argument| |to|
		 UNDO |must| |be| |positive| |.| CR))
	 ((NULL (CDR STATE-STACK))
	  (BLURB NIL |**NO| |CHANGE**| |--| |Already| |at| |the| |start| |.| CR))
	 ((MATCH (ACCESS STATE INSTRUCTION (CAR STATE-STACK))
		 (START START-MARKER))
	  (BLURB (START-MARKER) |**NO| |CHANGE**| |--| |Already| |at| |the|
		 |start| |of| |the| |current| |session| |,| |which| |is|
		 |marked| |by| (!PPR START-MARKER NIL) |.|
		 |Type| UNDO |to| |get| |past| |this| |barrier| |.| CR))
	 (T
	  (SETQ OLD-SS STATE-STACK)
	  (ITERATE FOR I FROM 1 TO X WITH NEXT-INSTR
		WHILE (AND	;notice this condition IS met when I = 1
			(CDR STATE-STACK)
			(NOT (MATCH (SETQ NEXT-INSTR
					  (ACCESS STATE INSTRUCTION (CAR STATE-STACK)))
				    (START START-MARKER))))
		DO (PROGN (SETQ STATE-STACK (CDR STATE-STACK))
			  (SETQ LAST-INSTR NEXT-INSTR)))
	  (BLURB (LAST-INSTR) |Last| |instruction| |undone| |:|
		 (!PPR LAST-INSTR (QUOTE |.|)) CR)
	  (COND
	    ((NULL (CDR STATE-STACK))
	     (BLURB NIL CR |Back| |to| |the| |start| |.| CR))
	    (START-MARKER
	     (BLURB (START-MARKER) CR |Cannot| |undo| |further| |back| |than|
		    |the| |start| |of| |the| |current| |session| |,| |which|
		    |is| |marked| |by| (!PPR START-MARKER (QUOTE |.|)) |Type|
		    UNDO |next| |if| |you| |wish| |to| |go| |past| |this|
		    |barrier| |.| NOTE |that| |instructions| |have| |been|
		    |undone| |until| |this| |barrier| |was| |reached--| |submit|
		    RESTORE |if| |that| |is| |not| |what| |was| |intended|
		    |.| CR)))
	  T)))
      ((MATCH INSTR (UNDO X))	;Find the index of the bookmark
       (ITERATE FOR STATE IN STATE-STACK AS I FROM 1
	     WITH FLG
	     UNTIL
	       (SETQ FLG
		     (EQUAL (ACCESS STATE INSTRUCTION STATE)
			    (LIST (QUOTE BOOKMARK) X)))
	     FINALLY
	       (RETURN
		 (COND
		   (FLG
		    (COND
		      ((INT= I 1)
		       (BLURB (X) |**NO| |CHANGE**| |--| |The| |bookmark|
			      (!PPR X NIL) |is| |already| |the| |previous|
			      |command| |.| |Type| UNDO |to| |undo|
			      |past| |this| |bookmark| |.| CR))
		      (T
		       (CHANGE-STATE-STACK-WITH-UNDO `(UNDO ,(1- I))))))
		   (T
		    (BLURB (X) |**NO| |CHANGE**| |--| |The| |bookmark|
			   (!PPR X NIL) |was| |not| |found| |.| CR)))))))))

;;************************************************************;;
;;                      HELP FUNCTIONS                        ;;
;;************************************************************;;

(DEFUN PRINT-HELP-WITH-COMMANDS (INSTR)
  (LET (N OVERFLOW-FLG LNG)
    (COND
      ((ATOM INSTR)
       (SETQ N (LENGTH STATE-STACK)))
      ((AND (INTEGERP (CADR INSTR))
	    (> (CADR INSTR) 0))
       (SETQ N
	     (IF (SETQ OVERFLOW-FLG (> (CADR INSTR) (SETQ LNG (LENGTH STATE-STACK))))
		 LNG
		 (CADR INSTR))))
      (T
       (BLURB NIL |The| |argument| |to| COMMANDS |must| |be|
	      |a| |positive| |integer| |.| CR)))
    (COND
      (N
       (COND
	 ((ATOM INSTR)
	  (BLURB NIL |The| |commands| |thus| |far| |(in| |reverse|
		 |order| |,| |i.e.| |last| |one| |first)| |have| |been| |:| CR))
	 (T
	  (BLURB (N)
	      (COND
		((= N 1)
		 |The| |previous| |command| |was|)
		(T
		 |The| |last| (!PPR N NIL) |commands| |,| |in| |reverse| |order|
		 |(i.e.| |last| |one| |first)| |,| |were|))
	      |:| CR)))
       (ITERATE FOR S IN STATE-STACK AS I FROM 1
	     UNTIL (> I N)
	     DO
	 (BLURB (I (INSTR (ACCESS STATE INSTRUCTION S)))
		(!PPR I (QUOTE |.|))
		(!PPR INSTR NIL) CR)
	     FINALLY
	       (COND (OVERFLOW-FLG
		      (BLURB NIL |There| |are| |no| |more| |commands| |.| CR))))))))

(DEFUN PRINT-HELP-WITH-GOALS (INSTR)
  (LET ((STATE (CAR STATE-STACK)))
       (LET-FIELDS
	GOAL-NAME DEPENDS-ON CONC	;for the compiler
	(LET ((ALL-GOALS (ALL-GOALS STATE)))
	     (COND
	      ((ATOM INSTR)
	       (ITERATE FOR GL IN ALL-GOALS
			WHEN (NOT (EQUAL (ACCESS GOAL CONC GL) TRUE))
			DO (PROGN (PRINC (ACCESS GOAL GOAL-NAME GL))
				  (TERPRI NIL))))
	      ((EQ (CADR INSTR) 'TREE)
	       (PRINT-GOAL-NAMES-AS-TREE
		(FIND-GOAL 'MAIN ALL-GOALS)
		(ITERATE FOR GL IN ALL-GOALS
			 WHEN (NOT (EQUAL (ACCESS GOAL CONC GL) TRUE))
			 COLLECT GL)
		0))
	      ((EQ (CADR INSTR) 'ALL)
	       (PRINT-GOAL-NAMES-AS-TREE (FIND-GOAL 'MAIN ALL-GOALS) ALL-GOALS 0))
	      (T
	       (BLURB NIL |Sorry| |--| |The| |only| |legal| |keywords| |for|
		      |the| GOALS |command| |are| TREE |and| ALL |.| CR)))))))

(DEFUN PRINT-HELP-WITH-HELP (HELP-INSTRS)
  (COND
   ((CONSP HELP-INSTRS)			;and therefore (CONSP (CDR HELP-INSTRS))
    (ITERATE FOR INSTR IN (CDR HELP-INSTRS)
	     WITH MACRO-DEFN AND FLG AS I FROM 1
	     DO
	     (PROGN
	      (COND ((NULL FLG)
		     (SETQ FLG T))
		    (T (IPRINC "----------------------------------------------------" NIL)
		       (BLURB NIL CR NIL)))
	      (SETQ MACRO-DEFN NIL)
	      (SETQ INSTR (INSTR-NAME (FIX-INSTR INSTR))) ;need INSTR-NAME e.g. for 17 
	      (COND
	       ((NOT (SYMBOLP INSTR))
		(BLURB (I) |***NOTE***| |Argument| |number| (!PPR I NIL)
		       |was| |not| |given| |as| |a| |symbol| |,| |so| |it| |has| |been|
		       |skipped| |.| CR))
	       ((OR (GET-INSTR-INFO INSTR)
		    (SETQ MACRO-DEFN (MACRO-COMMAND-DEF INSTR)))
		(COND
		 (MACRO-DEFN
		  (IPRINC "Macro Command" NIL)
		  (IPRINC "  [Use HELP-LONG to see the definition]" NIL)
		  (ITERPRI NIL)
		  (IPRINC (CONS (CADR MACRO-DEFN) (CADDR MACRO-DEFN)) NIL)
		  (ITERPRI NIL)
		  (EVAL `(BLURB ,@(MAKE-BLURB-BODY (GET INSTR '*PC-MACRO*-DOC)) CR))) 
		 (T (EVAL `(BLURB ,@(MAKE-BLURB-BODY (SHORT-HELP INSTR)) CR)))))
	       (T
		(BLURB (INSTR) (!PPR INSTR NIL) |is| |an| |invalid|
		       |instruction| |.| CR))))))
   (T
    (IPRINC
     "For more commands and more information, use HELP-LONG in place of HELP.
     Also try (HELP cmd1 ... cmdN) or (HELP-LONG cmd1 ... cmdN).
STATE CHANGE (PROOF) COMMANDS:
  Subterm manipulation commands:
	DIVE UP NX BK TOP
  Goal manipulation commands:
	CHANGE-GOAL HIDE-HYPS HIDE-GOVS SHOW-HYPS SHOW-GOVS CLAIM
	PROMOTE DEMOTE DROP CONTRADICT USE-LEMMA
  Goal creation commands:
	PUSH GENERALIZE INDUCT SPLIT
	(and sometimes CLAIM, =, REWRITE, and BASH)
  Simplification/replacement commands:
	X X-DUMB SUBV S S-PROP = PROVE REWRITE BASH
  Fast rewrite enable/disable commands:
        ENABLE DISABLE ENABLE-THEORY DISABLE-THEORY
  Free Variable commands:  PUT COPY (free variables also affectable by
                           REWRITE, USE-LEMMA, INDUCT, and GENERALIZE)
  Bookmark command:  BOOKMARK COMMENT
  Abbreviation commands:  ADD-ABBREVIATION REMOVE-ABBREVIATIONS
HELP COMMANDS:
        PP P PP-TOP GOALS HYPS SHOW-REWRITES SHOW-INDUCTIONS
        COMMANDS SHOW-ABBREVIATIONS UNDONE-INSTRUCTIONS
        HELP HELP-LONG SHOW-FREE-VARIABLES
META COMMANDS:
        UNDO RESTORE LISP DO-ALL
Exit command:  EXIT"
     NIL)
    (BLURB (basic-macro-command-names)
	   CR |Some| |basic| |macro| |commands| |:|
	   (cond ((not basic-macro-command-names)
		  |[None| |given| |at| |the| |moment]|)
		 (t (!LIST basic-macro-command-names) |.|))
	   CR))))

(DEFUN PRINT-HELP-WITH-HELP-LONG (HELP-INSTRS)
  (COND
    ((CONSP HELP-INSTRS)	;and therefore (CONSP (CDR HELP-INSTRS))
     (ITERATE FOR INSTR IN (CDR HELP-INSTRS)
	   WITH MACRO-DEFN AND FLG AS I FROM 1
	   DO 
       (PROGN
	 (COND ((NULL FLG)
		(SETQ FLG T))
	       (T (IPRINC "----------------------------------------------------" NIL)
		  (BLURB NIL CR NIL)))
	 (SETQ MACRO-DEFN NIL)
	 (SETQ INSTR (INSTR-NAME (FIX-INSTR INSTR)))	;need INSTR-NAME e.g. for 17
	 (COND
	  ((NOT (SYMBOLP INSTR))
	   (BLURB (I) |***NOTE***| |Argument| |number| (!PPR I NIL)
		  |was| |not| |given| |as| |a| |symbol| |,| |so| |it| |has| |been|
		  |skipped| |.| CR))
	  ((OR (GET-INSTR-INFO INSTR)
	       (SETQ MACRO-DEFN (MACRO-COMMAND-DEF INSTR)))
	   (COND
	    (MACRO-DEFN
	     (IPRINC "Macro Command:" NIL)
	     ;; **** The following was (BLURB (MACRO-DEFN) CR (!PPR MACRO-DEFN NIL) CR CR)
	     ;; but we violate our convention of using Boyer-Moore printing routines in order
	     ;; to get a printed version that can be read by the reader.
	     (pprint macro-defn) 
	     (iterpri nil)
	     (EVAL `(BLURB ,@(MAKE-BLURB-BODY (GET INSTR '*PC-MACRO*-DOC)) CR)))
	    (T (EVAL `(BLURB ,@(MAKE-BLURB-BODY (LONG-HELP INSTR)) CR)))))
	  (T
	   (BLURB (INSTR) (!PPR INSTR NIL) |is| |an| |invalid|
		  |instruction| |.| CR))))))
    (T
     (IPRINC "For help on individual commands, type (HELP-LONG cmd1 ... cmdN) (for
detailed help) or (HELP cmd1 ... cmdN) (for basic help, with examples).
STATE CHANGE (PROOF) COMMANDS:
  Subterm manipulation commands:
	DIVE UP NX BK TOP
  Goal manipulation commands:
	CHANGE-GOAL HIDE-HYPS HIDE-GOVS SHOW-HYPS SHOW-GOVS CLAIM
	PROMOTE DEMOTE DROP CONTRADICT USE-LEMMA
  Goal creation commands:
	PUSH GENERALIZE INDUCT SPLIT
	(and sometimes CLAIM, =, REWRITE, and BASH)
  Simplification/replacement commands:
	X X-DUMB SUBV S S-PROP = PROVE REWRITE BASH
  Fast rewrite enable/disable commands:
        ENABLE DISABLE ENABLE-THEORY DISABLE-THEORY
  Free Variable commands:  PUT COPY (free variables also affected by
                           REWRITE, USE-LEMMA, INDUCT, and GENERALIZE)
  Bookmark command:  BOOKMARK COMMENT
  Abbreviation commands:  ADD-ABBREVIATION REMOVE-ABBREVIATIONS
HELP COMMANDS:
        PP P PP-TOP GOALS HYPS SHOW-REWRITES SHOW-INDUCTIONS
        COMMANDS SHOW-ABBREVIATIONS UNDONE-INSTRUCTIONS
        HELP HELP-LONG SHOW-FREE-VARIABLES
META COMMANDS:
       	UNDO RESTORE LISP BIND DO-ALL PROG2 NEGATE PROTECT SUCCEED
Exit command:  EXIT"
		NIL)
     (BLURB (MACRO-COMMANDS) CR
	    |Macro| |commands| |:| (!LIST MACRO-COMMANDS) CR))))

(DEFUN PRINT-HELP-WITH-HYPS (INSTR)
  (LET ((STATE (CAR STATE-STACK)))
    (LET-FIELDS
      (COND
	((AND (CONSP INSTR)
	      (OR (AND (NOT (CONSP (CADR INSTR)))
		       (NOT (MEMBER-EQ (CADR INSTR) (QUOTE (NIL ALL))))
		       (PROGN (BLURB NIL |The| |first| |argument| |to| HYPS |must|
				     |be| |a| |list| |or| |the| |atom| ALL |.| CR)
			      T))
		  (AND (NOT (CONSP (CADDR INSTR)))
		       (NOT (MEMBER-EQ (CADDR INSTR) (QUOTE (NIL ALL))))
		       (PROGN (BLURB NIL |The| |second| |argument| |to| HYPS |must|
				     |be| |a| |list| |or| |the| |atom| ALL |.| CR)
			      T))
		  (AND (CONSP (CADR INSTR))
		       (ITERATE FOR X IN (CADR INSTR)
			     WITH HYPS-LNG = (LENGTH HYPS)
			     WHEN
			       (OR (NOT (INTEGERP X))
				   (< X 1)
				   (> X HYPS-LNG))
			       COLLECT
				 (PROGN (BLURB (X) |No| |hypothesis| |is| |numbered|
					       (!PPR X (QUOTE |.|)) CR)
					X)))
		  (AND (CONSP (CADDR INSTR))
		       (ITERATE FOR X IN (CADDR INSTR)
			     WITH GOVS-LNG = (LENGTH GOVERNORS)
			     WHEN
			       (OR (NOT (INTEGERP X))
				   (< X 1)
				   (> X GOVS-LNG))
			       COLLECT
				 (PROGN (BLURB (X) |No| |governor| |is| |numbered|
					       (!PPR X (QUOTE |.|)) CR)
					X))))))
	(T
	 (COND
	   ((ATOM INSTR)
	    (BLURB NIL |***| |Active| |top-level| |hypotheses| |:| CR))
	   (T
	    (BLURB NIL |***| |Top-level| |hypotheses| |(with| |"*"|
		   |when| |hidden)| |:| CR)))
	 (ITERATE FOR X IN HYPS AS INDEX FROM 1
	       WITH FLG
	       WHEN
		 (COND
		   ((ATOM INSTR)
		    (EQ (CDR X) (QUOTE A)))
		   ((EQ (CADR INSTR) (QUOTE ALL))
		    T)
		   (T
		    (MEMBER-EQUAL INDEX (CADR INSTR))))
	       WITH LABEL
	       DO (PROGN
		    (SETQ FLG T)
		    (SETQ LABEL (PACK (LIST
					(IF (EQ (CDR X) (QUOTE A))
					    (QUOTE H)
					    (QUOTE *H))
					INDEX (QUOTE |.  |))))
		    (IPRINC LABEL NIL)
		    (PPRIND-ABBREVIATIONS (CAR X) 5 0 NIL ABBREVIATIONS)
		    (ITERPRI NIL))
	       FINALLY (COND ((NULL FLG)
			      (BLURB NIL |There| |are| |no| |top-level|
				     |hypotheses| |to| |display| |.| CR))))
	 (COND
	   ((ATOM INSTR)
	    (BLURB NIL CR |***| |Active| |governors| |:| CR))
	   (T
	    (BLURB NIL CR |***| |Governors| |(with| |"*"|
		   |when| |hidden)| |:| CR)))
	 (ITERATE FOR X IN GOVERNORS AS INDEX FROM 1
	       WITH FLG
	       WHEN
		 (COND
		   ((ATOM INSTR)
		    (EQ (CDR X) (QUOTE A)))
		   ((EQ (CADDR INSTR) (QUOTE ALL))
		    T)
		   (T
		    (MEMBER-EQUAL INDEX (CADDR INSTR))))
	       WITH LABEL
	       DO (PROGN
		    (SETQ FLG T)
		    (SETQ LABEL (PACK (LIST
					(IF (EQ (CDR X) (QUOTE A))
					    (QUOTE G)
					    (QUOTE *G))
					INDEX (QUOTE |.  |))))
		    (BLURB (LABEL) (!PPR LABEL NIL))
		    (PPRIND-ABBREVIATIONS (CAR X) 5 0 NIL ABBREVIATIONS)
		    (ITERPRI NIL))
	       FINALLY (COND ((NULL FLG)
			      (BLURB NIL |There| |are| |no| |governors|
				     |to| |display| |.| CR)))))))))

(DEFUN PRINT-HELP-WITH-P (INSTR)
  (LET ((STATE (CAR STATE-STACK)))
    INSTR
    (LET-FIELDS
      (PPRIND-ABBREVIATIONS CURRENT-TERM 0 0 NIL ABBREVIATIONS)
      (ITERPRI NIL))))

(DEFUN PRINT-HELP-WITH-PP (INSTR)
  (LET ((STATE (CAR STATE-STACK)))
    INSTR
    (LET-FIELDS
      (PPRIND-WEAK? CURRENT-TERM 0 0 NIL ABBREVIATIONS)
      (ITERPRI NIL))))

(DEFUN PRINT-HELP-WITH-PP-TOP (INSTR)
  (LET ((STATE (CAR STATE-STACK)))
    INSTR
    (LET-FIELDS
      (HIGHLIGHT-PPRIND CONC 0 0 NIL CURRENT-ADDR-R CURRENT-TERM ABBREVIATIONS)
      (ITERPRI NIL))))

(DEFUN PRINT-HELP-WITH-SHOW-ABBREVIATIONS (INSTR)
  (LET ((ABBREVIATIONS (ACCESS STATE ABBREVIATIONS (CAR STATE-STACK))))
    (COND
      (ABBREVIATIONS
       (ITERATE FOR PAIR IN ABBREVIATIONS
	     DO
	 (COND
	   ((OR (ATOM INSTR)
		(MEMBER-EQ (CAR PAIR) (CDR INSTR)))
	    (BLURB ((VAR (CAR PAIR))
		    (TERM-ABB (UNTRANSLATE-ABB
				(CDR PAIR)
				(DEL-ASSQ (CAR PAIR) ABBREVIATIONS)))
		    (TERM (UNTRANSLATE (CDR PAIR))))
		   CR (!PPR VAR NIL) |corresponds| |to| CR |#|
		   (!PPR TERM-ABB NIL) CR
		   (COND ((NOT (EQUAL TERM TERM-ABB))
			  |#| | i.e.| |to| CR |#|
			  (!PPR TERM NIL) CR))))))
       (COND
	 ((CONSP INSTR)
	  (ITERATE FOR VAR IN (CDR INSTR)
		WHEN (NOT (ASSOC-EQ VAR ABBREVIATIONS))
		  DO (BLURB (VAR) CR NOTE |:| (!PPR VAR NIL) |is| |not| |an|
			    |abbreviation| |.| CR)))))
      (T
       (BLURB NIL CR |There| |are| |currently| |no| |abbreviations| |.| CR)))))

(DEFUN PRINT-HELP-WITH-SHOW-FREE-VARIABLES (INSTR)
  (LET ((FREE-VARIABLES (ACCESS STATE FREE-VARIABLES (CAR STATE-STACK))))
       (IF FREE-VARIABLES
	   (COND
	    ((ATOM INSTR)
	     (BLURB NIL |#| |Free| |variables| |and| |where| |they| |appear| |:| CR)
	     (SHOW-VARIABLES-WITH-GOALS FREE-VARIABLES (CAR STATE-STACK)))
	    (T
	     (BLURB (FREE-VARIABLES)
		    |The|
		    (PLURAL? FREE-VARIABLES (PROGN |free| |variables|)
			     (PROGN |only| |free| |variable|))
		    |in| |the| |current| |proof| |state|
		    (PLURAL? FREE-VARIABLES |are| |is|)
		    (!PPR-LIST FREE-VARIABLES (QUOTE |.|)) CR)))
	   (BLURB NIL |There| |are| |no| |free|
		  |variables| |in| |the| |current| |proof| |state| |.| CR))))

(DEFUN PRINT-HELP-WITH-SHOW-INDUCTIONS (INSTR)
  INSTR
  (LET ((STATE (CAR STATE-STACK)))
    (LET-FIELDS
      (COND
	(CURRENT-ADDR-R
	 (BLURB NIL CR WARNING |:| |You| |must| |be| |at| |the| |top| |of|
		|the| |conclusion| |to| |induct| |,| |and| |the| |following|
		|information| |assumes| |that| |you| |move| |to| |the| |top|
		|before| |attempting| |to| INDUCT |.| CR CR)))
      (LET ((SMRP (SCHEMA-MEASURE-RELATION-POSSIBILITIES HYPS CONC)))
	(BLURB (SMRP) |The| |heuristically| |chosen| |induction|
	       (PLURAL? SMRP |schemes are| |scheme is|)
	       |:| CR)
	(ITERATE FOR X IN SMRP
	      AS INDEX FROM 1
	      DO
	  (PROGN
	    (COND
	      ((CDR SMRP)
	       (blurb ((a (PACK (LIST INDEX (QUOTE |. |)))))
		      (!ppr a nil)))
	      (T (ISPACES 3 NIL)))
	    (pprind (car x) 3 0 nil)
	    (blurb nil cr)))))))

(DEFUN PRINT-HELP-WITH-SHOW-REWRITES (INSTR)
  (LET (APPLICABLE-REWRITES ASSUMPTIONS TYPE-ALIST PRINTED-HYPS
			    (STATE (CAR STATE-STACK)))
       (LET-FIELDS
	(COND
	 ((VARIABLEP CURRENT-TERM)
	  (BLURB (CURRENT-TERM) |It| |is| |impossible| |to| |rewrite| |a|
		 |variable| |,| |and| |the| |current| |subterm| |is| |the|
		 |variable| (!PPR CURRENT-TERM (QUOTE |.|)) CR))
	 (T
	  (SETQ ASSUMPTIONS
		(ASSUMPTIONS2 HYPS GOVERNORS))
	  (SETQ TYPE-ALIST
		(LET (TYPE-ALIST)
		     (FLESH-OUT-TYPE-ALIST
		      (MY-TYPE-ALIST-CLAUSE
		       (ITERATE FOR X IN ASSUMPTIONS
				COLLECT (DUMB-NEGATE-LIT X))))))
	  (COND
	   ((EQ TYPE-ALIST 'CONTRADICTION)
	    (BLURB NIL CONTRADICTION IN THE CURRENT ACTIVE HYPOTHESES AND GOVERNORS EXCL CR
		   |The| |command| S |should| |be| |useful| |in| |the| |current| |situation|
		   |.| CR))
	   (T
	    (SETQ ASSUMPTIONS
		  (FLESH-OUT-ASSUMPTIONS ASSUMPTIONS))
	    (SETQ APPLICABLE-REWRITES
		  (APPLICABLE-REWRITES
		   CURRENT-TERM
		   (TERM-ID-IFF CONC (REVERSE CURRENT-ADDR-R) (QUOTE IFF))))
	    (COND
	     ((NULL APPLICABLE-REWRITES)
	      (BLURB (INSTR (FLG (ATOM INSTR)))
		     |No| |rewrite| |rules| |apply| |to| |the| |current| |term|
		     (COND (FLG |.|)
			   (T |,| |even| |without| |the| |restriction| |you|
			      |supplied| |in| (!PPR INSTR (QUOTE |.|))))
		     CR))
	     ((AND (CONSP INSTR)
		   (NOT (SYMBOLP (CADR INSTR)))
		   (OR (NOT (INTEGERP (CADR INSTR)))
		       (< (CADR INSTR) 1)))
	      (BLURB NIL CR |Sorry| |,| |but| |the| SHOW-REWRITES |command|
		     |must| |either| |have| |no| |arguments| |,| |or| |else|
		     |must| |have| |an| |argument| |which| |is| |either| |T|
		     |a| |positive| |integer| |,| |or| |the| |name| |of| |an|
		     |applicable| |rewrite| |rule| |.| |Use|
		     |(HELP-LONG SHOW-REWRITES)| |for| |details| |.| CR))
	     (T
	      (ITERATE FOR LEMMA-SUBST IN APPLICABLE-REWRITES
		       AS I FROM 1
		       WHEN
		       (COND ((ATOM INSTR) T)
			     ((EQ (CADR INSTR) T) T)
			     ((SYMBOLP (CADR INSTR))
			      (EQ (ACCESS REWRITE-RULE NAME (CAR LEMMA-SUBST))
				  (CADR INSTR)))
			     (T (INT= I (CADR INSTR))))
		       WITH NEWTERM AND LEMMA-VARS-HYPS AND APOLOGY-FLG = T
		       DO
		       (PROGN
			(SETQ APOLOGY-FLG NIL)
			(SETQ LEMMA-VARS-HYPS
			      (LEMMA-VARS-HYPS-REMAINING
			       (CAR LEMMA-SUBST)
			       (CDR LEMMA-SUBST)
			       NIL ASSUMPTIONS TYPE-ALIST T))
			(SETQ NEWTERM
			      (SUBLIS-VAR 
			       (CDR LEMMA-SUBST)
			       (FARGN (ACCESS REWRITE-RULE CONCL
					      (CAR LEMMA-VARS-HYPS))
				      2)))
			(SETQ NEWTERM
			      (COND ((EQUAL INSTR '(SHOW-REWRITES T))
				     (COND
				      ((VARIABLEP NEWTERM)
				       NEWTERM)
				      ((FQUOTEP NEWTERM)
				       (UNTRANSLATE NEWTERM))
				      (T
				       (CONS (FFN-SYMB NEWTERM)
					     '(|<arguments>|)))))
				    (T (UNTRANSLATE-ABB NEWTERM ABBREVIATIONS))))
			(SETQ PRINTED-HYPS
			      (COND ((EQUAL INSTR '(SHOW-REWRITES T))
				     (let ((len (LENGTH (CDDR LEMMA-VARS-HYPS))))
				       (LIST '|There|
					     (if (int= len 1) '|is| '|are|)
					     len
					     '|top-level|
					     (if (int= len 1) '|hypothesis.| '|hypotheses.|))))
				    (T
				     (ITERATE FOR H IN (CDDR LEMMA-VARS-HYPS)
					      COLLECT
					      (UNTRANSLATE-ABB H ABBREVIATIONS)))))
			(BLURB (I
				(NAME (ACCESS REWRITE-RULE NAME
					      (CAR LEMMA-SUBST)))
				(disabled-flg
				 (disabledp
				  (ACCESS REWRITE-RULE NAME (CAR LEMMA-SUBST))))
				NEWTERM
				PRINTED-HYPS)
			       (!PPR I (QUOTE |.|))
			       (!PPR NAME NIL)
			       (cond (disabled-flg |(disabled)|))
			       CR |#| |New| |term| |:| (!PPR NEWTERM NIL)
			       CR |#| |Hypotheses| |:|
			       (COND (PRINTED-HYPS (!PPR PRINTED-HYPS NIL)) (T |<none>|))
			       CR))
		       FINALLY
		       (COND
			(APOLOGY-FLG
			 (BLURB ((N (LENGTH APPLICABLE-REWRITES))
				 (ARG (CADR INSTR))
				 (CONDITION (INTEGERP (CADR INSTR))))
				CR |There| (PLURAL? N |are| |is|) (!PPR N NIL)
				|rewrite| (PLURAL? N |rules| |rule|)
				|which| (PLURAL? N |apply| |applies|)
				|to| |the| |current| |term| |.|
				|Unfortunately| |,| |you| |asked| |for|
				(COND (CONDITION
				       |rule| |number| (!PPR ARG (QUOTE |.|)))
				      (T
				       |a| |rule| |named| (!PPR ARG NIL) |,|
				       |and| |no| |such| |rule| |applies| |.|))
				CR)))))))))))))

(DEFUN PRINT-HELP-WITH-UNDONE-INSTRUCTIONS (&OPTIONAL INSTR)
  INSTR
  (BLURB (UNDONE-INSTRS)
	 |Here| |is| |the| |list| |of| |instructions| |which| |were|
	 |not| |run| |successfully| |at|
	 |the| |last| |invocation| |of| |a| |command| |which| |keeps|
	 |track| |of| |undone| |instructions| |,| |including|
	 |for| |example| |PROVE-LEMMA| |:|
	 (COND (UNDONE-INSTRS
		(!PPR UNDONE-INSTRS (QUOTE |.|)) CR)
	       (T
		|<none>| CR))))

(DEFUN UNDONE-INSTRUCTIONS () (PRINT-HELP-WITH-UNDONE-INSTRUCTIONS))
